/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
#include "fm.h"
#include <icons/dirwIcon>

static char SccsID[] = "@(#)windows.c	1.2\t1/27/93";            /* sccs keyword */

/* placeWindow: **************************************************************
 * move the widget to the specified location and raise it to the top 
 ****************************************************************************/

placeWindow(w, x, y)
     dirWindow *w;
     int x,y;
{
  XtMoveWidget(w->fmTop, x, y);
  XMapRaised(XtDisplay(w->fmTop), XtWindow(w->fmTop));
}

/* closeDirWindow: ************************************************************
 * unmap the given dir window and map the associated icon instead
 ****************************************************************************/

/*ARGSUSED*/
void closeDirWindow(wd, client_data, call_data)
     Widget wd;
     caddr_t client_data;
     caddr_t call_data;
{
  dirWindow *wnd = (dirWindow *) client_data;

  if (! wnd->dirwIcon.iconShell) 
    defineIcon(wnd);

  XtPopup(wnd->dirwIcon.iconShell, XtGrabNone);
  XtPopdown(wnd->fmTop);

  XDefineCursor(XtDisplay(Head->root), 
		XtWindow(wnd->dirwIcon.icon), PanelCursor);
}

/* openDirWindow: ************************************************************
 * unmap the given icon and map the associated window instead; preserve
 * the location of the icon widget 
 ****************************************************************************/
/*ARGSUSED*/
void openDirWindow(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
  dirWindow *wnd = (dirWindow *) client_data;

  XtPopup(wnd->fmTop, XtGrabNone);
  XtPopdown(wnd->dirwIcon.iconShell);

  XDefineCursor(XtDisplay(wnd->workArea),
		XtWindow(wnd->workArea), PanelCursor);
}

/* redrawIcon: **************************************************************
 * callback function that is called when the icon is exposed: it redraws the 
 * icon bitmap;
 ****************************************************************************/
/*ARGSUSED*/
void redrawIcon(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
  XCopyArea(XtDisplay(Head->root), iconPixmap, XtWindow(w), Head->fm_gc,
	    0, 0, dirwIcon_width, dirwIcon_height, 0, 0);
}
/*ARGSUSED*/
/* closeTree: ************************************************************
 * delete the tree window
 ****************************************************************************/
/*ARGSUSED*/
void closeTree(wd, client_data, call_data)
     Widget wd;
     caddr_t client_data;
     caddr_t call_data;
{
  deleteTreeWindow((dirWindow *) client_data);
}

/* hideAttrDisplay: **********************************************************
 * callback function of the attribute editor's close box to unmap the 
 * attribute display
 ****************************************************************************/

/* ARGSUSED */
void hideAttrDisplay(wd, client_data, call_data)
     Widget wd;
     caddr_t client_data;
     caddr_t call_data;
{
  XtPopdown(Head->attrShell);
}

/* attrEditHide: *************************************************************
 * attribute editor function to unmap the attribute editor
 ****************************************************************************/

/* ARGSUSED */
attrEditHide(editor)
     Editor editor;
{
  XtPopdown(Head->attrShell);
}  

/* displayDirWindow: *********************************************************
 * display the specified directory window and fill in some values that may be
 * obtained only once the widget has been realized and the corresponding 
 * windows have been created. This is done however only when the widget is
 * mapped for the first time.
 *****************************************************************************/

displayDirWindow(wnd)
     dirWindow *wnd;
{
  XtPopup(wnd->fmTop, XtGrabNone);
  XFlush(XtDisplay(wnd->fmTop));
  setDirBoxSizeHints(wnd);
}

/* redrawTreeWindow: *********************************************************
 * callback for exposure events that occur with respect to the plane widget
 * that holds all the buttons 
 ****************************************************************************/

/* ARGSUSED */
void redrawTreeWindow(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String params;
     int *num_params;
{
  treeWindow *tw;
  Arg args[1];
  int cnt;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &tw); cnt++;
  XtGetValues(w, args, cnt);

  if (tw ) {
    XDefineCursor(XtDisplay(Head->root), XtWindow(tw->w), TreeCursor);
    XDrawSegments(XtDisplay(Head->root), XtWindow(tw->plane), Head->fm_gc, 
		  			tw->connections, tw->conCnt);
  }
}

/* fmRedraw: ****************************************************************
 * callback that is called when a configure event occures with respect to the 
 * work area; 
 ****************************************************************************/

void fmRedraw(w, event, params, num_params)
     Widget w;
     XExposeEvent *event;
     String *params;
     Cardinal *num_params;
{
  int cnt;
  Arg args[2];
  dirWindow *wnd = NULL;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(w, args, cnt);

  if (wnd)
    setDirBoxSizeHints(wnd);	
}

/* fmSetCursor: **************************************************************
 * sets the cursor when the panel window is created
 ****************************************************************************/

void fmSetCursor(w, event, params, num_params)
     Widget w;
     XExposeEvent *event;
     String *params;
     Cardinal *num_params;
{
  int cnt;
  Arg args[2];
  dirWindow *wnd = NULL;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &wnd); cnt++;
  XtGetValues(XtParent(w), args, cnt);

  if (! wnd)
    return;

  XDefineCursor(XtDisplay(wnd->workArea),
		  XtWindow(wnd->workArea), PanelCursor);
}
