/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
#include <stdio.h>
#include "ttys.h"

#ifdef DEBUG
#include <mnemosyne.h>
#endif

static void resize();
static Setenv();
static char *strindex();
static void Alert();

/**************************************************************************
 *                      Public Functions
 ***************************************************************************/

/* removeCtrlTerminal: *******************************************************
 * remove the association between the esh-process and its current control 
 * terminal 
 ****************************************************************************/

removeCtrlTerminal()
{
  int tty;
  /* open control terminal */
  if ((tty = open ("/dev/tty", O_RDWR, 0)) < 0) {
    exError("removeCtrlTerminal", "Cannot open control terminal!", NULL);
    return(-1);
  }
  
  /* remove association between esh and its control terminal */
  if (ioctl (tty, TIOCNOTTY, 0) == -1) {
    exError("removeCtrlTerminal", "Cannot unset control terminal!", NULL);
    return(-1);
  }
  
  return(close (tty));
}

/* adjustEnvironment: ********************************************************
 * set some new values in the environemnt of the current process, namely the 
 * requested size (cols and lines) of the terminal.
 ****************************************************************************/

adjustEnvironment(lines, cols, name)
     int lines, cols;
     char *name;
{
  int i;
  char termcap [1024];
  char newtc [1024];
  char *ptr;
  char **envnew;			/* new environment */
  char *TermName = NULL;
  
  /* first set new termcap in environ to reflect window-size for curses */
  
  ptr = termcap;
  
  if(tgetent(ptr, XTERM) != 1)
    fatal("adjustEnvironment","No termcap entry found", NULL);
  
  TermName = XTERM;
  resize(cols, lines, TermName, termcap, newtc, name);
  
  /* copy the environment before Setenving */
  for (i = 0 ; environ [i] != NULL ; i++) ;
  /*
   * The `5' is the number of Setenv() calls which may add
   * a new entry to the environment.  The `1' is for the
   * NULL terminating entry.
   */
  envnew = (char **) Calloc (i + (5 + 1), sizeof(char *));
  bcopy((char *)environ, (char *)envnew, i * sizeof(char *));
  environ = envnew;
  Setenv ("TERM=", TermName);
  if(!TermName)
    *newtc = 0;
  Setenv("TERMCAP=", newtc);
}

/* ttyGet: *******************************************************************
 * get several of the tty settings and modify them slightly 
 ****************************************************************************/

void ttyGet()
{
  int tty;
  
  if ((tty = open ("/dev/tty", O_RDWR, 0)) < 0) {
    if (errno != ENXIO) 
      Alert("ttyGet: open control terminal", errno);
    else {
      sg = d_sg;
      tc = d_tc;
      discipline = d_disipline;
      ltc = d_ltc;
      lmode = d_lmode;
    }
  } else {
    /* get a copy of the current terminal's state */
    
    if(ioctl(tty, TIOCGETP, &sg) == -1)
      Alert("ttyGet (TIOCGETP)", errno);
    if(ioctl(tty, TIOCGETC, &tc) == -1)
      Alert("ttyGet (TIOCGETC)", errno);
    if(ioctl(tty, TIOCGETD, &discipline) == -1)
      Alert("ttyGet (TIOCGETD)", errno);
    if(ioctl(tty, TIOCGLTC, &ltc) == -1)
      Alert("ttyGet (TIOCGLTC)", errno);
    if(ioctl(tty, TIOCLGET, &lmode) == -1)
      Alert("ttyGet (TIOCGET)", errno);
    close (tty);
  }
  
  /* set the new terminal's state to be the old one's 
     with minor modifications for efficiency */
  
  sg.sg_flags &= ~(ALLDELAY | XTABS | CBREAK | RAW);
  sg.sg_flags |= ECHO | CRMOD;
  /* make sure speed is set on pty so that editors work right*/
  sg.sg_ispeed = B9600;
  sg.sg_ospeed = B9600;
  /* reset t_brkc to default value */
  tc.t_brkc = -1;
}

/* ttySet: *******************************************************************
 * set the settings that have been obtained by a call to function ttyGet
 ****************************************************************************/

void ttySet(tty)
     int tty;
{
  if (ioctl (tty, TIOCSETP, &sg) == -1)
    Alert("ttySet (TIOCSETP)", errno);
  if (ioctl (tty, TIOCSETC, &tc) == -1)
    Alert("ttySet (TIOCSETC)", errno);
  if (ioctl (tty, TIOCSETD, &discipline) == -1)
    Alert("ttySet (TIOCSETD)", errno);
  if (ioctl (tty, TIOCSLTC, &ltc) == -1)
    Alert("ttySet TIOCSLTC)", errno);
  if (ioctl (tty, TIOCLSET, &lmode) == -1)
    Alert("ttySet (TIOCLSET)", errno);
}

/* setWinSize: ***************************************************************
 * inform the tty controler of the new window size
 ****************************************************************************/

void setWinSize(tty, lines, cols)
     int tty, lines, cols;
{
#ifdef sun
  struct ttysize ts;
  /* tell the tty of our son the window size */
  
  ts.ts_lines = lines;
  ts.ts_cols  = cols;
  ioctl (tty, TIOCSSIZE, &ts);
#endif
}

/**************************************************************************
 *                      Private Functions
 ***************************************************************************/

/* resize: *******************************************************************
 * replace the values for the li# and co# capabilities in the termcap string 
 * of the current process to reflect the new colsss and lines.
 ****************************************************************************/

static void resize (cols, lines, TermName, oldtc, newtc, name)
     int cols, lines;
     char *TermName;
     register char *oldtc, *newtc;
     char *name;				/* application name */
{
  register char *ptr1, *ptr2;
  register int i;
  register int li_first = 0;
  register char *temp;
  char *index(), *strindex();
  
  if ((ptr1 = strindex (oldtc, "co#")) == NULL)
    fatal(name, "can't find co# in termcap string -- ", TermName, " --", NULL);

  if ((ptr2 = strindex (oldtc, "li#")) == NULL)
    fatal(name, "can't find li# in termcap string -- ", TermName, " --", NULL);

  if(ptr1 > ptr2) {
    li_first++;
    temp = ptr1;
    ptr1 = ptr2;
    ptr2 = temp;
  }
  ptr1 += 3;
  ptr2 += 3;
  strncpy (newtc, oldtc, i = ptr1 - oldtc);
  newtc += i;
  sprintf (newtc, "%d", li_first ? lines : cols);
  newtc += strlen(newtc);
  ptr1 = index (ptr1, ':');
  strncpy (newtc, ptr1, i = ptr2 - ptr1);
  newtc += i;
  sprintf (newtc, "%d", li_first ? cols : lines);
  ptr2 = index (ptr2, ':');
  strcat (newtc, ptr2);
}

/* Setenv: ******************************************************************
 * sets the value of var to be arg in the Unix 4.2 BSD environment env.
 * Var should end with '=' (bindings are of the form "var=value").
 * This procedure assumes the memory for the first level of environ
 * was allocated using calloc, with enough extra room at the end so not
 * to have to do a realloc().
 ****************************************************************************/

static Setenv (var, value)
     register char *var, *value;
{
  extern char **environ;
  register int indx = 0;
  register int len = strlen(var);
  
  while (environ [indx] != NULL) {
    if (strncmp (environ [indx], var, len) == 0) {
      /* found it */
      environ[indx] = (char *)Malloc (len + strlen (value) + 1);
      strcpy (environ [indx], var);
      strcat (environ [indx], value);
      return;
    }
    indx ++;
  }
  
  environ [indx] = (char *) malloc (len + strlen (value) + 1);
  strcpy (environ [indx], var);
  strcat (environ [indx], value);
  environ [++indx] = NULL;
}

/* strindex: ****************************************************************
 * returns a pointer to the first occurrence of s2 in s1,
 * or NULL if there are none.
 ****************************************************************************/

static char *strindex (s1, s2)
     register char	*s1, *s2;
{
  register char	*s3;
  char	*index();
  
  while ((s3=index(s1, *s2)) != NULL) {
    if (strncmp(s3, s2, strlen(s2)) == 0)
      return (s3);
    s1 = ++s3;
  }
  return (NULL);
}

/* Alert: ********************************************************************
 * output an error message that consists of a user specified string, the value 
 * of the errno variable and the system error string
 ****************************************************************************/

static void Alert(message, code)
     char *message;
     int code;
{
  extern char *sys_errlist[];
  char m[512];
  
  sprintf(m,"%s <--> (%d) %s\n",message, code, sys_errlist[code]);
  fprintf(stderr, m);
}



