/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
#include <exLayout.h>
#include <exBasics.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <exThermometer.h>
#include "exEdit.h"

static char msg[200];
static struct stat sbuf;
static Boolean show_therm = False;		/* show thermometer or  ot */

char *ETtext();

/****************************************************************************
 * EIOreadf - read file into buffer
 * RETURN: Sucessful read?
 ****************************************************************************/
Boolean EIOreadf(bf)
     register BufInfo *bf;
{
  FILE *f;
  char s[MAXLINE + 2]; 			/* room for newline */
  int slen, size, sc, cnt = 0;
  Boolean err = False;
  
  if ((f = fopen(bf->filename, "r")) == NULL) {
    sprintf(msg, "Cant't open -- %s -- !", bf->filename);
    ESmsg(bf, msg, True);
    (void)ETappend(bf, "");
    return(False);
  }

  if (fstat(fileno(f), &sbuf) < 0) {
    size = -1;
    show_therm = False;
  } else {
    size = sbuf.st_size / THERM_UPDATE_INCR;
    sprintf(msg, "Reading file --%s-- (%d bytes)", bf->filename, sbuf.st_size);
    initThermometer(XtDisplay(bf->w), bf->w, 100, 100, strlen(msg), msg, NULL);
    show_therm = True;
  }
  
  sc = 0;
  while (fgets(s, sizeof(s), f) != NULL) {
    slen = strlen(s) - 1;
    if (show_therm) {
      cnt += slen; 
      if (cnt > size * sc) {
	setThermometer(-1, -1, (int)(sc * 100 / THERM_UPDATE_INCR));
	sc++;
      }
    }
    if (s[slen] != '\n') {
      ESmsg(bf, "Line too long", True);
      err = True;
      break;
    }
    s[slen] = '\0';
    if (!ETappend(bf, s)) {
      err = True; 
      break;
    }
  }

  if (show_therm) {
    setThermometer(-1, -1, 100);
    hideThermometer(10);
  }

  if(ferror(f) && !err) {
    ESmsg(bf, "Read error", True);
    err = True;
  }
  if (fclose(f) != 0 && !err) {
    ESmsg(bf, "Read error", True);
    err = True;
  }
  if (!err) {
    sprintf(msg, "Successfully read -- %s -- (%d bytes) !", 
	    bf->filename, cnt);
    ESmsg(bf, msg, False);
  } else
    Bell();
  if(ETnumlines(bf) < 1)
    (void)ETappend(bf, "");
  return((Boolean)(!err));
}


/*---------------------------------------------------------------------------
  EIOinsertf - insert file into buffer
  RETURN: Sucessful read?
  ---------------------------------------------------------------------------*/
Boolean EIOinsertf(bf, filename)
     register BufInfo *bf;
     char *filename;
{

  FILE *f;
  char s[MAXLINE + 2]; 			/* room for newline */
  int slen, size, sc, cnt = 0;
  Boolean err = False;
  int flags;

  if ((f = fopen(filename, "r")) == NULL) {
    sprintf(msg, "Cant't open -- %s -- !", filename);
    ESmsg(bf, msg, True);
    return(False);
  }

  if (fstat(fileno(f), &sbuf) < 0) {
    size = -1;
    show_therm = False;
  } else {
    size = sbuf.st_size / THERM_UPDATE_INCR;
    sprintf(msg, "Inserting file --%s-- (%d bytes)", filename, sbuf.st_size);
    initThermometer(XtDisplay(bf->w), bf->w, 
			         100, 100, strlen(msg), msg, NULL);
    show_therm = True;
  }
  
  flags = bf->flags;
  bf->flags &= ~ED_MAPPED;		/* don't show visible changes */
  sc = 0;
  while (fgets(s, sizeof(s), f) != NULL) {
    slen = strlen(s) - 1;
    if (show_therm) {
      cnt += slen; 
      if (cnt > size * sc) {
	setThermometer(-1, -1, (int)(sc * 100 / THERM_UPDATE_INCR));
	sc++;
      }
    }

    if (s[slen] != '\n') {
      ESmsg(bf, "Line too long", True);
      err = True;
      break;
    }
    if (! insertText(bf, s, &bf->curline, bf->curpos)) {
      err = True;
      break;
    }
  }

  if (show_therm) {
    setThermometer(-1, -1, 100);
    hideThermometer(10);
  }

  bf->flags = flags;

  if(ferror(f) && !err) {
    ESmsg(bf, "Insertion error", True);
    err = True;
  }
  if (fclose(f) != 0 && !err) {
    ESmsg(bf, "Read error", True);
    err = True;
  }
  if (!err) {
    sprintf(msg, "Successfully inserted -- %s -- (%d bytes)!", filename, cnt);
    ESmsg(bf, msg, True);
  } else
    Bell();

  if(ETnumlines(bf) < 1)
    (void)ETappend(bf, "");
  return((Boolean)(!err));
}


/****************************************************************************
 * EIOwritef - write buffer to file
 * RETURN: Successful write?
 ***************************************************************************/

Boolean EIOwritef(bf)
     register BufInfo *bf;
{
  FILE *f;
  register line, sc, size, cnt = 0;
  register char *t;
  Boolean err = False;

  if ((f = fopen(bf->filename, "w")) == NULL) {
    sprintf(msg, "Can't open -- %s -- !", bf->filename);
    ESmsg(bf, msg, True);
    return(False);
  }

  sc = 0;

  for (line = 0, size = 0; (t = ETtext(bf, line)) != NULL; line++)
    size += strlen(t);
  sprintf(msg, "Writing --%s-- (%d bytes)", bf->filename, size);
  initThermometer(XtDisplay(bf->w), bf->w, 100, 100, strlen(msg), msg, NULL);
  size /= THERM_UPDATE_INCR;

  for (line = 0; (t = ETtext(bf, line)) != NULL; line++) {
    cnt += strlen(t);
    if (cnt > size * sc) {
      setThermometer(-1, -1, (int)(sc * 100 / THERM_UPDATE_INCR));
      sc++;
    }
    
    if (fputs(t, f) == EOF || putc('\n', f) == EOF) {
      err = True;
      break;
    }
  }
  
  if (fclose(f) == EOF)
    err = True;

  if(err) {
    ESmsg(bf, "Write error", True);
    Bell();
    return(False);
  }

  setThermometer(-1, -1, 100);
  hideThermometer(10);

  sprintf(msg, "Successfully wrote -- %s -- (%d bytes)!", 
	  bf->filename, cnt);
  ESmsg(bf, msg, True);
  return(True);
}
