/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
#include <X11/IntrinsicP.h>
#include <Xm/Form.h>
#include <Xm/DrawingA.h>
#include <Xm/ScrollBar.h>
#include <X11/cursorfont.h>
#include <exBasics.h>
#include <exLayout.h>

#include "exEdit.h"
#include "display.h"

static Cursor hairCursor;		/* the editor's cursor */
static Cursor scrollCursor;		/* scrollbar cursor */

extern char *ETtext();
extern void exEditInitGCs();
void csrChrForget();

extern XLookupString();

void PSend();
void HandleButtons();
void redrawWindow();
void normalize();		/* */
void inschar();			/* insert a character at the cursor position */
static Boolean hastab();	/* look if tabs exist in string */
Boolean insline();	
Boolean checkScroll();
void delline();			/* */
static Boolean clearbuf();	/* */
static Boolean chgchk();	/* */
BufInfo *bufinfoptr();		/* */
void status();			/* */
void splitline();		/* split line in two at specified column */
void joinlines();		/* append the next line to the current */
void delchar();			/* */
void readfile();		/* */
void writefile();		/* */
void killbuf();			/* */
Boolean wrdchr();
void exEditSetScrollbar();

/* editor translation functions */
void redrawEditor();
void exCursorPos();
void exCursorSelect();
void exCursorDeselect();
void exCursorCopy();
void exCursorCut();
void exCursorPaste();
void exKey();
void exFocusIn();
void exFocusOut();
void configureEditor();
char *EMtext();

/* scrollbar callback functions */
void doScroll();

static BufInfo bufinfo[MAXBUF];	/* the editor buffer table */

static char msg[200];

static char editor_translations[] = 
  " <Expose>			: redrawEditor()     \n \
    <ConfigureNotify>		: configureEditor()  \n \
    <KeyPress>                	: exKey()            \n \
    Shift <Btn1Down>		: exCursorCopy()     \n \
    Shift <Btn2Down>		: exCursorCut()      \n \
    Shift <Btn3Down>		: exCursorPaste()    \n \
    <Btn1Down>			: exCursorPos()      \n \
    <Btn2Down>			: exCursorSelect()   \n \
    <Btn3Down>          	: exCursorDeselect()";

static char main_translations[] = 
  " <EnterWindow>		: exFocusIn()      \n \
    <LeaveWindow>		: exFocusOut()";

static XtActionsRec editor_actions[] = {
  { "configureEditor",    configureEditor },
  { "redrawEditor",       redrawEditor },
  { "exCursorPos",        exCursorPos },
  { "exCursorSelect",     exCursorSelect },
  { "exCursorDeselect",   exCursorDeselect },
  { "exCursorCopy",       exCursorCopy },
  { "exCursorCut",        exCursorCut },
  { "exCursorPaste",      exCursorPaste },
  { "exKey",              exKey },
  { "exFocusIn",          exFocusIn },
  { "exFocusOut",         exFocusOut }
};

static XtTranslations ptt = NULL;
static XtTranslations ptt1 = NULL;

static char text[MAXLINE + 1]; 

/* _initialise: ***************************************************************
 * do some initialisation 
 ****************************************************************************/

static _initialise(d)
     Display *d;
{
  static Boolean _initialised = False;

  if (_initialised)
    return;

  exEditInitGCs(d);

  hairCursor   = XCreateFontCursor(d, XC_xterm);
  scrollCursor = XCreateFontCursor(d, XC_sb_v_double_arrow);

  _initialised = True;
}

/* bufinfoptr: ****************************************************************
 * convert buffer number to pointer to info structure
 *****************************************************************************/

BufInfo *bufinfoptr(bnum) 
     int bnum;					/* buffer number */ 
{ 
  if (bnum < 0 || bnum >= MAXBUF) 
    fatal("bufinfoptr - bad buffer num", "bufinfoptr", NULL); 
  return(&bufinfo[bnum]); 
}

/* phys2screen: **************************************************************
 * convert a position within a text-buffer into the corresponding position on 
 * the screen;
 ****************************************************************************/

Boolean phys2screen(bf, line, pos, row_ret, col_ret, lm)
     BufInfo *bf;
     int line;		/* position in text buffer */
     int pos;		/* position in text buffer line */
     int *row_ret;	/* row on screen */
     int *col_ret;	/* col on screen */
     Boolean lm;	/* check left margin */
{
  int col, prevcol, prevrow, p, row, l;
  register char *s;
  int eswidth;

  if (! (bf->flags & ED_LINEBREAK))
    eswidth = ED_MAXWIDTH;
  else
    eswidth = ESwidth(bf);

  row = 0;
  for (l = bf->topline; l < line; l++) 
    row += screenLines(bf, l);
  
  if (row >= bf->lines) {
    *row_ret = row;
    *col_ret = 0;
    return(False);
  }

  s = ETtext(bf, line);

  if (! s) {
    *col_ret = 0;
    *row_ret = row;
    return(False);
  }

  col = prevcol = 0;
  prevrow = row;

  for (p = 0; p <= pos; p++) {
    prevcol = col;
    prevrow = row;
    if (s[p] == EOS)
      break;
    if (s[p] == '\t')
      col = col + TABINTERVAL - (col % TABINTERVAL);
    else 
      col++;

    if ((col >= eswidth) && s[p + 1]) {
      col = 0;
      row++;
    }
  }

  if ((! (bf->flags & ED_LINEBREAK)) && lm) {	/* check left margin */
    while (prevcol < bf->lmarg)
      bf->lmarg -= MARGIN_JUMP;
    while (prevcol >= bf->lmarg + bf->cols)
      bf->lmarg += MARGIN_JUMP;
  }

  *col_ret = prevcol - bf->lmarg;
  *row_ret = prevrow;

  return(*row_ret < bf->lines ? True : False);
}

/* screen2phys: **************************************************************
 * convert a screen position into the corresponding position in the text buffer
 ****************************************************************************/

Boolean screen2phys(bf, row, col, ret_line, ret_pos, lm)
     BufInfo *bf;
     int row;			/* screen line */
     int col;			/* screen column */
     int *ret_line;		/* text buffer line */
     int *ret_pos;		/* text buffer pos */
     Boolean lm;		/* correct for left margin */
{
  register char *s;
  int pos, offset, cc, l, sl, r,  
  	w = ESwidth(bf),
  	numlines = ETnumlines(bf);

  /* calculate physical line */
  r = 0;
  for (l = bf->topline; ; l++) { 
    if (l >= numlines) {			/* out of text-lines */
      *ret_line = -1;
      *ret_pos = -1;
      return(False);
    }
    r += (sl = screenLines(bf, l));
    if (r > row)
      break;
  }

  if (! (s = ETtext(bf, l))) {
    *ret_pos = -1;
    *ret_line = -1;
    return(False);
  }

  if (bf->flags & ED_LINEBREAK) {		/* multi-line editor */
    offset = row - (r - sl); 	  		/* vertical offset */
    col = col % w;
  } else 
    offset = row = 0;
  
  for (cc = 0, pos = 0; s[pos]; pos++) {
    if (offset == 0) {				/* single row, no linebreak */
      if (cc >= col + bf->lmarg)
	break;
    } else if (cc >= w) {
      cc = 0;
      offset--;
    }

    if (s[pos] == '\t')
      cc += TABINTERVAL - cc % TABINTERVAL;
    else 
      cc++;
  }

  if (! s[pos] && lm) 				/* at end of line */
    while (cc < bf->lmarg)		      	/* adjust left margin */
      bf->lmarg -=  MARGIN_JUMP;

  *ret_pos = pos;
  *ret_line = l;

  return(True);
}

/* setcol: ********************************************************************
 * set current column corresponding to current position; check that row is
 * in visible area and decrement current line appropriately.
 *****************************************************************************/

void setcol(bf) 
     register BufInfo *bf;
{ 
  while (! phys2screen(bf, bf->curline, bf->curpos, 
		       			&bf->currow, &bf->curcol, True)){
    bf->curline--;

    if (bf->curline < 0) {
      bf->curline = 0;
      return;
    }
  }
}

/* setpos: ********************************************************************
 * set current position corresponding to current column
 *****************************************************************************/

void setpos(bf) 
     register BufInfo *bf;
{ 
  int err = 0;

  screen2phys(bf, bf->currow, bf->curcol, &bf->curline, &bf->curpos, True);

  if (bf->curline < 0) {
    err = 1;
    bf->curline = ETnumlines(bf) - 1;
  } 
  
  if (bf->curpos < 0) {
    err = 1;
    bf->curpos = ETlen(bf, bf->curline);
  }

  if (err)
    setcol(bf);
}

/* detab: ********************************************************************
 * replace tabs with blanks; return True if expanded string did fit, False 
 * otherwise
 ****************************************************************************/

static Boolean detab(src, dst, dstsize) 
     char *src;				/* source string */ char
          *dst;				/* destination string */ 
     int dstsize;			/* size of destination string */ 
{ 
  int srcpos, dstpos, j; 

  dstpos = 0;
  for (srcpos = 0; src[srcpos] != EOS; srcpos++) 
    if (src[srcpos] == '\t') 
      for (j = TABINTERVAL - dstpos % TABINTERVAL; j > 0; j--) {
	if (dstpos >= dstsize - 1) { 	 
	  dst[dstpos] = EOS;
	  return(False); 
	} 	
	dst[dstpos++] = ' '; 
      } else { 
	if (dstpos >= dstsize - 1) { 	
	  dst[dstpos] = EOS; 	
	  return(False); 
	} dst[dstpos++] = src[srcpos]; 
      }

  dst[dstpos] = EOS; return(True);
}

/* EMtext: *******************************************************************
 * get text corresponding to line; return text or blank line if beyond buffer;
 * expand tabs before returning the line; EMtext returns a pointer to a static
 * text-buffer, that is overwritten each time EMtext is called. 
 *****************************************************************************/

char *EMtext(bf, line, verbose) 
     register BufInfo *bf;
     int line;			/* abs. phys. line */ 
     Boolean verbose;
{ 
  char *t;
  static char tabtext[MAXLINE + 1];

  if ((t = ETtext(bf, line)) == NULL) 
    return(""); 
  if (!hastab(t)) 
    return(t); 
  if (! detab(t, tabtext, sizeof(tabtext)) && verbose) {
      ESmsg(bf, "Line too long !", False); 
      Bell(); 
  }
  return(tabtext); 
}

/* hastab: ********************************************************************
 * determine if the string contains a tab
 *****************************************************************************/

static Boolean hastab(s) 
     char *s; 
{ 
  char *strchr();
  return((Boolean)(strchr(s, '\t') != NULL)); 
}

/* normalize: *****************************************************************
 * locate the cursor sensibly and make sure it is on the screen; handle left
 * and right shifts 
 *****************************************************************************/

void normalize(bf) 
     BufInfo *bf;
{ 
  int len, i;
  static prev_lmarg = 0;
  RECT rect;

  if (prev_lmarg < bf->lmarg) {				/* move left */
    ESshowcur(bf, False, bf->oldcol, bf->oldrow);	/* hide cursor */
    csrChrForget(bf);					
    RASG(&rect, 0, bf->lmarg - prev_lmarg, bf->lines, bf->cols); 
    PSslide(bf, &rect, DIR_LEFT, bf->lmarg - prev_lmarg);
    for (i = 0; i < bf->lines; i++)			/* draw vis. area */
      /*ESshowrowrange(bf, i, bf->cols - (bf->lmarg - prev_lmarg) - 1, bf->cols);*/
      ESdisplayLine(bf, i, bf->curpos - (bf->lmarg - prev_lmarg) - 1);
  } else if (prev_lmarg > bf->lmarg) {          	/* move right */
    ESshowcur(bf, False, bf->oldcol, bf->oldrow);
    csrChrForget(bf);
    RASG(&rect, 0, 0, bf->lines, bf->cols - bf->lmarg); 
    PSslide(bf, &rect, DIR_RIGHT, prev_lmarg - bf->lmarg);
    for (i = 0; i < bf->lines; i++)
      /*ESshowrowrange(bf, i, 0, prev_lmarg - bf->lmarg)*/
      ESdisplayLine(bf, i, bf->curpos - bf->cols - 1);
  }

  prev_lmarg = bf->lmarg;

  len = ETlen(bf, bf->curline); 
  if (bf->curpos > len)  
    bf->curpos = len;
  setcol(bf); 
}

/* csrChrForget ************************************************************* 
 * sets old cursor position to current one so that no redrawing of stored
 * cursor character takes place
 ***************************************************************************/

void csrChrForget(bf) 
     BufInfo *bf; 
{ 
  bf->oldrow = -1;
  bf->oldcol = -1; 
}

/* splitline: ****************************************************************
 * split line in two lines
 ****************************************************************************/

void splitline(bf) 
     register BufInfo *bf;
{ 
  int len = ETlen(bf, bf->curline), sl; 
  int nextLine = bf->curline + 1;
  
  if (len > MAXLINE) 
    fatal("splitline - line length !", "splitline", NULL);

  strcpy(text, ETtext(bf, bf->curline));
  sl = screenLines(bf, bf->curline);

  if (! ETinsline(bf, &nextLine))
    return;

  if ((bf->currow + 1) >= ESheight(bf))
    ESscroll(bf, 1, DIR_UP);
  else
    bf->currow++;

  if (bf->curpos < len) { 
    ETreplace(bf, nextLine, &text[bf->curpos]); 
    text[bf->curpos] = EOS; 
    ETreplace(bf, bf->curline, text); 
    if (sl == screenLines(bf, bf->curline))
      ESinsrow(bf, bf->currow);
    PSclear(bf, bf->currow - 1, bf->curcol);
    ESshowrow(bf, bf->currow, 0);
    csrChrForget(bf);
  }
  else
    ESinsrow(bf, bf->currow);    

  bf->curline++;
  bf->curcol = bf->curpos = 0;
}


/* joinlines: ****************************************************************
 * join two adjacent lines
 *****************************************************************************/

void joinlines(bf) 
     register BufInfo *bf;
{ 		   
  int line, sl, sldiff, sladd;
  register i;

  line = bf->curline; 
  if (line > ETnumlines(bf) - 2) { 
    Bell(); 
    return; 
  } /* last but one line is the last useful to perform operation on */

  if (ETlen(bf, line) + ETlen(bf, line + 1) > MAXLINE) {
    ESmsg(bf, "Line too long !", True);
    Bell();
    return;
  }
  
  sl = screenLines(bf, line + 1);
  sladd = screenLines(bf, line) + sl;          /* former range of rows */

  strcpy(text, ETtext(bf, line));
  strcat(text, ETtext(bf, line + 1));
  ETreplace(bf, line, text);
  ETdelline(bf, line + 1);

  ESshowrow(bf, bf->currow, bf->curcol);
  sldiff = sladd - screenLines(bf, line);
  for (i = 0; i < sldiff; i++)
    ESdelrow(bf, bf->currow + sl - i);
  
  csrChrForget(bf);

  setKillBuffer(bf, strCopy("\n"));
}


/* showline: ****************************************************************
 * show specified line with cursor on specified pos (scroll if necessary);
 * return True if screen is ok, False otherwise; 
 ****************************************************************************/

Boolean showline(bf, line, pos) 
     register BufInfo *bf;
     int line;			/* phys. line */
     int pos;			/* phys. pos */
{ 
  register numlines = ETnumlines(bf);
  int newtop, l;

  if (line < 0)
    line = 0;
  else if (line >= numlines)
    line = numlines - 1;

  if (line < bf->topline) 		/* line to top , scroll down */
    newtop = line;
  else {				/* line to bottom */
    int ll = lastLineOnScreen(bf);
    
    if (line <= ll) 			/* requested line already displayed */
      newtop = bf->topline;
    else 
      newtop = (bf->flags & ED_JUMPSCROLL) ? 
	line - (ll - bf->topline) / 2 : line - (ll - bf->topline);
  }

  bf->curline = line;
  bf->curpos = MIN(pos, ETlen(bf, line));

  if (newtop == bf->topline) {		/* need no scrolling */
    phys2screen(bf, bf->curline, bf->curpos, &bf->currow, &bf->curcol, True);
    setpos(bf);
    return(True);
  }

  /* scroll so that requested line is displayed in the screen window */

  l = bf->topline - newtop;		/* phys. lines to scroll */
  
  if (ABS(l) >= bf->lines / 2) {	/* redraw screen */
    bf->topline = newtop;
    phys2screen(bf, bf->curline, bf->curpos, &bf->currow, &bf->curcol, True);
    setpos(bf);
    csrChrForget(bf);
    ESshowall(bf);
    return(True);
  }

  /* scroll lines so that requested line fits on screen */

  if (l < 0) {				/* scroll up */
    /* requested line is to be displayed at the bottom; check if it fits 
     * on the screen 
     */
    
    l = bf->topline - topLine(bf, line);    
    scrollLines(bf, ABS(l), DIR_UP);
    return(True);
  } else { 				/* scroll down */
    scrollLines(bf, l, DIR_DOWN);
    return(True);
  }
}

/* delchar ********************************************************************
 * delete the character at the cursor position
 *****************************************************************************/

void delchar(bf) 
     register BufInfo *bf;
{ 
  int line, rowNo = 1, row, col; 
  Boolean ISTAB = False;

  line = bf->curline;
  if (bf->flags & ED_LINEBREAK)
    rowNo = screenLines(bf, line);

  if (ETgetchar(bf, line, bf->curpos) == '\t')
    ISTAB = True;

  if (!ETdelchar(bf, line, bf->curpos)) 
    return;

  if (! (bf->flags & ED_LINEBREAK) && (bf->curcol > ESwidth(bf)))
    return;

  ESdelchar(bf, bf->currow, bf->curcol, rowNo, ISTAB);

  if (ISTAB) {
    phys2screen(bf, line, 0, &row, &col, True);    
    PSclear(bf, row + screenLines(bf, line) - 1, ESrestLen(bf, line) - 1);
    ESdisplayLine(bf, line, bf->curpos);
  }

  csrChrForget(bf);
}

/* inschar: ******************************************************************
 * insert a character at the cursor position
 ****************************************************************************/

void inschar(bf, ch) 
     register BufInfo *bf;
     char ch; 
{ 
  int line;

  if (ch == '\n') { 
    splitline(bf); 
    return; 
  }

  line = bf->curline; 

  if (ETlen(bf, line) > MAXLINE) {		/* physical line limit */
    Bell();
    return;
   }

  ETinschar(bf, line, bf->curpos, ch); 	/* insert into physical line */

  if (! (bf->flags & ED_LINEBREAK) && bf->curcol > ESwidth(bf))
    return;

  ESinschar(bf, bf->currow, bf->curcol, ch);    /* insert into visible row */

  if (ch == '\t')
    ESdisplayLine(bf, line, bf->curpos); 

  bf->curpos++; 
  phys2screen(bf, bf->curline, bf->curpos, &bf->currow, &bf->curcol, True);
  
  csrChrForget(bf);
} 

/* insline: ******************************************************************
 * insert a line; return True if enough room, False otherwise
 ****************************************************************************/

Boolean insline(bf) 
     register BufInfo *bf;
{ 
  if (! ETinsline(bf, &bf->curline))
    return(False); 

  if (bf->currow >= ESheight(bf)) { 
    bf->currow--; 
    ESscroll(bf, 1, DIR_UP);
  } else 
    ESinsrow(bf, bf->currow); 

  bf->curcol = bf->curpos = 0;
  
  return(True); 
}


/* delline: ******************************************************************
 * delete the current line
 ****************************************************************************/

void delline(bf) 
     register BufInfo *bf;
{ 
  register i, sl = screenLines(bf, bf->curline);
  
  ETdelline(bf, bf->curline);

  for (i = 0; i < sl; i++)
    ESdelrow(bf, bf->currow + i); 

  bf->curcol = bf->curpos = 0; 
}

/* insertText: *************************************************************** 
 * insert text into the editor buffer
 ****************************************************************************/

insertText(bf, buffer, line, pos)
     register BufInfo *bf;
     char *buffer;              	/* text to append */
     int *line, pos;                  	/* position for insertion in buffer */
{
  register char *ptr = buffer;
  int lines = 0;			/* number of lines inserted */
  Boolean nl = False;
  
  text[0] = EOS;

  if (pos < ETlen(bf, *line))		/* save rest of line */
    strcpy(text, &bf->bufPtr->text[*line][pos]);

  while (*ptr) {    		     	/* insert string into buffer */
    if (*ptr == '\n') {
      nl = True;
      lines++;
      ETinschar(bf, *line, pos, EOS); 	/* insert text between newlines */
      (*line)++;
      if (! ETinsline(bf, line))
	return(-1);
      pos = 0;
    } else {
      if (*ptr != '\r')
	ETinschar(bf, *line, pos++, *ptr);
    }
    ptr++;
  }

  if (text[0] && nl) 
    ETinsstr(bf, *line, pos, text);
  
  if (bf->topline + bf->lines < *line)  /* change visible range */
    bf->topline = *line - bf->lines + 1;

  bf->curline = *line;  		/* cursor at end of inserted text */
  bf->curpos = pos;
  /*phys2screen(bf, bf->curline, bf->curpos, &bf->currow, &bf->curcol, True);*/
  /*setcol(bf);*/
  return(lines > 0 ? lines : 1);
}

/* appendText: **************************************************************
 * append text to the end of the editor buffer without updating the display
 ****************************************************************************/

appendText(editor, buffer)
     Editor editor;                      
     char *buffer;                               /* text to append */
{
  register col, lines;
  register char *ptr = buffer;
  register BufInfo *bf = bufinfoptr(editor);

  if (! buffer)
    return;

  lines = bf->bufPtr->numlines - 1;
  col = ETlen(bf, lines);
  
  /* append string to buffer */
  while (*ptr) {
    if (col >= MAXLINE) {		/* truncate line */
      col--;
      while (*ptr && *ptr != '\n')	/* position to next NL */
	ptr++;
      continue;
    } else if (*ptr == '\n') {          /* start new line */
      ETinschar(bf, lines, col, '\0');  /* finish old line */
      ETappend(bf, "");	                /* newline */
      col = 0;
      lines = bf->bufPtr->numlines - 1;	/* might be changed (autodelete) */
    } else if (*ptr == '\r') { 		/* don't insert CR */
    } else {				/* insert character */
      ETinschar(bf, lines, col++, *ptr);
    }
    ptr++;
  }
}

/* killbuf: ******************************************************************
 * clear buffer and give it one blank line 
 ****************************************************************************/

void killbuf(bf, forced)
     register BufInfo *bf;
     Boolean forced;
{
  if (clearbuf(bf, forced)) {
    ETappend(bf, "");
    ETunchanged(bf);
  }
}

/* clearbuf: ****************************************************************
 * clear buffer completely; return True of ok; False otherwise
 ***************************************************************************/

static Boolean clearbuf(bf, forced)
     register BufInfo *bf;
     Boolean forced;
{
  if (! forced && ! chgchk(bf))
    return(False);

  ETclear(bf);
  bf->topline = bf->currow = bf->curcol = bf->curpos = bf->curline = 0;
  reset_region(bf);
  csrChrForget(bf);
  bf->oldFnct = (int (*)()) 0;
  ESclear(bf);
  return(True);
}

/* chgchk: ******************************************************************
 * ask user to approve desctruction of changed buffer; return True if ok to 
 * destroy, False otherwise
 ***************************************************************************/

static Boolean chgchk(bf)
     BufInfo *bf;
{
  if (ETchanged(bf)) {
    sprintf(msg, "editor -- %s --\nChanges not saved!\nDestroy anyway ?",
	    							   bf->name); 
    return(exNotify2User(bf->w, -1, -1, msg) == ANSWERYES);
  }
  return(True);
}

/* status: ******************************************************************
 * show the editing status 
 ***************************************************************************/

void status (bf)
     register BufInfo *bf;
{
  sprintf(msg, "file: %s Column %d Row %d Line %d of %d (%d) Buffer %d", 
	  bf->filename ? bf->filename : "no file", 	/* file name */
	  bf->curcol + 1, 				/* col */
	  bf->currow + 1,				/* screen row */
	  bf->curline + 1, 				/* line */
	  ETnumlines(bf), 				/* lines in buffer */
	  bf->bufPtr->bufferLines,			/* bufferlines */
	  bf->bufnum + 1);				/* buffer number */
  ESmsg(bf, msg, True);
}

/* readfile: ***************************************************************
 * read file, prompt for file name if necessary
 ***************************************************************************/

void readfile(bf)
     register BufInfo *bf;
{

  if (! clearbuf(bf, True))
    return;

  if (! EIOreadf(bf))
    *(bf->filename) = EOS;
  ETunchanged(bf);
  ESshowall(bf);
  csrChrForget(bf);
  reset_region(bf);
}

/* insertfile: ***************************************************************
 * insert file; scroll until last line of inserted file is visible on the
 * screen
 ***************************************************************************/

void insertfile(bf, filename)
     BufInfo *bf;
     char *filename;
{
  if (! *filename)
    return;

  EIOinsertf(bf, filename);

  showline(bf, bf->curline, strlen(ETtext(bf, bf->curline)));
  ESshowall(bf);

  if (bf->curline == lastLineOnScreen(bf))
    ESscroll(bf, 1, DIR_UP);
}

/* writefile: ***************************************************************
 * write file
 ***************************************************************************/

void writefile(bf)
     register BufInfo *bf;
{
  if (! *bf->filename)
    return;

  if (EIOwritef(bf))
    ETunchanged(bf);
}

/* ESmsgclear: ****************************************************************
 * clear the message area
 *****************************************************************************/

ESmsgclear(bf)
     BufInfo *bf;
{
  if (bf->msg) 
    XClearArea(XtDisplay(bf->msg), XtWindow(bf->msg), 0, 0, 0, 0, False);
}

/* search: ********************************************************************
 * search for a given string 
 *****************************************************************************/

ESsearch(bf, dir)
     BufInfo *bf;
     int dir;
{
  register char c0, c1;
  int lininc;
  int l;
  register pos, t_pos, s_pos;
  char sbuffer[20];
  int numlines = ETnumlines(bf);
  register char *t = bf->sbuff;

  if (dir != DIR_UP && dir != DIR_DOWN)
    return;

  if (! bf->sbuff[0]) {
    ESmsg(bf, "Empty search string.", True);
    return;
  }

  lininc = (dir == DIR_DOWN) ? 1 : (-1);
  
  if (dir == DIR_DOWN)
    strcpy(sbuffer, t);
  else {					/* revert search string */
    register i, j;
    
    for (i = strlen(t) - 1, j = 0; i >= 0; i--, j++)
      sbuffer[j] = t[i];
    sbuffer[strlen(t)] = EOS;
  }

  for (l = bf->curline; l < numlines && l >= 0; l += lininc) {
    s_pos = 0;					/* firstpos in search buf */
    t = ETtext(bf, l);				/* current text line */
    pos = (l == bf->curline) ? bf->curpos : 
                     (lininc > 0) ? 0 : ETlen(bf, l) - 1;

    for (t_pos = pos; ; ) { 			/* search current line */
      c1 = sbuffer[s_pos];			/* char in search buffer */

      if (c1 == EOS) { 				/* found it */
	bf->curpos = (dir == DIR_DOWN) ? t_pos : t_pos + 1;
	if (l >= lastLineOnScreen(bf) || l < bf->topline) {
	  if (! showline(bf, l, bf->curpos)) {
	    bf->curline = l;
	    bf->topline = l;
	    phys2screen(bf, bf->curline, bf->curpos, 
					&bf->currow, &bf->curcol, True);
	    ESshowall(bf);
	  }
	  return;
	} else {				/* on screen */
	  bf->curline = l;
	  phys2screen(bf, bf->curline, bf->curpos, 
		      			&bf->currow, &bf->curcol, True);
	  ESsetcur(bf);
	  return;
	}
      }

      c0 = t[t_pos];				/* char in buffer */

      if (dir == DIR_DOWN) {
	if (c0 == EOS) 				/* end of line in buffer */
	  break;				/* next line in buffer */
      } else {
	if (t_pos < 0)				/* start of line */
	  break;				/* goto line above */
      }

      if (c0 != c1) {				/* chars dont match */
	t_pos += lininc;			/* try from next char */
	s_pos = 0;				/* beg of search buff */
	continue;
      }

      t_pos += lininc;				/* advance pointers */
      s_pos++;
    }
  }

  sprintf(msg,"%s: not found!", bf->sbuff);
  ESmsg(bf, msg, True);					/* announce failure */
}

/* ESwords: *******************************************************************
 * skip over words 
 *****************************************************************************/

ESwords(bf, dir)
     BufInfo *bf;
     int dir;
{
  register char *t;
  register pos, l;
  int incr, curline;

  if (dir != DIR_LEFT && dir != DIR_RIGHT)
    return;
  incr = (dir == DIR_LEFT) ? -1 : 1;

  t = ETtext(bf, bf->curline);
  l = ETlen(bf, bf->curline);

  pos = bf->curpos;

  if (dir == DIR_LEFT && ! wrdchr(t[pos - 1]))	/* first char of word */
    pos--;

  if (wrdchr(t[pos])) {	/* inside a word, pos to 1st or behind last char */
      while (wrdchr(t[pos]))
	if ((incr < 0 && pos <= 0) || (incr > 0 && pos >= l))
	  break;
	else
	  pos += incr;
      if (dir == DIR_LEFT && pos > 0 && pos < l)
	pos++;
      bf->curpos = pos;
      curline = bf->curline;
      setcol(bf);
      while (bf->curline < curline) {	      	/* curline has been changed */
	ESscroll(bf, 1, DIR_UP);
	bf->curline = MIN(ETnumlines(bf) - 1, bf->curline + 1);
      }
      return;
    }
      
  while (! wrdchr(t[pos]))	/* first skip over non-word chars */
    if ((pos >= l && incr > 0) || (pos <= 0 && incr < 0))
      break;
    else 
      pos += incr;

  while (wrdchr(t[pos]))
    if ((pos == 0 && incr < 0) || (pos == l && incr > 0))
      break;
    else
      pos += incr;

  if (dir == DIR_LEFT && pos > 0 && pos < l)
    pos++;

  bf->curpos = pos;
  curline = bf->curline;
  setcol(bf);
  while (bf->curline < curline) {	      	/* curline has been changed */
    ESscroll(bf, 1, DIR_UP);
    bf->curline = MIN(ETnumlines(bf) - 1, bf->curline + 1);
  }
}

/* wrdchr: *******************************************************************
 * determine if a character is a "word"     type     character 
 ****************************************************************************/

Boolean wrdchr(chr)
     char chr;
{ 

  if ((chr >= 'a' && chr <= 'z') || (chr >= 'A' && chr <= 'Z') ||
      (chr >= '0' && chr <= '9') || chr == '.')
    return (True);
  else
    return (False);
}

/* exEditCursor: *************************************************************
 * return the cursor for the main editor window 
 ****************************************************************************/

Cursor exEditCursor()
{
  return(hairCursor);
}

/* exScrollbarCursor: *******************************************************
 * return the cursor for the scrollbar 
 ***************************************************************************/

Cursor exScrollbarCursor()
{
  return(scrollCursor);
}

/* createEditor: ***********************************************************
 * create an editor instance according to the information contained in the 
 * description record 
 **************************************************************************/

Editor createEditor(parent, name, descr, b_args, b_cnt)
     Widget parent;
     char *name;
     EditorDescr *descr;
     Arg b_args[];
     int b_cnt;
{
  char *XLookupoString();
  static editorNumber = 0;
  register BufInfo *bf = bufinfoptr(editorNumber);
  Widget w;
  Arg args[25];
  register cnt;
  XmFontList fl;

  if (! ptt || !ptt1) {
    XtAddActions(editor_actions, 11);
    ptt = XtParseTranslationTable(editor_translations);
    ptt1 = XtParseTranslationTable(main_translations);
  }

  _initialise(XtDisplay(parent));

  for (cnt = 0; cnt < b_cnt; cnt++) {
    args[cnt].name = b_args[cnt].name;
    args[cnt].value = b_args[cnt].value;
  }
    
  XtSetArg(args[cnt], XmNuserData, bf); cnt++;
  XtSetArg(args[cnt], XmNshadowThickness, 0); cnt++;
  w = XtCreateManagedWidget(name, xmFormWidgetClass, parent, args, cnt);
  
  XtAddCallback(w, XmNdestroyCallback, PSend, bf);

  /* retrieve the font from the just created widget */
  cnt = 0;
  XtSetArg(args[cnt], XmNtextFontList, &fl); cnt++;
  XtGetValues(w, args, cnt);
  
  if (fl) {
    bf->font = getFont(fl);
    if (! bf->font) {
      exError("createEditor: cannot get font !", NULL);
      bf->font = exEditorFont(XtDisplay(parent));
    } else if (XTextWidth(bf->font, "i", 1) != XTextWidth(bf->font, "M", 1)) {
      exError("createEditor: not a fixed font !", NULL);
      bf->font = exEditorFont(XtDisplay(parent));
    }      
  } else
    bf->font = exEditorFont(XtDisplay(parent));

  bf->w = w;
  bf->name = (descr->name) ? strCopy(descr->name) : "no name";
  bf->msg = descr->msg;
  bf->cols = descr->cols;			/* actual cols */
  bf->lines = descr->lines;			/* actual lines */
  bf->topline = 0;
  bf->currow = 0;
  bf->curcol = 0;
  bf->curpos = 0;
  bf->lmarg = 0;
  bf->filename = strCopy(descr->file);
  bf->bufnum = editorNumber;
  bf->inputFnct = (descr->inputFnct) ? descr->inputFnct : XLookupString;
  bf->flags = descr->flags;
  bf->fh = exFontHeight(bf->font);
  bf->fw = exFontWidth(bf->font);
  bf->csrFrame.x = 0;
  bf->csrFrame.y = 0;
  bf->csrFrame.w = exFontWidth(bf->font);
  bf->csrFrame.h = LINEHEIGHT(bf->font);
  bf->baseline = bf->font->max_bounds.ascent;
  bf->cursor = hairCursor;

  /* calculate width and height of editor main area based on the given font */
  bf->width = bf->cols * exFontWidth(bf->font);
  if (bf->flags & ED_SCROLLBAR)
    bf->width += (3 * exFontHeight(bf->font) / 2);

  bf->height = bf->lines * LINEHEIGHT(bf->font);
  if (bf->lines > 1)
    bf->height += FONTGAP;
  
  /* set height and width of top-level widget */
  cnt = 0;
  XtSetArg(args[cnt], XmNwidth, bf->width); cnt++;
  XtSetArg(args[cnt], XmNheight, bf->height); cnt++;
  XtSetValues(w, args, cnt);

  /* create scrollbar (if needed) and main drawing area */
  cnt = 0;
  if (descr->flags & ED_SCROLLBAR) { 		/* make scrollbar */
    XtSetArg(args[cnt], XmNwidth, 3 * exFontHeight(bf->font) / 2); cnt++;
    XtSetArg(args[cnt], XmNmaximum, bf->height); cnt++;
    XtSetArg(args[cnt], XmNminimum, 0); cnt++;
    XtSetArg(args[cnt], XmNsliderSize, bf->height); cnt++;
    XtSetArg(args[cnt], XmNvalue, 0); cnt++;
    XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNrepeatDelay, 1); cnt++;

    bf->scrollbar = 
     XtCreateManagedWidget("vScrollBar", xmScrollBarWidgetClass, 
			                               w, args, cnt);

    /* add the scrollbar callbacks */
    XtAddCallback(bf->scrollbar, XmNincrementCallback,     doScroll, bf);
    XtAddCallback(bf->scrollbar, XmNdecrementCallback,     doScroll, bf);
    XtAddCallback(bf->scrollbar, XmNdragCallback,          doScroll, bf);
    XtAddCallback(bf->scrollbar, XmNpageDecrementCallback, doScroll, bf);
    XtAddCallback(bf->scrollbar, XmNpageIncrementCallback, doScroll, bf);
    XtAddCallback(bf->scrollbar, XmNtoTopCallback,         doScroll, bf);
    XtAddCallback(bf->scrollbar, XmNtoBottomCallback,      doScroll, bf);
    
    cnt = 0;
    XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
    XtSetArg(args[cnt], XmNleftWidget, bf->scrollbar); cnt++;
    XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNheight, bf->height); cnt++;
  } else {
    XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  }

  /* create main drawing area */
  XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
  XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, 0); cnt++;
  XtSetArg(args[cnt], XmNuserData, bf); cnt++;
  
  bf->edt = XtCreateManagedWidget("edit_area", 
			  xmDrawingAreaWidgetClass, w, args, cnt);
  
  XtOverrideTranslations(bf->edt, ptt);
  XtOverrideTranslations(bf->w, ptt1);

  ETbegin(editorNumber, descr);			/* initialise text buffer */
  reset_region(bf);
  exEditSetScrollbar(bf);		     	/* set scrolbar */
  return(editorNumber++);
}

/*****************************************************************************
 *
 *		scrollbar functions 
 *
 ****************************************************************************/

/* exEditSetScrollbar: *******************************************************
 * adjust scrollbar values 
 ****************************************************************************/
void exEditSetScrollbar(bf)
     BufInfo *bf;
{
  int maxline, value, size, incr, page_incr;
  Arg args[5];
  int cnt;

  if (! bf->scrollbar || !(bf->flags & ED_MAPPED))
    return;

  maxline = ETnumlines(bf);  

  size  = MIN((int) ((float) bf->height * (float) bf->lines / (float) maxline),
	      bf->height);
  value = MIN(bf->height - size, 
	      (int) ((float) (bf->height * bf->topline) / (float) maxline));
  incr = MAX(1, (int) ((float) bf->height / (float) maxline));
  page_incr = MIN(bf->height, MIN(bf->lines, ETnumlines(bf)) * incr);

  cnt = 0;
  XtSetArg(args[cnt], XmNmaximum, bf->height); cnt++;
  XtSetArg(args[cnt], XmNpageIncrement, page_incr); cnt++;
  XtSetArg(args[cnt], XmNsliderSize, size); cnt++;
  XtSetArg(args[cnt], XmNvalue, value); cnt++;
  XtSetArg(args[cnt], XmNincrement, incr); cnt++;

  XtSetValues(bf->scrollbar, args, cnt);
}

/* scrollLines: *************************************************************
 * scroll a specified amount of lines; 
 ***************************************************************************/

scrollLines(bf, lines, dir)
     BufInfo *bf;
     int lines;
     int dir;
{
  int numlines = ETnumlines(bf);
  int amount;

  if (lines < 1)
    return;

  switch (dir) {
  case DIR_DOWN:
    if (bf->topline < lines)
      amount = bf->topline;
    else 
      amount = lines;
    break;
  case DIR_UP:
    if (regionLines(bf, bf->topline, numlines) < bf->lines)
      return;

    if (bf->topline + lines < numlines)
      amount = lines;
    else
      amount = bf->topline + lines - numlines;
    break;
  }

  if (amount < 1)			/* nothing to scroll */
    return;

  if (amount > bf->lines / 2) {		/* redraw all */
    if (dir == DIR_UP) {
      bf->topline += amount;
      bf->curline = MIN(numlines - 1, bf->curline + amount);
    }
    else {
      bf->topline -= amount;
      bf->curline = MAX(0, bf->curline - amount);
    }
    setpos(bf);
    csrChrForget(bf);
    ESshowall(bf);
  } else {				/* scroll */
    preserveCol(bf);
    ESscroll(bf, amount, dir);
    if (dir == DIR_UP)
      bf->curline = MIN(numlines - 1, bf->curline + amount);
    else
      bf->curline = MAX(0, bf->curline - amount);
    setPreservedCol(bf);
    setcol(bf);
    setpos(bf);
  } 
}

/* doScroll: ****************************************************************
 * called when the user presses a mouse button inside the scrollbar area 
 ***************************************************************************/
/*ARGSUSED*/
void doScroll(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
  int topline;
  int maxline;
  register BufInfo *bf = (BufInfo *) client_data;
  XmScrollBarCallbackStruct *sb = (XmScrollBarCallbackStruct *) call_data;

  maxline = ETnumlines(bf);

  if (sb->reason == XmCR_TO_TOP) 
    topline = 0;
  else if (sb->reason == XmCR_TO_BOTTOM)
    topline = topLine(bf, maxline - 1) /*maxline - bf->lines*/;
  else {
    topline = 
      (int)((float)sb->value * (float)regionLines(bf, 0, maxline - 1)/*maxline*/ / (float)bf->height + 0.5);
  }
  
  if (topline < 0)
    topline = 0;
  else if (topline > topLine(bf, maxline - 1) /*maxline - bf->lines*/)
    topline = topLine(bf, maxline - 1) /*maxline - bf->lines*/;

  if (topline > bf->topline)
    scrollLines(bf, topline - bf->topline, DIR_UP);
  else if (topline < bf->topline)
    scrollLines(bf, bf->topline - topline, DIR_DOWN);

  PSsetcur(bf, bf->curline, bf->curpos);
  ESshowcur(bf, True, bf->curcol, bf->currow);
}

/* exFocusIn: ****************************************************************
 * activates the editor by changing the cursor appearance to a solid cursor;
 * is called, when the mouse enters the editor window
 ***************************************************************************/
/*ARGSUSED*/
void exFocusIn(w, event, params, num_params)
     Widget w;
     XFocusChangeEvent *event;
     String params;
     int *num_params;
{
  BufInfo *bf;
  Arg args[1];

  XtSetArg(args[0], XmNuserData, &bf);
  XtGetValues(w, args, 1);

  activateEditor(bf, True);
}

/* exFocusOut: ****************************************************************
 * deactivates the editor by changing the cursor appearance to a cursor frame;
 * is called, when the mouse leaves the editor window
 ***************************************************************************/
/*ARGSUSED*/
void exFocusOut(w, event, params, num_params)
     Widget w;
     XFocusChangeEvent *event;
     String params;
     int *num_params;
{
  BufInfo *bf;
  Arg args[1];

  XtSetArg(args[0], XmNuserData, &bf);
  XtGetValues(w, args, 1);
  
  activateEditor(bf, False);
}

/* yankKillBuffer: ***********************************************************
 * copy the killBuffer to the current location 
 ****************************************************************************/

yankKillBuffer(bf)
     BufInfo *bf;
{
  if (! bf->killBuff)
    return(0);

  return(insertText(bf, bf->killBuff, &bf->curline, bf->curpos));
}

/* setKillBuffer: **********************************************************
 * set kill buffer to specified buffer
 ****************************************************************************/

setKillBuffer(bf, buf)
     BufInfo *bf;
     char *buf;
{
    if (! buf || ! *buf)
      return;

    if (bf->killBuff)
      free(bf->killBuff);
    bf->killBuff = buf;
  }


/* screenLines: *************************************************************
 * calculate how many lines on screen corresponds to physical line 
 ***************************************************************************/

screenLines(bf, line)
     BufInfo *bf;
     int line;
{
  register char *s = EMtext(bf, line, False);

  if (! strlen(s))			/* empty line */
    return(1);

  if (! (bf->flags & ED_LINEBREAK))
    return(1);

  return(ROUNDOFF(strlen(s), ESwidth(bf)));
}

/* regionLines: *************************************************************
 * calculate how many screen lines correspond to a region of text lines 
 ****************************************************************************/

regionLines(bf, line1, line2)
     BufInfo *bf;
     int line1, line2;
{
  register i, cnt;
  int numlines = ETnumlines(bf);

  for (i = line1, cnt = 0; i < line2 && i < numlines; i++)
    cnt += screenLines(bf, i);

  return(cnt);
}

/* topLine: *****************************************************************
 * calculate topline if physical line is to be displayed at the bottom
 * of the screen
 ****************************************************************************/

topLine(bf, lastLine)
     BufInfo *bf;
     int lastLine;
{
  register cnt;
  int line;

  for (line = lastLine, cnt = 0; line >= 0; line--) {
    cnt += screenLines(bf, line);
    if (cnt >= bf->lines) {	              /* off bottom */
      if (cnt > bf->lines)
	line = MIN(line + 1, lastLine);
      break;
    }
  }

  return(line);
}

/* screenLine: *************************************************************** 
 * given a physical line, calculate in which line of this line's screen 
 * representation the cursor is (starting with 0).
 *****************************************************************************/

screenLine(bf)
     BufInfo *bf;
{
  int row, row1, col;

  if (! (bf->flags & ED_LINEBREAK))
    return(0);

  phys2screen(bf, bf->curline, 0, &row, &col, False);
  phys2screen(bf, bf->curline, bf->curpos, &row1, &col, False);

  return(row1 - row);
}

/* checkScroll: ************************************************************** 
 * check if amount of lines to scroll is reasonable; return FALSE if not
 *****************************************************************************/

Boolean checkScroll(bf, lines, dir)
     BufInfo *bf;
     int lines;
     int dir;
{
  if (lines < 1)
    return(False);

  switch (dir) {
  case DIR_UP:
    return(bf->topline + lines < bf->bufPtr->numlines);
  case DIR_DOWN:
    return(bf->topline >= lines);
  default:
    return(False);
  }
}

/* lastLineOnScreen: *********************************************************
 * calculate the last physical line that is totally displayed on the screen;
 ****************************************************************************/

int lastLineOnScreen(bf)
     BufInfo *bf;
{
  register i, l;
  int numlines = ETnumlines(bf);

  for (l = 0, i = bf->topline; i < numlines && l < bf->lines; i++)
    l += screenLines(bf, i);

  return((l <= bf->lines) ? i - 1 : i - 2);
}

/* configureEditor: **********************************************************
 * called when a configure notify event has occurred, i.e. the user has 
 * changed the editor window by means of the window manager. The function 
 * calculates the appropriate number of lines and updates the editor structur.
 ****************************************************************************/
/*ARGSUSED*/
void configureEditor(w, event, params, num_params)
     Widget w;
     XConfigureEvent *event;
     String params;
     int *num_params;
{
  Arg args[2];
  int cnt;
  BufInfo *bf;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &bf); cnt++;
  XtGetValues(w, args, cnt);

  /* dimensions did not change */
  if (bf->width == event->width && bf->height == event->height)
    return;

  bf->width = event->width;
  bf->height = event->height;

  /* calculate new values */
  bf->cols   = bf->width / bf->fw;
  bf->lines  = MAX(1, bf->height / (bf->fh + 2 * FONTGAP));

  /* adjust old values accordingly */

  if (bf->currow >= bf->lines) {     	/* beyond editor frame */

    bf->currow = bf->lines - 1;

    if (!screen2phys(bf, bf->currow, 0, &bf->curline, &bf->curpos, True))
      bf->curline = bf->topline;
    bf->curcol = MIN(bf->curcol, strlen(EMtext(bf, bf->curline, False)));
    bf->curcol = MIN(bf->curcol, bf->cols - 1);
    if (!screen2phys(bf, bf->currow, bf->curcol, &bf->curline, &bf->curpos, 
		     True)) {
      bf->currow = bf->topline;
      bf->curpos = 0;
    }
    phys2screen(bf, bf->curline, bf->curpos, &bf->currow, &bf->curcol, True);

    bf->oldcol = -1;
    bf->oldrow = -1;
  }

  if (bf->flags & ED_LINEBREAK) {           /* redisplay separately */
    ESshowall(bf);  /* optimieren */
  }

  ESsetcur(bf);
}












