/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
#include <X11/Intrinsic.h>

#include <ctype.h>
#include "exEdit.h"

#include <exBasics.h>

#define lexHungry -1  
#define KBD_RESET        ctrl('g')

static char PRINTCMD[10];     /* holds editor cmd to print it in msg win */
static printCmd();

/* edit: *********************************************************************
 * the main edit loop
 ****************************************************************************/

edit(bf, edStr, len)
     BufInfo *bf;
     char *edStr;
     int len;
{  
  int lexcnt;
  static char lexchr;
  editCmd *cmdPtr;
  
  while(1) {
    normalize(bf);
    ESsetcur(bf);				/* position cursor */

    if(lex(bf, edStr, len, &lexcnt, &lexchr, &cmdPtr) == EOS)
      return(EOS);

    if (lexchr == KBD_RESET) { 			/* intercept kbd-reset */
      bf->lex_state = ISNEUTRAL;
      cmdPtr= bf->cmdTable;
      continue;
    }

    if (bf->lex_state == ISNEUTRAL) {
      evaluate(bf, cmdPtr, lexchr, lexcnt);
      bf->lex_state = ISNEUTRAL;
    }   
  }
}

/* lexget: *******************************************************************
 * get a character for lex, possibly from repeat buffer
 ****************************************************************************/
 
Boolean lexget(bf,edStr, len, chr)
     BufInfo *bf;
     char *edStr;
     int len;
     char *chr;
{  
  char tmp;
  
  gkbd(bf, edStr, len, &tmp);			/* read from keyboard */
  if (tmp == EOS)
    return(False);
  *chr = tmp;
  return(True);
}

/* lex: **********************************************************************
 * gets input and scans for its lexical value; returns a count if given.
 ****************************************************************************/

lex(bf, edStr, len, lexcnt, lexchr, cmdPtr)
     BufInfo *bf;
     char *edStr;
     int len;
     int *lexcnt;
     char *lexchr;
     editCmd **cmdPtr;
{ 
  char chr;
  
  if(! lexget(bf, edStr, len, &chr))		    	/* fetch a char */
    return(EOS);					/* stop if EOS */
  
  if ((bf->lex_state == ISNEUTRAL) && (chr == ctrl('x'))) {
    bf->lex_state = ISCONTR_X;
    return(lexHungry);
  }

  if ((bf->lex_state == ISNEUTRAL) && (chr == ESC)) {
    bf->lex_state = ISESCAPE;
    return(lexHungry);
  }

  if ((bf->lex_state == ISNUMBER) && (isNumber(chr))) { 
    bf->lex_count = bf->lex_count * 10 + chr - '0'; 	/* convert to int */
    return(lexHungry);
  }

  if ((bf->lex_state == ISESCAPE) && isNumber(chr)) {
    bf->lex_count = chr - '0';
    bf->lex_state = ISNUMBER;
    return(lexHungry);
  }

  /* return good stuff */

  *lexcnt = 1;         					/* 1 is default */
  
  switch (bf->lex_state) {
  case ISESCAPE:
    *cmdPtr = bf->escTable;
    sprintf(PRINTCMD, "ESC ");
    break;
  case ISCONTR_X:
    *cmdPtr = bf->cxTable;
    sprintf(PRINTCMD, "C-x ");
    break;
  case ISNUMBER:
    *lexcnt = bf->lex_count;
    break;
  default:
    *cmdPtr = bf->cmdTable;
    break;
  }

  *lexchr = chr;
  bf->lex_state = ISNEUTRAL;

  return(chr);
}

/* evaluate: *****************************************************************
 * evaluate the given char with respect to the current command tables: if a 
 * corresponding command is found, it is executed,; otherwise the char is 
 * inserted
 ****************************************************************************/
/*ARGSUSED*/
evaluate(bf, CMDPtr, chr, cnt)
     BufInfo *bf;
     editCmd *CMDPtr;
     char chr;
     int cnt;
{
  int state = -1;
  register editCmd *cmdPtr = CMDPtr;

  if (! cmdPtr)
    return(state);
  
  while (cmdPtr->func != (int (*)()) 0) {		/* search command */
    if (cmdPtr->key == chr) {				/* cmd exists */
      ESmsgclear(bf);
      printCmd(bf, chr);                                /* show command */
      bf->lex_state = ISNEUTRAL;			/* reset parser */
      state = (*(cmdPtr->func))(bf->bufnum /*, cnt */);	/* execute cmd */
      bf->oldFnct = cmdPtr->func;			/* remember function */
      return(state);
    }
    cmdPtr++;
  }

 if (CMDPtr == bf->cmdTable) {				/* insert char */
    if (bf->flags & ED_VIEWMODE)
      ESmsg(bf, "Read only buffer!", False);
    else {
      if (chr >= ' ' && chr <= '~')
	inschar(bf, chr);
      state = 0;
    }
  } else 
    Bell();
  
  bf->lex_state = ISNEUTRAL;				/* reset parser */
  return(state);
}					

/* printCmd: *****************************************************************
 * build a printable form of the command and put it into the message window
 *******************************************************************************/

static printCmd(bf, chr)
     BufInfo *bf;
     char chr;
{
  register l;

  if (iscntrl(chr)) {
    strcat(PRINTCMD, "C-");
    chr = chr + '@';
    chr = tolower(chr);
  }

  l = strlen(PRINTCMD);
  PRINTCMD[l] = chr;
  PRINTCMD[l + 1] = EOS;
  ESmsg(bf, PRINTCMD, False);
  PRINTCMD[0] = EOS;
}
