/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
#include <stdio.h>
#include <sys/types.h>
#include <sys/file.h>

#include <Xm/Xm.h>
#include <Xm/DrawingA.h>
#include <Xm/Form.h>

#include <X11/keysym.h>
#include <X11/Xutil.h>

#include <exBasics.h>
#include <exErrors.h>
#include <exEdit.h>
#include <exLayout.h>
#include <exStxt.h>
#include <exMDS.h>
#include <exEntry.h>
#include <exMenu.h>
#include "exFkm.hh"

#include <cursors/fkmCursor>

static char SccsID[] = "@(#)fkm.c	1.2\t11/10/92";            /* sccs keyword */

static char fullName[256];
static struct FunctionKey *FKS[FK_MAXKEYS];
static FKSCNT = 0;

static Atom exFkmReturnValue;

char *getFKSetFile();
char *getFKey();
void displayFKSet();

static void destroyFunctionKey();
static void eraseFunctionKey();
static void buttonProc();
static void showBody();
static void redrawBody();
static void unmapBody();
static void destroyBody();

static Cursor FKM_CURSOR = (Cursor) NULL;
static Widget FK_WINDOW, FK_SHELL;

static GC normal_gc, bold_gc;

/* loadFKSet: ****************************************************************
 * load a set of fuction keys from a file and display them on the screen
 *****************************************************************************/

loadFKSet(parent, getFunctionKeyValue)
     Widget parent;
     void (* getFunctionKeyValue)();
{
  char *file;
  int fd, i, cnt, a_cnt;
  register struct FunctionKey *fk;
  Arg args[4];

  static char bar_translations[] = 
    " <PropertyNotify>	: getFunctionKeyValue()";

  static XtActionsRec bar_actions[] = {
    { "getFunctionKeyValue",   NULL }
  };
  
  static XtTranslations ptt = NULL;

  if (! ptt) {
    ptt = XtParseTranslationTable(bar_translations);
  }
  
  /* register function that is called when a function key is selected */

  bar_actions[0].proc = getFunctionKeyValue;
  XtAddActions(bar_actions, 1);

  if (FKSCNT > 0) {			/* set already loaded */
    setError(ERR_EXISTS);
    return(-1);
  }
    
  file = getFKSetFile(parent);

  if (! file || ! *file) {
    setError(ERR_NULL);
    return(-1);
  }  

  if (! fileExists(file)) {
    setError(ERR_NOEXIST);
    return(-1);
  }
  
  if ((fd = open(file, O_RDONLY)) < 0) {
    setError(ERR_OPEN);
    return(-1);
  }

  read(fd, &cnt, sizeof(int));

  if (cnt > FK_MAXKEYS || cnt < 0) {
    setError(ERR_VALUE);
    close(fd);
    return(-1);
  }
  
  for (i = 0, FKSCNT = 0; i < cnt; i++) {
    fk = (struct FunctionKey *) Malloc(sizeof(struct FunctionKey));
    FKS[i] = fk;
    FKSCNT++;
    read(fd, &(fk->sym), sizeof(KeySym));
    ReadString(fd, &fk->name);
    ReadString(fd, &fk->body);

    XRebindKeysym(XtDisplay(parent), fk->sym, NULL, 0, 
			(unsigned char *) fk->body, strlen(fk->body));
  }

  close(fd);

  a_cnt = 0;
  XtSetArg(args[a_cnt], XmNwidth, 1); a_cnt++;
  XtSetArg(args[a_cnt], XmNheight, 1); a_cnt++;
  XtSetArg(args[a_cnt], XmNkeyboardFocusPolicy, XmPOINTER); a_cnt++;

  FK_SHELL  = XtCreatePopupShell("fked show shell", 
		    transientShellWidgetClass, parent, args, a_cnt);

  FK_WINDOW = XtCreateManagedWidget(NULL, xmFormWidgetClass, 
				               FK_SHELL, args, a_cnt);
  XtOverrideTranslations(FK_WINDOW, ptt);

  displayFKSet(FK_WINDOW, FKS, FKSCNT, True);
  setError(ERR_NOERR);
  return(cnt);
}

/* displayFKSet: *************************************************************
 * display the set of function keys 
 ****************************************************************************/

void displayFKSet(fkw, fks, cnt, do_kill)
     Widget fkw;
     struct FunctionKey *fks[];
     int cnt;
     Boolean do_kill;
{
  int i, width = 0;
  Dimension wt;
  Widget w;
  register struct FunctionKey *fk;
  XFontStruct *f = exBaseFont(XtDisplay(fkw));
  Arg args[10];
  XmFontList fontlist;
  static XtTranslations ptt = NULL;
  int a_cnt;

  static char button_translations[] = 
    " <Btn3Down>	: fkmHelp()     \n\
      <PropertyNotify>	: ArmAndActivate()";

  static XtActionsRec button_actions[] = {
    { "fkmHelp", showBody},
  };

  if (! ptt) {
    XtAddActions(button_actions, 1);
    ptt = XtParseTranslationTable(button_translations);
  }

  fontlist = XmFontListCreate(f, XmSTRING_DEFAULT_CHARSET);

  for (i = 0; i < cnt; i++) {
    fk = fks[i];

    a_cnt = 0;
    XtSetArg(args[a_cnt], XmNtopAttachment, XmATTACH_FORM); a_cnt++;
    XtSetArg(args[a_cnt], XmNbottomAttachment, XmATTACH_FORM); a_cnt++;
    if (i == 0) { 			/* left most item */
      XtSetArg(args[a_cnt], XmNleftAttachment, XmATTACH_FORM); a_cnt++;
    } else {
      XtSetArg(args[a_cnt], XmNleftAttachment, XmATTACH_WIDGET); a_cnt++;
      XtSetArg(args[a_cnt], XmNleftWidget, fks[i - 1]->keywin); a_cnt++;
    }
    XtSetArg(args[a_cnt], XmNfontList, fontlist); a_cnt++;
    XtSetArg(args[a_cnt], XmNuserData, fk); a_cnt++;
    XtSetArg(args[a_cnt], XmNlabelString, 
	     XmStringCreate(fk->name, XmSTRING_DEFAULT_CHARSET)); a_cnt++;

    w = XtCreateManagedWidget("fkw button", xmPushButtonWidgetClass, 
			                              fkw, args, a_cnt);

    XtAddCallback(w, XmNdestroyCallback, 
		  (do_kill) ? destroyFunctionKey : eraseFunctionKey, fk);
    
    XtAddCallback(w, XmNactivateCallback, buttonProc, fk);
    XtOverrideTranslations(w, ptt);

    a_cnt = 0;
    XtSetArg(args[a_cnt], XmNwidth, &wt); a_cnt++;
    XtGetValues(w, args, a_cnt);
    
    width += wt;         

    fk->keywin = w;
    fk->fkBar = fkw;

    fk->bodyWin = (Widget) NULL;
  }

  XtConfigureWidget(XtParent(fkw), 
		    (XWidthOfScreen(XtScreen(w)) - width) / 2, 
		    XHeightOfScreen(XtScreen(w)) - (KEY_HEIGHT + 20),
		    width,
		    KEY_HEIGHT,
		    2);

  XtPopup(XtParent(fkw), XtGrabNone);
  XDefineCursor(XtDisplay(fkw), XtWindow(fkw), FKM_CURSOR);
}

/* closeFKSet: ***************************************************************
 * remove the displayed function keys from the screen and remove the 
 * association between a physical key and the defined string
 *****************************************************************************/

closeFKSet()
{
  if (FKSCNT > 0) {
    XtDestroyWidget(XtParent(FKS[0]->fkBar));	
    setError(ERR_NOERR);
    FKSCNT = 0;
    return(0);
  } else {
    setError(ERR_NOEXIST);
    return(-1);
  }
}

/* getFKSetFile: **************************************************************
 * get the name of a file that contains a function key set
 *****************************************************************************/
char *getFKSetFile(w)
     Widget w;
{
  char *getcwd();
  char *file;
  int cnt, i;
  char *getenv();
  char *fkmdir, *ptr;
  char *fkmTable[MDS_MAXITEM];
  MHmenu *mptr;
  MHitem *iptr;
  int x,y;
  exMenu  *fkmList;

  if (! (fkmdir = getenv("FKPATH"))) {		/* let the user enter a file */
    if (ptr = getenv("HOME"))
      file = strCopy(ptr);
    else 
      file = NULL;
    
    return(exGetFilename(w, -1, -1, "Open file: ", file, 40));
  } else {					/* show a list of description 
						   sets in the fkm directory */
    getMouse_xy(XtDisplay(w), &x, &y);
    if ((cnt = loadDirectory(fkmdir, fkmTable, MDS_MAXITEM)) < 0){
      if (ptr = getenv("HOME"))
	file = strCopy(ptr);
      else 
	file = NULL;

      return(exGetFilename(w, -1, -1, "Open key set: ", file, 40));
    }
    
    setWorkApplication();

    if (! (mptr = createMenu())) {
      setError(ERR_CREATE);
      return(NULL);
    }

    insertMenu(mptr);

    setSelCnt(mptr, 1);
    setSelType(mptr, MDS_CNT_EQ);
    initMenu(mptr, MDS_L_POPUP, "key sets");
    setMenuPos(mptr, x, y, MDS_POS_ABS);
    
    mptr->m_hlp->hlpmsg = strCopy("Please select one of the key sets.");
    mptr->m_key = EOS;
  
    for (i = 0; i < cnt; i++) {				/* make the items */
      if (makeFullName(fkmdir, fkmTable[i], fullName, sizeof(fullName)) < 0)
	continue;
      iptr = newItem(mptr, fkmTable[i], EOS);		/* create item */
      iptr->i_return->retPattern = strCopy(fullName);
      free(fkmTable[i]);
    }
    
    fkmList = exMakeMenu(NULL, NULL, w, mptr);

    if (fkmList == (exMenu *) 0)
      return(NULL);
    
    exCallMenu(fkmList);

    file = exGetMenuReturnValue(fkmList);
  
    freeMenu(fkmList);
    deleteWorkMenu(mptr);
  }

  return(file);
}

/* static destroyFunctionKey: *************************************************
 * free a function key's data structures
 *****************************************************************************/
/*ARGSUSED*/
static void destroyFunctionKey(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
  struct FunctionKey *fk = (struct FunctionKey *) client_data;

  if (fk->bodyWin) {
    XtDestroyWidget(XtParent(fk->bodyWin));
    fk->bodyWin = (Widget) NULL; 
  }

  fk->fkBar = (Widget) NULL;

  XRebindKeysym(XtDisplay(w), fk->sym, NULL, 0, (unsigned char *) "", 0);

  free(fk->name);
  free(fk->body);
  free(fk);
}

/* static eraseFunctionKey: ***************************************************
 * erase a function key from the screen
 *****************************************************************************/
/*ARGSUSED*/
static void eraseFunctionKey(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
  struct FunctionKey *fk = (struct FunctionKey *) client_data;

  XRebindKeysym(XtDisplay(w), fk->sym, NULL, 0, (unsigned char *) "", 0);

  if (fk->bodyWin) {
    XtDestroyWidget(XtParent(fk->bodyWin));
    fk->bodyWin = (Widget) NULL; 
  }

  fk->fkBar = (Widget) NULL;
}

/* buttonProc: ****************************************************************
 * is called when button is pressed in a function key window; left button means
 * execute function key, right button means show body (see showBody)
 *****************************************************************************/
/*ARGSUSED*/
static void buttonProc(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
  struct FunctionKey *fk = (struct FunctionKey *) client_data;

  XChangeProperty(XtDisplay(w), XtWindow(fk->fkBar),
		    exFkmReturnValue, XA_STRING, 8, PropModeReplace,
		    (unsigned char *) fk->body, strlen(fk->body));
}
     
/* showBody: ******************************************************************
 * display the body text of a function key in a window; create the window if 
 * necessary, otherwise map an already created body window; when the window is 
 * mapped, the function redrawBody is called automatically to draw the text
 * into the window; 
 *****************************************************************************/
/*ARGSUSED*/
static void showBody(wd, event, params, num_params)
     Widget wd;
     XLeaveWindowEvent *event;
     String *params;
     Cardinal *num_params;
{  
  Arg args[10];
  struct FunctionKey *fk;
  Widget win, swin;
  XFontStruct *f;
  int x, y, w, h;
  int dw, dh, cnt;
  static void unmapBody();

  static char body_translations[] = 
    " <BtnDown>	: unmapBody() 	\n \
      <Expose> 	: redrawBody()";

  static XtActionsRec body_actions[] = {
    { "unmapBody",   unmapBody },
    { "redrawBody",  redrawBody }
  };
  
  static XtTranslations ptt = NULL;

  if (! ptt) {
    XtAddActions(body_actions, 2);
    ptt = XtParseTranslationTable(body_translations);
  }

  XtSetArg(args[0], XmNuserData, &fk);
  XtGetValues(wd, args, 1);

  if (fk->flags & KEY_BODYDISPLAYED)		/* already displayed */
    return;

  f = exBaseFont(XtDisplay(wd));
  fk->bodyLines = exFormat(&fk->body, 1, 40, 0, &fk->fbodyText, False);
  
  dw = XWidthOfScreen(XtScreen(wd));
  dh = XHeightOfScreen(XtScreen(wd));

  w = 40 * exFontWidth(f) + 10;
  h = fk->bodyLines * exFontHeight(f) + exFontHeight(exBoldFont(XtDisplay(wd))) + 15;

  x = event->x_root - event->x;
  y = event->y_root - event->y - (h + 10);

  if (x < 0)
    x = 0;
  else if (x > dw - w)
    x = dw - w + 20;

  if (y < 0)
    y = 0;
  else if (y > dh - h)
    y = dh - h + 20;

  if (fk->bodyWin) {				/* already created, ... */
    XtMoveWidget(XtParent(fk->bodyWin), x, y);	/* ... so move and ... */
    XtPopup(XtParent(fk->bodyWin), XtGrabNone);		/* ... map it */
    fk->flags |= KEY_BODYDISPLAYED;
  } else {					/* create body window first */
    cnt = 0;
    XtSetArg(args[cnt], XmNx, x); cnt++;
    XtSetArg(args[cnt], XmNy, y); cnt++;
    XtSetArg(args[cnt], XmNwidth, w); cnt++;
    XtSetArg(args[cnt], XmNheight, h); cnt++;
    XtSetArg(args[cnt], XmNuserData, fk); cnt++;
    XtSetArg(args[cnt], XmNborderWidth, 4); cnt++;
    XtSetArg(args[cnt], XmNkeyboardFocusPolicy, XmPOINTER); cnt++;
    swin = XtCreatePopupShell("fkm body shell", transientShellWidgetClass, 
			                                     wd, args, cnt);
    
    cnt = 0;
    XtSetArg(args[cnt], XmNwidth, w); cnt++;
    XtSetArg(args[cnt], XmNheight, h); cnt++;
    XtSetArg(args[cnt], XmNuserData, fk); cnt++;

    win = XtCreateManagedWidget("fkm body", xmDrawingAreaWidgetClass, 
				swin, args, cnt);
    XtOverrideTranslations(win, ptt);
    XtAddCallback(win, XmNdestroyCallback, destroyBody, fk);
    fk->bodyWin = win;
    fk->flags |= KEY_BODYDISPLAYED;
    XtPopup(swin, XtGrabNone);
  }
}

/* redrawBody: ****************************************************************
 * is called when an exposure event occures within a window that displays the 
 * text associated with a function key; 
 *****************************************************************************/
/*ARGSUSED*/
static void redrawBody(wd, event, params, num_params)
     Widget wd;
     XCrossingEvent *event;
     String params;
     int *num_params;
{
  struct FunctionKey *fk;
  Arg args[1];
  int x, y, w, i, kw, nw;
  XFontStruct *f = exBaseFont(XtDisplay(wd));
  char k[10];
  Display *d = XtDisplay(wd);
  Window win;

  XtSetArg(args[0], XmNuserData, &fk);
  XtGetValues(wd, args, 1);

  win = XtWindow(fk->bodyWin);

  sprintf(k, "  (%s)", getFKey(fk));
  kw = exTextWidth(exBoldFont(XtDisplay(wd)), k);
  nw = exTextWidth(exBoldFont(XtDisplay(wd)), fk->name);
  w = 40 * exFontWidth(f) + 10;

  x = MAX(0, (w - (nw + kw)) / 2);
  y = exBoldFont(d)->max_bounds.ascent;

  XDrawImageString(d, win, bold_gc, x, y, fk->name, strlen(fk->name));
  XDrawImageString(d, win, bold_gc, x + nw, y, k, strlen(k));

  y = y + exFontHeight(exBoldFont(d)) + 5;

  for (i = 0; i < fk->bodyLines; i++) {
    XDrawImageString(d, win, normal_gc, 5, y, 
		     fk->fbodyText[i], strlen(fk->fbodyText[i]));

    y += exFontHeight(f);
  }
}

/* unmapBody: *****************************************************************
 * is called to unmap the body window of a function key;
 ****************************************************************************/
/*ARGSUSED*/
static void unmapBody(w, event, params, num_params)
     Widget w;
     XCrossingEvent *event;
     String params;
     int *num_params;
{
  struct FunctionKey *fk;
  Arg args[1];
  
  XtSetArg(args[0], XmNuserData, &fk);
  XtGetValues(w, args, 1);

  if (! fk->flags & KEY_BODYDISPLAYED)
    return;

  XtPopdown(XtParent(fk->bodyWin));
  fk->flags &= ~KEY_BODYDISPLAYED;
}
			   
/* destroyBody: **************************************************************
 * is called when a function key is destroyed; it frees the structures of a 
 * function keys' body
 *****************************************************************************/
/*ARGSUSED*/
static void destroyBody(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
  struct FunctionKey *fk = (struct FunctionKey *) client_data;
  register i;

  if (! fk->flags & KEY_BODYDISPLAYED)
    return;

  for (i = 0; i < fk->bodyLines; i++)
    free(fk->fbodyText[i]);
  free(fk->fbodyText);

  fk->flags &= ~KEY_BODYDISPLAYED;
}

/* getFKey: *******************************************************************
 * return the name of the physical key that is assigned to the current key
 *****************************************************************************/

char *getFKey(fk)
     struct FunctionKey *fk;
{
  KeySym sym;

  static char *keys[] = 
                 {"??", "F1", "F2", "F3", "F4", "F5", "F6", "F7", "F8", "F9"};
  if (! fk)
    return(NULL);

  sym = fk->sym;

  if (sym == (KeySym) XK_F1)
    return(keys[1]);
  else if (sym == (KeySym) XK_F2)
    return(keys[2]);
  else if (sym == (KeySym) XK_F3)
    return(keys[3]);
  else if (sym == (KeySym) XK_F4)
    return(keys[4]);
  else if (sym == (KeySym) XK_F5)
    return(keys[5]);
  else if (sym == (KeySym) XK_F6)
    return(keys[6]);
  else if (sym == (KeySym) XK_F7)
    return(keys[7]);
  else if (sym == (KeySym) XK_F8)
    return(keys[8]);
  else if (sym == (KeySym) XK_F9)
    return(keys[9]);
  else 
    return(keys[0]);
}

/* exInitFKM: ****************************************************************
 * create and initialize internally used variables 
 *****************************************************************************/

void exInitFKM(d)
     Display *d;
{
  Window w = XDefaultRootWindow(d);
  Pixmap pm;
  XColor fg, bg;

  normal_gc = XCreateGC(d, w, 0, 0);
  XSetBackground(d, normal_gc, exWhitePixel(d));
  XSetForeground(d, normal_gc, exBlackPixel(d));
  XSetFont(d, normal_gc, exBaseFont(d)->fid);

  bold_gc = XCreateGC(d, w, 0, 0);
  XSetBackground(d, bold_gc, exWhitePixel(d));
  XSetForeground(d, bold_gc, exBlackPixel(d));
  XSetFont(d, bold_gc, exBoldFont(d)->fid);

  exFkmReturnValue = XInternAtom(d, EX_FKM_VALUE_ATOM_NAME, False);
  
  pm = XCreateBitmapFromData (d, w, 
	       fkmCursor_bits, fkmCursor_width, fkmCursor_height);
  FKM_CURSOR = XCreatePixmapCursor(d, pm, pm, &fg, &bg, 1, 1);
}

/* getFKValue: ***************************************************************
 * given a Property-Notify event, this function tries to get the corresponding
 * string value from the given widget; The function is used, when the user 
 * selects a function key with the mouse and wants to retrieve the associated 
 * function key value.
 *****************************************************************************/

char *getFKValue(event)
     XPropertyEvent *event;
{
  Atom ret_type;
  int ret_format;
  unsigned long ret_items;
  unsigned long rem_bytes;
  unsigned char *value;
  char *ret;

  if (event->state != PropertyNewValue) {
    setError(ERR_TYPE);
    return(NULL);
  }

  if (event->atom != exFkmReturnValue) {
    setError(ERR_TYPE);
    return(NULL);
  }

  XGetWindowProperty(event->display, event->window, event->atom, 0L, 8192L, 
		     True, XA_STRING, &ret_type, &ret_format, 
		     &ret_items, &rem_bytes, &value);
  if (value) {
    ret = strCopy(value);
    XFree((char *) value);
  } else
    ret = NULL;

  setError(ERR_NOERR);
  return(ret);
}
