/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
/****************************************************************************
 * this files contains the functions provided to access various components of 
 * the MDS.
 ****************************************************************************/

#include <X11/Intrinsic.h>
#include <exBasics.h>
#include <exErrors.h>
#include <exLayout.h>
#include "exMDS.h"
#include "exMenu.h"

char *cuserid();

static curAppl = -1;	/* the current application (index into MDS - array) */
   
MHfunc MH_FUNCTIONS[MDS_MAXFUNC];

MHfunc MDS_functions[] = { /* holds pointers to some predefined functions */
       {(int (*)()) 0, ""},
       {(int (*)()) dsplInWnd, "dsplInWnd"},
       {(int (*)()) reqViaWnd, "reqViaWnd"},
       {(int (*)()) reqViaMenu,"reqViaMenu"},
       {(int (*)()) reqViaForm,"reqViaForm"},
       {(int (*)()) 0, ""}
};
     
/* MDS is the basic structure; it holds for each application an identifier 
   and a reference to the corresponding application node 
*/
     
ApplicationNode *MDS[MDS_MAXAPPL];

/* initMDSarray: *************************************************************
 * initialize the global MDS-array holding the application nodes
 *****************************************************************************/

initMDSarray()
{
  register i;
  for (i = 0; i < MDS_MAXAPPL; i++)
    MDS[i] = (ApplicationNode *) 0;
}

/* currentApplication: ********************************************************
 * return application node with index 'curAppl' in MDS-array
 *****************************************************************************/

ApplicationNode *currentApplication()
{
  if (curAppl < 0 || curAppl > MDS_MAXAPPL)
    return((ApplicationNode *) 0);
  
  return(MDS[curAppl]);
}

/* setCurrentApplication: ****************************************************
 * search applnode in MDS-array and set static variable 'curAppl' to the 
 * corresponding index 
 *****************************************************************************/

setCurrentApplication(aptr)
     ApplicationNode *aptr;
{
  register i;
  
  CHECK(aptr);
  
  for (i = 0; i < MDS_MAXAPPL; i++)
    if (aptr == MDS[i]) {
      curAppl = i;
      return(setError(ERR_NOERR));
    }
  
  return(setError(ERR_NOEXIST));
}

/****************************************************************************
 * 		initialise the corresponding globals 
 ****************************************************************************/

Boolean resetMDS()
{
  register i;
  for (i = 0; i < MDS_MAXAPPL; i++)
    clearApplNode(i);
  
  return(True);
}

/****************************************************************************
 * 		create, insert, delete application nodes 
 ****************************************************************************/

/* createApplNode: **********************************************************
 * provide storage for application node structure and set members to NULL
 ****************************************************************************/

ApplicationNode *createApplNode()
{
  ApplicationNode *aNptr;
  register i;
  
  aNptr = (ApplicationNode *) Malloc(sizeof(ApplicationNode));
  
  for (i = 0; i < MDS_MAXFUNC; i++) {
    aNptr->ApplFunc[i].func = (int (*)()) 0;
    aNptr->ApplFunc[i].funcName = (char *) 0;
  }
  
  aNptr->menuCnt = 0;
  aNptr->a_access = MDS_CHNGACC;
  
  aNptr->m_table = (MHmenu **) Calloc(MDS_MAXMENU,sizeof(MHmenu *));
  for (i = 0; i < MDS_MAXMENU; i++) 
    aNptr->m_table[i] = (MHmenu *) 0;
  
  aNptr->file = (char *) 0;
  
  aNptr->applicationName = (char *) 0;
  return(aNptr);
}

/* initApplNode: *************************************************************
 * initialize application node structure with 'reasonable' default values
 *****************************************************************************/

initApplNode(aptr, name)
     ApplicationNode *aptr;
     char *name;
{
  /* stay that up to now no modified menus have to be saved */
  aptr->_changed = False;
  
  /* initialise the application structure */
  aptr->menuCnt = 0;
  aptr->a_access = MDS_CHNGACC;		/* no access permitted */
  
  /* generate a file name out of the applicationname */
  aptr->file = Malloc(strlen(name) + 1 + sizeof(MDS_FILESUFFIX));
  strcpy(aptr->file, name);
  strcat(aptr->file, MDS_FILESUFFIX);
  
  aptr->applicationName = strCopy(name);
}

/* insertApplNode: ***********************************************************
 * insert the given application node into the MDS-array into the first free 
 * slot;
 ****************************************************************************/

int insertApplNode(ptr)
     ApplicationNode *ptr;
{
  register i;
  int oldAppl = curAppl;
  
  initMDS();					/* initialise MDS array */
  for (i = 0; i < MDS_MAXAPPL; i++)
    if (MDS[i] == (ApplicationNode *) 0) {
      MDS[i] = ptr;
      curAppl = i;
      return(ERR_NOERR);
    }
  
  curAppl = oldAppl;
  deleteApplNode(ptr);
  return(ERR_MAX);
}

/* newApplicationNode: ********************************************************
 * create, initialize and insert application node into MDS-array
 *****************************************************************************/

int newApplicationNode(name)
     char *name;
{
  ApplicationNode *aptr = createApplNode();
  initApplNode(aptr,name);
  return(insertApplNode(aptr));
}

/* removeApplNode: ***********************************************************
 * search for given applnode in MDS-array and delete it 
 *****************************************************************************/

removeApplNode(aptr)
     ApplicationNode *aptr;
{
  register i;
  
  if (!aptr) 
    return;
  
  for (i = 0; i < MDS_MAXAPPL; i++)
    if (MDS[i] == aptr) {
      clearApplNode(i);
      return;
    }
}

/* clearApplNode: *************************************************************
 * delete application node stored in MDS-array under given index and clear the
 * corresponding MDS-array element
 *****************************************************************************/

clearApplNode(indx)
     int indx;
{
  ApplicationNode *aptr = MDS[indx];
  
  if (aptr == (ApplicationNode *) 0)
    return;
  
  deleteApplNode(aptr);
  MDS[indx] = (ApplicationNode *) 0;
}

/* deleteApplNode: ***********************************************************
 * delete substructures and free storage of application node structure 
 * (allowed to owner or if change access is set)
 *****************************************************************************/

deleteApplNode(aptr)
     ApplicationNode *aptr;
{
  register i;
  
  CHECK(aptr);
  
  if (Access(aptr,MDS_CHNGACC) == False)
    return(setError(ERR_ACCESS));
  
  for (i = 0; i < MDS_MAXMENU; i++)
    deleteMenu(aptr,aptr->m_table[i]);
  
  for (i = 0; i < MDS_MAXFUNC; i++)
    free(aptr->ApplFunc[i].funcName);
  
   free(aptr->applicationName);
  free(aptr->m_table);
  free(aptr->file);
  
  free(aptr);
  BYE;
}

/* getCurAppl ****************************************************************
 * provides value of 'curAppl' (the index of the current application in the 
 * MDS-array 
 ****************************************************************************/

getCurAppl()
{
  if (curAppl < 0 || curAppl > MDS_MAXAPPL)
    return(-1);
  
  return(curAppl);
}

/* getMDSindex ****************************************************************
 * gets index of 'aptr' in MDS-array
 *****************************************************************************/

getMDSindex(aptr)
     ApplicationNode *aptr;
{
  register i;
  
  CHECK(aptr);
  
  for (i = 0; i < MDS_MAXAPPL; i++)
    if (aptr == MDS[i]) {
      return(i);
    }
  
  return(setError(ERR_NOEXIST));
}

/* getApplByName: ************************************************************
 * search for application node with a given name in MDS-array
 ****************************************************************************/

ApplicationNode *getApplByName(name)
     char *name;
{
  register i;
  register ApplicationNode *aptr;
  
  if (name == (char *) 0)
    return ((ApplicationNode *) 0);
  
  for (i = 0; i < MDS_MAXAPPL; i++) 
    if ((aptr = MDS[i]) != (ApplicationNode *) 0)
      if (aptr->applicationName && strcmp(aptr->applicationName, name) == 0)
  	return (MDS[i]);
  
  return((ApplicationNode *) 0);
}

/*****************************************************************************
 *			create, insert, delete menus 
 *****************************************************************************/

/* createMenu: ***************************************************************
 * provide storage for menu structure and set members to NULL
 *****************************************************************************/

MHmenu *createMenu()
{
  register MHmenu *mptr;
  register i;
  
  mptr = (MHmenu *) Malloc(sizeof(MHmenu));
  
  mptr->m_repr = (MHmRepr *) Malloc(sizeof(MHmRepr));
  mptr->m_repr->reprType = MDS_RTYP_STRING;
  mptr->m_repr->name = NULL;
  
  mptr->m_type = MDS_M_NULL;
  mptr->m_flags = MDS_CHNGACC;
  mptr->itemCnt = 0;
  mptr->i_selCnt = 0;			/* select only one item per time */
  mptr->i_selType = MDS_CNT_NULL; 	/* select exactly one item */
  mptr->m_pos.x = mptr->m_pos.y = -1;	/* no position defined up to now */
  mptr->m_pos.posType = MDS_POS_NULL;	/* pos. is absolute position */
  
  mptr->i_table = (MHitem **) Calloc(MDS_MAXITEM, sizeof(MHitem *));
  for (i = 0; i < MDS_MAXITEM; i++)
    mptr->i_table[i] = (MHitem *) 0;
  
  mptr->m_hlp = (MHhlp *) Malloc(sizeof(MHhlp));
  
  /* initialise the help structure */
  mptr->m_hlp->hlpmsg = (char *) 0;
  mptr->m_hlp->hlpFnct = (int (*) ()) 0;
  
  return((MHmenu *) mptr);
}

/* initMenu: *****************************************************************
 * initialize menu structure with 'reasonable' default values
 *****************************************************************************/

initMenu(mptr, type, name)
     MHmenu *mptr;
     MDS_MENU_TYPE type;
     char *name;
{
  mptr->m_type = type;
  mptr->m_flags = MDS_CHNGACC;
  mptr->itemCnt = 0;
  mptr->maxItem = MDS_MAXITEM;	/* max. number of items to be displayed
				   per menu (if more items are defined, the 
				   menu is scrollable */
  mptr->i_selCnt = 1;		/* select only one item per time */
  mptr->i_selType = MDS_CNT_EQ; /* select exactly one item */
  
  AM_ONES(&mptr->activityMask);
  
  mptr->m_pos.x = mptr->m_pos.y = 0;	/* no position defined up to now */
  mptr->m_pos.posType = MDS_POS_ABS;    /* absolute position */
  
  /* initialise the help structure */
  mptr->m_hlp->hlpmsg = MDS_M_HELP;
  mptr->m_hlp->hlpFnct = dsplInWnd;
  
  /* insert the name as the represenation of the menu */
  
  mptr->m_repr->reprType = MDS_RTYP_STRING;
  mptr->m_repr->name = strCopy(name);
  
  /* initialise the list of keywords with the keyword supplied to allow the 
   * menu to be activated by the user 
   */
  mptr->m_key = EOS;
}

/* newMenu: ******************************************************************
 * create, initialize and insert menu into the current application node; return
 * menu pointer
 ****************************************************************************/

MHmenu *newMenu(name)
     char *name;
{
  MHmenu *mptr;
  mptr = createMenu();
  initMenu(mptr, MDS_M_POPUP, name);
  insertMenu(mptr);
  
  return(mptr);
}


/* insertMenu: **************************************************************
 * insert the menu into the current application node
 ****************************************************************************/

insertMenu(m)
     MHmenu *m;
{
  register MHmenu **mptr;
  register i;
  
  if (curAppl < 0)
    return(setError(ERR_NOEXIST));
  
  if (Access(MDS[curAppl],MDS_CHNGACC) == False)
    return(setError(ERR_ACCESS));
  
  if (MDS[curAppl]->menuCnt > MDS_MAXMENU) 
    return(setError(ERR_MAX));
  
  mptr = MDS[curAppl]->m_table;	   
  
  for (i = 0; i < MDS_MAXMENU; i++)
    if (mptr[i] == (MHmenu *)0) {
      mptr[i] = m;
      MDS[curAppl]->menuCnt++;
      MDS[curAppl]->_changed = True;
      return(OK);
    }
  return(ERR_MAX);
}

/* clearMenu: ***************************************************************
 * delete substructures and free storage of menu structure
 ****************************************************************************/

clearMenu(mptr)
     MHmenu *mptr;
{
  register i;
  
  if (mptr->m_repr != (MHmRepr *) 0) {
    free(mptr->m_repr->name);
    free(mptr->m_repr);
  }
  
  for (i = 0; i < MDS_MAXITEM; i++)
    clearItem(mptr->i_table[i]);
  free(mptr->i_table);
  
  free(mptr->m_hlp->hlpmsg);
  free(mptr->m_hlp);
  
  free(mptr);
}

/* checkMenuConsistency ******************************************************
 * This functioon does some pragmatic consistency checking on the given menu 
 * description. Checked are:
 * - whether more items are defined than could be displayed in the menu
 *   in this case the type of the menu should be LIST.
 * - if the mptr denotes a list no keys or requests may be defined.
 * 
 ****************************************************************************/

Boolean checkMenuConsistency(mptr)
     MHmenu *mptr;
{
  Boolean isList;
  register i;

  if (! mptr) {
    setError(ERR_NULL);
    return(False);
  }
    
  
  if (mptr->m_type == MDS_L_POPUP || mptr->m_type == MDS_L_PULLD)
    isList = True;
  else 
    isList = False;
  
  if (mptr->itemCnt > mptr->maxItem) {
    if (isList == False)
      return(False);
  }
  
  if (isList == True) {
    for (i = 0; i < mptr->itemCnt; i++) {
      if (mptr->i_table[i]->i_return->reqCnt > 0)
	return(False);
      if (mptr->i_table[i]->i_repr->key != EOS)
	return(False);
    }
  }
  
  return(True);
}


/*****************************************************************************
 *			create, insert, delete items 
 ****************************************************************************/

/* createItem: ***************************************************************
 * provide storage for item structure and set members to NULL
 ****************************************************************************/

MHitem *createItem()
{
  register MHitem *iptr;
  register i;
  
  iptr = (MHitem *) Malloc(sizeof(MHitem));
  
  iptr->flag = MH_NODEFAULT;
  
  iptr->i_hlp = (MHhlp *) Malloc(sizeof(MHhlp));
  iptr->i_return = (MHiRet *) Malloc(sizeof(MHiRet));
  
  /* define pos as not defined */
  iptr->i_Pos.x = iptr->i_Pos.y = -1;
  iptr->i_Pos.posType = MDS_POS_NULL;
  
  /* initialise the request pointers to nil (no requests specified ) */
  for (i = 0; i < MDS_MAXREQ; i++)
    iptr->i_requests[i] = (MHiReq *) 0;
  
  /* define the item representations */
  iptr->i_repr  = new_iRepr("",EOS);
  
  /* initialise the help structure */
  iptr->i_hlp->hlpmsg = (char *) 0;
  iptr->i_hlp->hlpFnct = (int (*)()) 0;
  
  /* initialise the return record */
  iptr->i_return = newRetPattern((char *) 0);
  
  return(iptr);
}

/* initItem: *****************************************************************
 * initialize item structure with 'reasonable' default values
 ****************************************************************************/

initItem(iptr, name, key)
     MHitem *iptr;
     char *name;
     char key;
{
  /* define pos as not defined */
  iptr->i_Pos.x = iptr->i_Pos.y = -1;
  iptr->i_Pos.posType = MDS_POS_ABS;
  
  /* set all items selectable */
  AM_ONES(&iptr->activityMask);
  
  /* define the item representations */
  iptr->i_repr  = new_iRepr(name, key);
  
  /* initialise the help structure */
  iptr->i_hlp->hlpmsg = MDS_I_HELP;
  iptr->i_hlp->hlpFnct = dsplInWnd;
}

/* newItem: *****************************************************************
 * create, initialize and insert item; return item pointer
 ****************************************************************************/

MHitem *newItem(mptr,name,key)
     MHmenu *mptr;
     char *name;
     char key;
{
  MHitem *iptr = createItem();
  initItem(iptr, name, key);
  insertItem(mptr, iptr, MDS_MAXITEM);
  return(iptr);
}

/* clearItem: *****************************************************************
 * delete substructures and free storage of item structure
 *****************************************************************************/

clearItem(item)
     MHitem *item;
{
  register i;
  
  CHECK(item);
  
  free(item->i_repr->name);
  free(item->i_repr);
  
  free(item->i_hlp->hlpmsg);
  free(item->i_hlp);
  
  for (i = 0; i < MDS_MAXREQ; i++) 
    deleteRequest(item,i);
  
  free(item->i_return->retPattern);
  free(item->i_return);
  
  free(item);
  
  BYE;
}

/* insertItem: ***************************************************************
 * insert the item into the slot determined by where; care is taken that
 * no gaps occur (i.e. the highest slot number is the slot after the 
 * last item;
 *****************************************************************************/

insertItem(m,item,where)
     MHmenu *m;
     MHitem *item;
     int where;
{
  register i,j,k;
  int slot = MIN(where, MDS_MAXITEM);
  
  CHECK(m && item);
  
  if (curAppl < 0)
    return(setError(ERR_NOEXIST));
  
  if (Access(MDS[curAppl],MDS_CHNGACC) == False)
    return(ERR_ACCESS);
  
  if (m->m_type == MDS_L_POPUP || m->m_type == MDS_L_PULLD) {
    if (item->i_repr->key != EOS) {
      setItemKey(item, EOS);
      setError(ERR_VALUE);
    }
    
    if (item->i_return->reqCnt != 0) {
      setError(ERR_SIZE);
      for (i = 0; i < item->i_return->reqCnt; i++)
	deleteRequest(item, i);
    }
  }
  
  for (i = 0; i < slot; i++) 		
    if (! m->i_table[i]) {   			/* insert in 1st free slot */
      m->i_table[i] = item;
      m->itemCnt++;
      AM_SET(&m->activityMask, i);
      for (j = 0; j < MDS_MAXITEM; j++) /* same for all items */
	if (m->i_table[j])
	  AM_SET(&m->i_table[j]->activityMask, i);
      MDS[curAppl]->_changed = True;
      BYE;
    }
  
  if (! m->i_table[MDS_MAXITEM - 1]) {		/* room for one add. item */
    for (j = MDS_MAXITEM - 1; j >= slot; j--) {
      m->i_table[j] = m->i_table[j - 1];
      if (m->i_table[j]) {
	for (k = MDS_MAXITEM - 1; k >= slot; k--)
	  if (AM_ISSET(&m->i_table[j]->activityMask, k - 1))
	    AM_SET(&m->i_table[j]->activityMask, k);
	  else
	    AM_CLR(&m->i_table[j]->activityMask, k);
	    
	AM_SET(&m->i_table[j]->activityMask, slot - 1);
      }
      if (AM_ISSET(&m->activityMask, j -1))
	AM_SET(&m->activityMask, j);
      else
	AM_CLR(&m->activityMask, j);
    }

    m->i_table[slot - 1] = item;
    AM_SET(&m->activityMask, slot - 1);
    m->itemCnt++;
    MDS[curAppl]->_changed = True;
    BYE;
  }
  
  return(ERR_MAX);
}

/****************************************************************************
 *			insert, lookup, delete keywords in items
 ***************************************************************************/

/* isItemKey: ****************************************************************
 *
 ****************************************************************************/

Boolean isItemKey(item,key)
     MHitem *item;
     char key;
{
  return((item->i_repr->key == key) ? True : False);
}

/* getItemByKey: *************************************************************
 *
 ****************************************************************************/

MHitem *getItemByKey(mptr,key)
     MHmenu *mptr;
     char key;
{
  register i;
  register MHitem *iptr;
  
  if (mptr == (MHmenu *)0 || key == EOS)
    return((MHitem *)0);
  
  for (i = 0; i < MDS_MAXITEM; i++) 
    if (iptr = mptr->i_table[i])
      if (isItemKey(iptr,key))
	return(iptr);
  
  return((MHitem *) 0);
}

/*****************************************************************************
 *			create, insert, delete requests
 ****************************************************************************/

/* deleteRequest: ************************************************************
 * delete a request with a given index and reduce the request count
 *****************************************************************************/

deleteRequest(iptr,rcnt)
     MHitem *iptr;
     int rcnt;
{
  if (rcnt < 0 || rcnt > MDS_MAXREQ)
    return(ERR_SIZE);
  
  CHECK(iptr);
  
  if (iptr->i_requests[rcnt]) {
    free(iptr->i_requests[rcnt]->reqData);
    free(iptr->i_requests[rcnt]->initialValue);
    free(iptr->i_requests[rcnt]->prelimInitialValue);
    free(iptr->i_requests[rcnt]);
    iptr->i_requests[rcnt] = (MHiReq *) 0;
    if (iptr->i_return->reqCnt > 0)
      iptr->i_return->reqCnt--;
  }
  BYE;
}

/* checkRequests: ************************************************************
 *
 *****************************************************************************/

checkRequests(iptr)
     MHitem *iptr;
{
  register char *ptr;
  register i;
  
  CHECK(iptr);
  
  if (! iptr->i_return)
    return(setError(ERR_NOERR));
  
  if (iptr->i_return->retPattern && *(iptr->i_return->retPattern)) {
    ptr = iptr->i_return->retPattern;
    iptr->i_return->reqCnt = countRequestsInPattern(ptr);
    for (i = 0; i < iptr->i_return->reqCnt; i++) /* look through requests */
      if (! iptr->i_requests[i])
	return(setError(ERR_SIZE));
    return(setError(ERR_NOERR));
  } else 
    return(setError(ERR_NOERR));
}

/* setItemKey: **************************************************************
 * 
 ****************************************************************************/

setItemKey(iptr, key)
     MHitem *iptr;
     char key;
{
  CHECK(iptr);
  
  iptr->i_repr->key = key;

  return(setError(ERR_NOERR));
}

/* deleteFunction: ************************************************************
 * delete application function in applnode
 *****************************************************************************/

deleteFunction(aptr, func)
     ApplicationNode *aptr;
     int (* func)();
{
  int i;
  CHECK(aptr);
  CHECK(func);
  
  for (i = 1; i < MDS_MAXFUNC; i++) {
    if (aptr->ApplFunc[i].func == func) {
      aptr->ApplFunc[i].func = (int (*)()) 0;
      free(aptr->ApplFunc[i].funcName);
      return(i);
    }
  }
  
  return(-1);
}

/******************************************************************************
 *			retrieve components
 *****************************************************************************/

/* getiPos: ******************************************************************
 *
 ****************************************************************************/

getiPos(iptr,x,y,t)
     MHitem *iptr;
     int *x, *y;
     MDS_POS_TYPE *t;
{
  CHECK(iptr);
  
  *x = iptr->i_Pos.x;
  *y = iptr->i_Pos.y;
  *t = iptr->i_Pos.posType;
  
  BYE;
}

/*getmPos : *****************************************************************
 *
 ****************************************************************************/

getmPos(mptr,x,y)
     MHmenu *mptr;
     int *x, *y;
{
  CHECK(mptr);
  
  *x = mptr->m_pos.x;
  *y = mptr->m_pos.y;
  
  BYE;
}

/*****************************************************************************
 *		    check access rights 
 ****************************************************************************/

/* setMenuFlag: **************************************************************
 * add menu flags (access rights, menu state)
 *****************************************************************************/

setMenuFlag(mptr, flags)
     MHmenu *mptr;
     int flags;
{
  CHECK(mptr);
  mptr->m_flags |= flags;
  BYE;
}

/* copyItemDescr ************************************************************
 * copy the description of an item 
 ****************************************************************************/

copyItemDescr(from,to)
     MHitem *from, *to;
{
  int i;
  
  if (from == (MHitem *) 0 || to == (MHitem *)0)
    return(ERR_NULL);
  
  to->flag = from->flag;
  
  free(to->i_repr->name);
  free(to->i_repr);
  
  to->i_repr = new_iRepr(from->i_repr->name, from->i_repr->key);
  
  to->i_Pos.x = from->i_Pos.x;
  to->i_Pos.y = from->i_Pos.y;
  to->i_Pos.posType = from->i_Pos.posType;
  
  AM_CPY(&from->activityMask, &to->activityMask);
  
  to->i_repr->key = from->i_repr->key;
  
  free(to->i_hlp->hlpmsg);
  to->i_hlp->hlpmsg = strCopy(from->i_hlp->hlpmsg);
  to->i_hlp->hlpFnct = from->i_hlp->hlpFnct;
  
  free(to->i_return->retPattern);
  to->i_return->retPattern = strCopy(from->i_return->retPattern);
  to->i_return->reqCnt = from->i_return->reqCnt;
  
  for (i = 0; i < MDS_MAXREQ; i++) {
    if (from->i_requests[i] != (MHiReq *) 0) {
      
      if (to->i_requests[i] == (MHiReq *) 0)
	to->i_requests[i] = newRequest();
      
      copyRequestDescr(from->i_requests[i], to->i_requests[i]);
    }
    else if (to->i_requests[i] != (MHiReq *) 0) {
      deleteRequest(to, i);
      to->i_requests[i] = (MHiReq *) 0;
    }
  }
  return(ERR_NOERR);
}

/* copyRequestDescr **********************************************************
 * copy the description of a request 
 ****************************************************************************/

copyRequestDescr(from, to)
     register MHiReq *from, *to;
{
  if (from == (MHiReq *) 0 || to == (MHiReq *) 0)
    return(setError(ERR_NULL));
  
  to->reqFnct = from->reqFnct;
  to->flags = from->flags;
  if (to->reqData)
    free(to->reqData);
  to->reqData  = strCopy(from->reqData);
  to->reqPtr = from->reqPtr;
  if (to->initialValue)
    free(to->initialValue);
  to->initialValue  = strCopy(from->initialValue);
  to->prelimInitialValue  = strCopy(from->prelimInitialValue);
  to->cnt = from->cnt;
  to->size = from->size;
  to->reqCntType = from->reqCntType;
  
  return(setError(ERR_NOERR));
}

/* cmpMenuDescr *************************************************************
 * compare two menu descriptions; returns 'True' if differences occur
 ****************************************************************************/

Boolean cmpMenuDescr(mptr1, mptr2)
MHmenu *mptr1, *mptr2;
{
  register i;

  /* if cmpMenuDescr() = False no differences occurred */

  if (mptr1 != (MHmenu *) 0) {

    if (mptr2 == (MHmenu *) 0) {
      setError(ERR_NULL);		    
      return(True);
    }

    if (mptr1->m_repr->reprType != mptr2->m_repr->reprType)
      return(True);
    if (strcmp(mptr1->m_repr->name, mptr2->m_repr->name) != 0)
      return(True);
 
    if ((mptr1->m_type != mptr2->m_type)
	|| (mptr1->m_flags != mptr2->m_flags)
	|| (mptr1->itemCnt != mptr2->itemCnt)
	|| (mptr1->maxItem != mptr2->maxItem)
	|| (mptr1->i_selCnt != mptr2->i_selCnt)
	|| (mptr1->i_selType != mptr2->i_selType))
      return(True);
  
    if (bcmp((char *) &mptr1->activityMask, (char *) &mptr2->activityMask,
	     sizeof(act_mask)) != 0)
      return(True);

    if ((mptr1->m_pos.x != mptr2->m_pos.x)
	|| (mptr1->m_pos.y != mptr2->m_pos.y))
      return(True);
    if (mptr1->m_pos.posType != mptr2->m_pos.posType)
      return(True);
 
    if (mptr1->m_key != mptr2->m_key)
      return(True);
  
  
    /* compare items */
    for (i = 0; i < MDS_MAXITEM; i++) {
      if (cmpItemDescr(mptr1->i_table[i], mptr2->i_table[i]) == True)
	return(True);
    }

    if (strcmp(mptr1->m_hlp->hlpmsg, mptr2->m_hlp->hlpmsg) != 0)
      return(True);
    if (mptr1->m_hlp->hlpFnct != mptr2->m_hlp->hlpFnct)
      return(True);
  }
  else if (mptr2 != (MHmenu *) 0) {
    setError(ERR_NULL);
    return(True);
  }

  return(False);
}

/* cmpItemDescr **************************************************************
 * compare two item descriptions; returns 'True' if differences occur
 ****************************************************************************/

Boolean cmpItemDescr(iptr1, iptr2)
MHitem *iptr1, *iptr2;
{
  register i;

  /* if cmpItemDescr() = False no differences occurred */

  if (iptr1 != (MHitem *) 0) {

    if (iptr2 == (MHitem *) 0) {
      setError(ERR_NULL);
      return(True);
    }
    
    if (iptr1->flag != iptr2->flag)
      return(True);

    if (strcmp(iptr1->i_repr->name, iptr2->i_repr->name) != 0)
      return(True);
    if (iptr1->i_repr->key != iptr2->i_repr->key)
      return(True); 

    if ((iptr1->i_Pos.x != iptr2->i_Pos.x)
	|| (iptr1->i_Pos.y != iptr2->i_Pos.y))
      return(True);
    if (iptr1->i_Pos.posType != iptr2->i_Pos.posType)
      return(True);
    
    if (bcmp((char *) &iptr1->activityMask, (char *) &iptr2->activityMask,
	     sizeof(act_mask)) != 0)
      return(True);

    if (strcmp(iptr1->i_hlp->hlpmsg, iptr2->i_hlp->hlpmsg) != 0)
      return(True);
    if (iptr1->i_hlp->hlpFnct != iptr2->i_hlp->hlpFnct)
      return(True);

    if (strCmp(iptr1->i_return->retPattern, iptr2->i_return->retPattern))
      return(True);
    if (iptr1->i_return->reqCnt != iptr2->i_return->reqCnt)
      return(True);

    for (i = 0; i < MDS_MAXREQ; i++) {

      if (cmpReqDescr(iptr1->i_requests[i], iptr2->i_requests[i]) == True)
	return(True);
    }
  }
  else if (iptr2 != (MHitem *) 0) {
    setError(ERR_NULL);
    return(True);
  }

  return(False);
}

/* cmpReqDescr ***************************************************************
 * compare two request descriptions; returns 'True' if differences occur
 ****************************************************************************/

Boolean cmpReqDescr(req1, req2)
MHiReq *req1, *req2;
{
  /* if cmpReqDescr() = False no differences occurred */

  if (req1 != (MHiReq *) 0) {

    if (req2 == (MHiReq *) 0) {
      setError(ERR_NULL);
      return(True);
    }

    if (req1->flags != req2->flags || req1->size != req2->size)
      return(True);

    if (req1->reqFnct != req2->reqFnct)
      return(True);

    if (strCmp(req1->initialValue, req2->initialValue) == True)
      return(True);

    if (strCmp(req1->prelimInitialValue, req2->prelimInitialValue) == True)
      return(True);

    if (strCmp(req1->reqData, req2->reqData) == True)
      return(True);

    if ((req1->cnt != req2->cnt)|| (req1->reqCntType != req2->reqCntType))
      return(True);
  }
  else if (req2 != (MHiReq *) 0) {
    setError(ERR_NULL);
    return(True);
  }

  return(False);
}

/* mds_function: **************************************************************
 * look if the specified function is a function defined in the MDS_functions
 * structure; otherwise the function must be a user defined function;
 *****************************************************************************/

mds_function(fnct)
     int (* fnct)();
{
  int i;
  
  for (i = 1; MDS_functions[i].func != (int (*)()) 0; i++)
    if (fnct == MDS_functions[i].func)
      return(i);
  
  return(0);
}

/* user_function *************************************************************
 * searches for index of input function in Array ApplFunc of Application node
 ****************************************************************************/

user_function(fnct, aptr)
     int (* fnct)();
     ApplicationNode *aptr;
{
  int i;
  
  for (i = 1; aptr->ApplFunc[i].func != (int (*)()) 0; i++)
    if (fnct == aptr->ApplFunc[i].func)
      return(i);
  
  return(0);
}

/* checkFunc ******************************************************************
 * function returns True if application node has application functions defined
 *****************************************************************************/

Boolean checkFunc(aptr)
ApplicationNode *aptr;
{
  register i;

  if (aptr == (ApplicationNode *) 0) {
    setError(ERR_NULL);
    return(False);
  }

  for (i = 1; i < MDS_MAXFUNC; i++)
    if (aptr->ApplFunc[i].func)
      return(True);

  return(False);
}

/* getFuncByName ************************************************************
 * function to get the user defined function via its name; returns index
 * if found                 
 ****************************************************************************/

getFuncByName(aptr, name)
ApplicationNode *aptr;
char *name;
{
  register i;

  if (name == (char *) 0)
    return (0);

  for (i = 1; i < MDS_MAXFUNC; i++)
    if (aptr->ApplFunc[i].funcName 
	&& strcmp(aptr->ApplFunc[i].funcName, name) == 0)
      return (i);

  return(0);
}

