/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
#include <exLayout.h>
#include <exFonts.h>
#include <exErrors.h>

#include <X11/cursorfont.h>

#include "exMenu.h"
#include "m_int.h"
#include <pixels/tick.bmp>
#include <pixels/Arrow.bmp>
#include <pixels/close.bmp>

static char SccsID[] = "@(#)exMenu.c	1.5\t1/18/93";            /* sccs keyword */

static Atom exMenuReturnValue;
static Atom exMenuToggleButton;

static Pixmap enabledTick  = (Pixmap) NULL;
static Pixmap disabledTick  = (Pixmap) NULL;
static Pixmap enabledArrow  = (Pixmap) NULL;
static Pixmap disabledArrow  = (Pixmap) NULL;
static Pixmap closePix = (Pixmap) NULL;

static GC normal_gc = (GC) NULL;
static GC select_gc = (GC) NULL;
static GC disable_gc = (GC) NULL;
static Cursor exMenuCursor;
static Cursor exMenuBarCursor;
static Cursor exMenuClockCursor;

static Widget grabbedWidget = NULL;	/* widget that has been grabbed */

static Display *display;

/* external interface: exported functions */
Boolean exCallMenu();
exMenubar *exCreateMenubar();
char *exMenuGetReturnValue();

/* forward declarations of functions associated with the menu pane */
static void exMenuEnterProc();
static void exMenuLeaveProc(); 
static void exMenuBtnDownProc(); 
static void exMenuBtnUpProc(); 
static void exMenuExposeProc(); 
static void exMenubarExposeProc();
static void exConfigurePane();
static void exMenuMotionProc();

/* foward declarations for functions associated with menubar buttons */
static void exButtonDownProc();
static void exButtonHelpProc();

/* forward declarations of the functions associated with scrollbars */
static void exScrollbar();
static void exScrollMenu();
static void exSetScrollbar();

static void exPanelScroll();

void exGrabMenu();
void postMenu();
static exMenu *getActiveMenu();
static void dummy();
static void exTimerEvent();
static void destroyMenu();
static exMenu *pullMenuDown();
static Widget findButton();
static void exMakeArrowButton();
static void drawDoneBox();
static void selectDoneBox();
static void drawArrow();
static Boolean requestFromUser();
static void redrawDoneBox();
void displayItem();
static void displayTitle();
static void drawColSep();
static void exDisplayMenu();
static void redrawMenu();
static void assembleReturnValues();
static void clearMenuDescr();
static void eraseMenu();
static findItem();
static void helpUser();
static Boolean exMakeDefaultItem();
static Boolean nestedItem();
static Boolean checkDefaultMenu();
static drawTick();
static drawItem();
static verticalDist();
static Boolean checkactMask();
void checkMenuActivityMask();
static setInitialValue();
exMenu *exCreateMenuStruct();
static Boolean exCreateMenuRepr();
Boolean exCreatePulldownMenuRepr();
Boolean exCreatePopupMenuRepr();
static Boolean thisMenuActive();
void setMenuActive();
static Boolean inArrowRegion();
static Boolean displayable();

static menuEntry *exCreateMenuEntry();
static Boolean selectNestedItem();
static Boolean nestedItem();

static void Layout();
static void exPanelLayout();
static void exMenuLayout();

static titleHeight();
static doneHeight();
static sepHeight();

void _exInitMenus(d)
     Display *d;
{
  static Boolean init = False;
  Window w;
  GC gc;
  int depth;

  display = d;

  if (init == True)
    return;

  w = XDefaultRootWindow(d);

  normal_gc = XCreateGC(d, w, 0, 0);
  XSetBackground(d, normal_gc, exWhitePixel(d));
  XSetForeground(d, normal_gc, exBlackPixel(d));
  XSetFunction(d, normal_gc, GXcopy);
  XSetFillStyle(d, normal_gc, FillSolid);
  XSetLineAttributes(d, normal_gc, 1, LineSolid, CapButt, JoinMiter);

  select_gc = XCreateGC(d, w, 0, 0);
  XSetBackground(d, select_gc, exBlackPixel(d));
  XSetForeground(d, select_gc, exWhitePixel(d));
  XSetFunction(d, select_gc, GXcopy);
  XSetFillStyle(d, select_gc, FillSolid);
  XSetLineAttributes(d, select_gc, 1, LineSolid, CapButt, JoinMiter);

  disable_gc = XCreateGC(d, w, 0, 0);
  XSetBackground(d, disable_gc, exWhitePixel(d));
  XSetForeground(d, disable_gc, exBlackPixel(d));
  XSetFunction(d, disable_gc, GXcopy);
  XSetFillStyle(d, disable_gc, FillOpaqueStippled);
  XSetStipple(d, disable_gc, exLight_greyBM(d));
  XSetLineAttributes(d, disable_gc, 1, LineSolid, CapButt, JoinMiter);

  /* create tick pixmaps */
  gc = XCreateGC(d, w, 0, 0);
  depth = DefaultDepth(d, DefaultScreen(d));

  XSetFunction(d, gc, GXcopy);
  enabledTick = XCreatePixmapFromBitmapData(d, w, tick_bits, tick_width, 
			tick_height, exBlackPixel(d), exWhitePixel(d), depth);
  disabledTick = XCreatePixmap(d, w, tick_width, tick_height, depth);
  XCopyArea(d, exGreyPM(d), disabledTick, gc, 
	    0, 0, tick_width, tick_height, 0, 0);
  XSetFunction(d, gc, GXand);
  XCopyArea(d, enabledTick, disabledTick, gc, 
	    0, 0, tick_width, tick_height, 0, 0);
/*  enabledTick = XCreateBitmapFromData(d, w,tick_bits, tick_width, tick_height);*/
  
  /* create the arrow pixmaps */
  XSetFunction(d, gc, GXcopy);

  enabledArrow = 
    XCreatePixmapFromBitmapData(d, w, Arrow_bits, Arrow_width, Arrow_height,
				exBlackPixel(d), exWhitePixel(d), depth);
  disabledArrow = XCreatePixmap(d, w, Arrow_width, Arrow_height, depth);
  XCopyArea(d, exGreyPM(d), disabledArrow, gc, 
	    0, 0, Arrow_width, Arrow_height, 0, 0);
  XSetFunction(d, gc, GXand);
  XCopyArea(d, enabledArrow, disabledArrow, gc, 
	    0, 0, Arrow_width, Arrow_height, 0, 0);

  /* create pixmap for done box */
  XSetFunction(d, gc, GXcopy);
  closePix = 
    XCreatePixmapFromBitmapData(d, w, close_bits, close_width, close_height,
				exBlackPixel(d), exWhitePixel(d), depth);

  XFreeGC(d, gc);

  /* create the menu-cursor */
  exMenuCursor = XCreateFontCursor(d, XC_hand1);
  exMenuClockCursor = XCreateFontCursor(d, XC_watch);
  exMenuBarCursor = XCreateFontCursor(d, XC_dotbox);

  /* create the atoms for the property names for return values */
  exMenuReturnValue = XInternAtom(d, EX_MENU_RETURN_VALUE_ATOM_NAME, False);
  exMenuToggleButton = XInternAtom(d, EX_MENU_TOGGLE_BUTTON_ATOM_NAME, False);
  
  init = True;
}

/***************************************************************************
 *
 *		PRIVATE PROCEDURES FOR MENU CREATION
 *
 ***************************************************************************/

/* createMenuBar: **********************************************************
 * creates a menu bar to be used with pulldown menus; A function is registered 
 * that is called whenever a ProperNotify event occurs. By this mechanism 
 * Return values are provided to the application (see exButtonDownProc())
 ****************************************************************************/

exMenubar *createMenuBar(parent, mb_args, mb_cnt, getReturnValue)
     Widget parent;
     Arg mb_args[];
     int mb_cnt;
     void (* getReturnValue)();
{
  Arg args[20];
  static XtTranslations ptt = NULL;
  exMenubar *mb;
  int cnt;
  static char menubar_translations[] = 
    " <PropertyNotify>  	: getReturnValue() \n\
      <Expose>			: exMenubarExposeProc()";

  static XtActionsRec menubar_actions[] = {
    { "getReturnValue", NULL },
    { "exMenubarExposeProc", exMenubarExposeProc }
  };

  /***************************************************************************
   *to make the toolkit happy, we add a function even if the caller did not 
   *provide one; otherwise the toolkit complains the missing action.
   **************************************************************************/

  menubar_actions[0].proc = (getReturnValue) ? getReturnValue : dummy;

  XtAddActions(menubar_actions, 2);

  if (! ptt)
    ptt = XtParseTranslationTable(menubar_translations);

  /* create menubar as row-column widget */

  /* get provided arguments */
  for (cnt = 0; cnt < mb_cnt; cnt++) {
    args[cnt].name = mb_args[cnt].name;
    args[cnt].value = mb_args[cnt].value;
  }

  /* define additional arguments */
  XtSetArg(args[cnt], XmNpacking, XmPACK_TIGHT); cnt++;
  XtSetArg(args[cnt], XmNorientation, XmHORIZONTAL); cnt++;

  mb = (exMenubar *) Malloc(sizeof(exMenubar));
  
  mb->ptt = ptt;
  mb->num_buttons = 0;
  if (! (mb->menubar = XtCreateManagedWidget("menubar", 
			    xmRowColumnWidgetClass, parent, args, cnt))) {
    setError(ERR_CREATE);
    return((exMenubar *) NULL);
  }

  XtOverrideTranslations(mb->menubar, ptt);
  
  return(mb);
}

/* exCreateMenuStruct: **************************************************
 * create a menu from a menu description; 
 ***************************************************************************/

exMenu *exCreateMenuStruct(aptr, mptr)
     ApplicationNode *aptr;
     MHmenu *mptr;
{
  register exMenu *m;
  register i, cnt;

  if (! (m = (exMenu *) Malloc(sizeof(exMenu)))) {
    setError(ERR_ALLOC);
    return((exMenu *) NULL);
  }

  if (isPopup(mptr))
    m->flags = POPUP;
  else if (isPulldown(mptr))
    m->flags = PULLDOWN;
  
  m->flags |= MH_FIRST_MAP;		/* menu mapped the first time */
  m->mapped = False;			/* not displayed initially */
  m->erase = True;			/* erase initial values */
  m->selType = mptr->i_selType;
  m->prev = -1;
  m->itemCnt = 0;
  m->selItemCnt = 0;
  m->hlpMsg = strCopy(mptr->m_hlp->hlpmsg);
  m->hlpFnct = mptr->m_hlp->hlpFnct;
  m->mds_structure = mptr;
  m->name = strCopy(mptr->m_repr->name);
  m->applName = aptr ? strCopy(aptr->applicationName) : 
    strCopy(currentApplication()->applicationName);
  m->itemFont = NULL;
  m->titleFont = NULL;
  m->sb_width = 0;
  m->selCnt = mptr->i_selCnt;
  m->xs = m->ys = m->yb = 0;
  m->titleHeight = m->doneHeight = m->sepHeight = 0;
  m->titleFontHeight = 0;

  AM_CPY(&mptr->activityMask, &m->initialMask);
  AM_CPY(&m->initialMask, &m->activityMask);
  
  for (i = 0, cnt = 0; i < MDS_MAXITEM; i++) {
    if (mptr->i_table[i] == (MHitem *) 0)
      continue;
    
    if (! (m->entries[cnt] = exCreateMenuEntry(mptr->i_table[i]))) {
      setError(ERR_CREATE);
      return((exMenu *) NULL);
    }

    m->itemCnt++;
    
    if (mptr->i_table[i]->flag & MH_DEFAULTITEM)
      exMakeDefaultItem(m, cnt);
    
    cnt++;
    
  }
  
  m->key = '\0';

  checkMenuActivityMask(m, False);
  
  m->itemDisplayCnt =
    MIN(mptr->maxItem, mptr->itemCnt);
  m->topItem = 0;
  if (checkDefaultMenu(mptr) == True)
    m->erase = False;
  
  return(m);
}

/* exCreatePulldownMenuRepr: **************************************************
 * creates the visual representation and the button for the pulldown-menu
 ***************************************************************************/

Boolean exCreatePulldownMenuRepr(menubar, m)
     exMenubar *menubar;
     exMenu *m;
{
  Arg args[5];
  int cnt;
  XmFontList fl = NULL;

  static char button_translations[] = 
    " Button1 <EnterWindow>	: Arm() exButtonDownProc() \n \
      <EnterWindow>             : Arm() \n \
      <LeaveWindow>		: Activate() \n \
      <Btn1Down>		: Arm() exButtonDownProc() \n \
      <Btn1Up>          	: Activate() \n \
      <Btn3Down>                : exButtonHelpProc() \n \
      <PropertyNotify>  	: Activate()";

  static XtActionsRec button_actions[] = {
    { "exButtonDownProc", exButtonDownProc },
    { "exButtonHelpProc", exButtonHelpProc }
  };

  static XtTranslations ptt = NULL;

  if (! ptt) {
    XtAddActions(button_actions, 2);
    ptt = XtParseTranslationTable(button_translations);
  }

  cnt = 0;
  XtSetArg(args[cnt], XmNlabelString, 
	XmStringCreate(m->name, XmSTRING_DEFAULT_CHARSET)); cnt++; 
  XtSetArg(args[cnt], XmNuserData, (caddr_t) m); cnt++;

  m->button = 
    XtCreateManagedWidget(m->name, xmPushButtonWidgetClass, 
			              menubar->menubar, args, cnt);
  if (! m->button) {
    setError(ERR_CREATE);
    return(False);
  }

  /* get font resource from buttons */
  cnt = 0;
  XtSetArg(args[cnt], XmNfontList, &fl); cnt++;
  XtGetValues(m->button, args, cnt);

  if (! fl) {
    m->titleFont = exBoldFont(display);
    exError("Create pulldown menu", " cannot obtain title font!", NULL);
  } else 
    m->titleFont= getFont(fl);

  m->titleFontHeight = exFontHeight(m->titleFont);
  m->menubar = menubar;
  
  XtOverrideTranslations(m->button, ptt);

  /* update menubar information */
  menubar->buttons[menubar->num_buttons].menu = m; 
  menubar->buttons[menubar->num_buttons].button = m->button; 
  menubar->buttons[menubar->num_buttons].ptt = ptt;
  menubar->num_buttons++;

  m->cols = 1;
  m->itemsPerCol = m->itemCnt;
  if (m->itemCnt > m->itemDisplayCnt) 
    m->flags |= SCROLLABLE;		/* create a list of items */

  if (! (m->selCnt == 1 && m->selType == MDS_CNT_EQ))
    m->flags |= MULTI_SEL;

  return(exCreateMenuRepr(menubar, m->button, m));
}

/* exCreatePopupMenuRepr: **************************************************
 * creates the visual representation for the popup-menu
 ***************************************************************************/

Boolean exCreatePopupMenuRepr(parent, m)
     Widget parent;
     exMenu *m;
{
  m->menubar = (exMenubar *) 0;

  m->cols = 1;
  m->itemsPerCol = m->itemCnt;
  if (m->itemCnt > m->itemDisplayCnt) 
    m->flags |= SCROLLABLE;		/* create a list of items */

  if (! (m->selCnt == 1 && m->selType == MDS_CNT_EQ))
    m->flags |= MULTI_SEL;

  return(exCreateMenuRepr(NULL, parent, m));
}

/* exCreatePanelRepr: ******************************************************
 * creates the visual representation for the panel
 ***************************************************************************/

exCreatePanelRepr(parent, m, cols, width)
     Widget parent;
     exMenu *m;
     int cols;
     int width;
{
  m->menubar = (exMenubar *) 0;
  m->flags &= ~(POPUP | PULLDOWN);
  m->flags |= PANEL;
  m->cols = cols;
  m->itemsPerCol =  ROUNDOFF(m->itemCnt, cols);
  m->itemDisplayCnt = MIN(m->itemDisplayCnt, m->itemsPerCol);

  if (m->itemsPerCol > m->itemDisplayCnt) 
    m->flags |= SCROLLABLE;		/* create a list of items */

  if (width > 0) {
    m->flags |= FIXED_WIDTH;
    m->colWidth = width;
  }
  
  if (! (m->selCnt == 1 && m->selType == MDS_CNT_EQ))
    m->flags |= MULTI_SEL;

  return(exCreateMenuRepr(NULL, parent, m));
}

/* exCreateMenuRepr: ********************************************************
 * create the visual representation for the menu-body
 ****************************************************************************/

static Boolean exCreateMenuRepr(menubar, parent, m)
     exMenubar *menubar;
     Widget parent;
     exMenu *m;
{
  int width, height, s_height;
  Arg args[12];
  int cnt;
  XmFontList fl = NULL;

  static char exMenu_translations[] = 
   "<EnterWindow>	: exMenuEnterProc()   \n \
    <LeaveWindow>	: exMenuLeaveProc()   \n \
    <ButtonPress>   	: exMenuBtnDownProc() \n \
    <Btn1Up>     	: exMenuBtnUpProc()   \n \
    <Expose>		: exMenuExposeProc()  \n \
    <Configure>   	: exConfigurePane()   \n \
    <Motion>		: exMenuMotionProc()";

  static XtActionsRec exMenu_actions[] = {
    { "exMenuEnterProc",  exMenuEnterProc},
    { "exMenuLeaveProc",  exMenuLeaveProc}, 
    { "exMenuBtnDownProc",exMenuBtnDownProc},
    { "exMenuBtnUpProc",  exMenuBtnUpProc},
    { "exMenuExposeProc", exMenuExposeProc},
    { "exMenuMotionProc", exMenuMotionProc},
    { "exConfigurePane",  exConfigurePane}
  };

  static char exPanel_translations[] = 
   "<Btn1Down>   	: exPanelScroll()    \n \
    <Btn1Up>     	: exPanelScroll()    \n \
    <Motion>		: exPanelScroll()";

  static XtActionsRec exPanel_actions[] = {
    { "exPanelScroll",    exPanelScroll}
  };

  static XtTranslations ptt = NULL, panel_ptt = NULL;

  if (! ptt || !panel_ptt) {
    XtAddActions(exMenu_actions, 7);
    ptt = XtParseTranslationTable(exMenu_translations);
    XtAddActions(exPanel_actions, 1);
    panel_ptt = XtParseTranslationTable(exPanel_translations);
  }

  m->ptt = ptt;
  m->menubar = menubar;

  if (m->flags & PANEL) {			/* panel */
    m->w = parent;
    m->shape = parent;
    cnt = 0;
    XtSetArg(args[cnt], XmNuserData, (caddr_t) m); cnt++;
    XtSetValues(m->w, args, cnt);		
  } else {					/* pulldown or popup */
    cnt = 0;
    /* create popup menu pane */
    XtSetArg(args[cnt], XmNborderWidth, 0); cnt++;
    XtSetArg(args[cnt], XmNx, m->pos.x); cnt++;
    XtSetArg(args[cnt], XmNy, m->pos.y); cnt++;
    XtSetArg(args[cnt], XmNkeyboardFocusPolicy, XmPOINTER); cnt++;
    
    m->w = XtCreatePopupShell(m->name,
			 overrideShellWidgetClass, parent, args, cnt);
    
    if (! m->w) {
      setError(ERR_CREATE);
      return(False);
    }

    XtAddCallback(m->w, XmNdestroyCallback, destroyMenu, m);

    /* create top-level form */
    cnt = 0;
    XtSetArg(args[cnt], XmNborderWidth, 0); cnt++;
    XtSetArg(args[cnt], XmNshadowThickness, 0); cnt++;
    XtSetArg(args[cnt], XmNtopShadowPixmap, exWhitePM(display)); cnt++;
    XtSetArg(args[cnt], XmNbottomShadowPixmap, exWhitePM(display)); cnt++;
    
    m->shape = 
      XtCreateManagedWidget("shape", xmFormWidgetClass, m->w, args, cnt);
    
  }

  cnt = 0;
  if (! (m->flags & PANEL)) {		/* leave room for shadow */
    XtSetArg(args[cnt], XmNrightOffset, SHADOW_WIDTH); cnt++;
    XtSetArg(args[cnt], XmNbottomOffset, SHADOW_WIDTH); cnt++;
  }
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, (m->flags & PANEL) ? 0 : 1); cnt++;

  m->form = 
    XtCreateManagedWidget(m->flags & PANEL ? m->name : "form", 
			  xmFormWidgetClass, m->shape, args, cnt);

  if (m->flags & POPUP) {	      		/* popup menu */
    /* create title */
    cnt = 0;
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNalignment, XmALIGNMENT_CENTER); cnt++;
    XtSetArg(args[cnt], XmNlabelString, 
	     XmStringCreate(m->name, XmSTRING_DEFAULT_CHARSET)); cnt++; 
    m->title = XtCreateManagedWidget("title", 
				     xmLabelWidgetClass, m->form, args, cnt);
    /* get title font resource */
    cnt = 0;
    XtSetArg(args[cnt], XmNfontList, &fl); cnt++;
    XtGetValues(m->title, args, cnt);
    
    if (! fl) {
      m->titleFont = exBaseFont(display);
      exError("Create popup menu", "cannot obtain title font!", NULL);
    } else
      m->titleFont = getFont(fl);

    m->titleFontHeight = exFontHeight(m->titleFont);
    

    /* create separator */
    cnt = 0;
    XtSetArg(args[cnt], XmNseparatorType, XmDOUBLE_LINE); cnt++;
    XtSetArg(args[cnt], XmNorientation, XmHORIZONTAL); cnt++;
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
    XtSetArg(args[cnt], XmNtopWidget, m->title); cnt++;
    m->title_sep = XtCreateManagedWidget("title_sep", 
			 xmSeparatorWidgetClass, m->form, args, cnt);
    
  } else {					/* pulldown menu */
    m->title = NULL;
    m->title_sep = NULL;
  }
  
  if (m->flags & MULTI_SEL && !(m->flags & PANEL)) {/* create done box */
    /* create done box */
    cnt = 0;
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
    XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;
    XtSetArg(args[cnt], XmNborderWidth, 0); cnt++;
    XtSetArg(args[cnt], XmNheight, 
	     exFontHeight(m->titleFont) + 2 * MH_MARGIN); cnt++;
    m->done_box = XtCreateManagedWidget("done_box", 
			     xmDrawingAreaWidgetClass, m->form, args, cnt);
    XtAddCallback(m->done_box, XmNexposeCallback, redrawDoneBox, m);

    /* create separator */
    cnt = 0;
    XtSetArg(args[cnt], XmNseparatorType, XmDOUBLE_LINE); cnt++;
    XtSetArg(args[cnt], XmNorientation, XmHORIZONTAL); cnt++;
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
    XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_WIDGET); cnt++;
    XtSetArg(args[cnt], XmNbottomWidget, m->done_box); cnt++;
    m->done_sep = XtCreateManagedWidget("done_sep", 
			 xmSeparatorWidgetClass, m->form, args, cnt);
  }

  if (m->flags & SCROLLABLE) {			/* create scrollbar */
    m->slider_pos = -1;
    cnt = 0;
    XtSetArg(args[cnt], XmNborderWidth, 2); cnt++;
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
    if (m->flags & POPUP) {
      XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
      XtSetArg(args[cnt], XmNtopWidget, m->title_sep); cnt++;
    } else {
      XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
    }
    
    if (m->flags & MULTI_SEL) {
      XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_WIDGET); cnt++;
      XtSetArg(args[cnt], XmNbottomWidget, m->done_sep); cnt++;
    } else {
      XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
    }

    XtSetArg(args[cnt], XmNminimum, 0); cnt++;
    XtSetArg(args[cnt], XmNvalue, 0); cnt++;
    if (m->flags & PANEL) {
      XtSetArg(args[cnt], XmNuserData, (caddr_t) m); cnt++;
    }

    m->scrollbar = 
     XtCreateManagedWidget("scroll_bar", xmScrollBarWidgetClass, 
			                               m->form, args, cnt);
    if (! m->scrollbar) {
      setError(ERR_CREATE);
      return(False);
    }

    if (m->flags & PANEL)
      XtOverrideTranslations(m->scrollbar, panel_ptt);
  }

  /* create item area */
  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, (caddr_t) m); cnt++;
  if (m->flags & SCROLLABLE) {
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
    XtSetArg(args[cnt], XmNleftWidget, m->scrollbar); cnt++;
  } else {
    XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  }

  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;

  if (m->flags & POPUP) {
    XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
    XtSetArg(args[cnt], XmNtopWidget, m->title_sep); cnt++;
  } else {
    XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  }

  if (m->flags & MULTI_SEL && !(m->flags & PANEL)) {
    XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_WIDGET); cnt++;
    XtSetArg(args[cnt], XmNbottomWidget, m->done_sep); cnt++;
  } else {
    XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  }
  
  m->pane = XtCreateWidget("pane",
			  xmBulletinBoardWidgetClass, m->form, args, cnt);   
  
  /* get menu font resource */
  cnt = 0;
  XtSetArg(args[cnt], XmNtextFontList, &fl); cnt++;
  XtGetValues(m->pane, args, cnt);
  
  if (! fl) {
    exError("Create menu", "cannot obtain item font!", NULL);
    m->itemFont = exBaseFont(display);
  } else
    m->itemFont = getFont(fl);

  if (m->flags & SCROLLABLE) 
    m->sb_width = exFontHeight(m->itemFont);

  Layout(m, &width, &height);

  m->menuHeight = height;			/* height of menu pane */
  m->menuWidth  = width;			/* width of item area */
  
  cnt = 0;
  XtSetArg(args[cnt], XmNwidth, width); cnt++;
  XtSetArg(args[cnt], XmNheight, height); cnt++;
  XtSetValues(m->pane, args, cnt);

  /* set scrollbar values */
  if (m->flags & SCROLLABLE) {			/* scrollbar */
    s_height = m->menuHeight;

    cnt = 0;
    XtSetArg(args[cnt], XmNwidth, m->sb_width); cnt++;
    XtSetArg(args[cnt], XmNsliderSize, (int)((float) s_height * 
			    ((float) m->itemDisplayCnt / 
			     (float) ROUNDOFF(m->itemCnt, m->cols)))); cnt++;
    XtSetArg(args[cnt], XmNmaximum, s_height); cnt++;
    XtSetValues(m->scrollbar, args, cnt);
  }

  XtManageChild(m->pane);
  XtOverrideTranslations(m->pane, ptt);

  setError(ERR_NOERR);
  return(True);
}
 
/* exCreateMenuEntry: *******************************************************
 * create an entry structure by copying appropiate values from the provided
 * MDS-item structure 
 ***************************************************************************/

static menuEntry *exCreateMenuEntry(iptr)
     MHitem *iptr;
{
  register i, cnt;
  
  register menuEntry *eptr;

  if (! (eptr = (menuEntry *) calloc(1, sizeof(menuEntry)))) {
    setError(ERR_CREATE);
    return((menuEntry *) NULL);
  }

  for (i = 0, cnt = 0; i < iptr->i_return->reqCnt; i++) {
    if (iptr->i_requests[i]->initialValue)
      cnt++;
    eptr->requests[i] = (MHiReq *) Malloc(sizeof(MHiReq));
    copyRequestDescr(iptr->i_requests[i], eptr->requests[i]);
  }
  
  eptr->flags = iptr->flag;
  
  if (cnt == iptr->i_return->reqCnt)	/* all requests have initial values */
    eptr->flags |= MH_INITVALUE;
  
  eptr->name = strCopy(iptr->i_repr->name);
  eptr->key = iptr->i_repr->key;
  AM_CPY(&iptr->activityMask, &eptr->activityMask);
  
  eptr->i_return = strCopy(iptr->i_return->retPattern);
  eptr->reqCnt = iptr->i_return->reqCnt;
  
  eptr->hlpFnct = iptr->i_hlp->hlpFnct;
  eptr->hlpMsg = strCopy(iptr->i_hlp->hlpmsg);
  
  if (nestedItem(eptr))
    eptr->flags |= NESTED_ITEM;
  
  return(eptr);
}

/* Layout: *******************************************************************
 * determine representation and position on the screen
 ****************************************************************************/

static void Layout(m, w, h)
     exMenu *m;
     int *w, *h;
{
  if (m->flags & PANEL)
    exPanelLayout(m, w, h);
  else
    exMenuLayout(m, w, h);
}

/* exMenuLayout: *************************************************************
 * determine menu representation and position on the screen
 ****************************************************************************/
static void exMenuLayout(m, w, h)
     exMenu *m;
     int *w, *h;
{
  int arrow, indx, x, y, l, len, key, 
  	itemWidth, menuHeight, fontHeight;
  register menuEntry *eptr;
  static char str[4] = {'^', ' ', ' ', EOS};
  
  XFontStruct *f, *tf;
  Boolean nested;
  
  tf = m->titleFont;
  f  = m->itemFont;

  fontHeight = exFontHeight(f);

  x = ITEM_INDENT + MH_BORDER;

  if (m->flags & POPUP) {
    m->sepHeight = sepHeight(m->title_sep);
    m->titleHeight = titleHeight(m);
  }

  if (m->flags & MULTI_SEL) {   
    m->sepHeight = sepHeight(m->done_sep);
    m->doneHeight = doneHeight(m);
  }
  
  m->baseLine = tf->max_bounds.ascent;
  key = 0;
  arrow = 0;

  if (m->flags & POPUP) {		/* calculate title for popups */
    itemWidth = exTextWidth(tf, m->name);
  } else {				/* no title with pulldowns */
    itemWidth = 0;
  }

  y = MH_BORDER;
  menuHeight = 0;

  m->lm = x;				/* start of item area */

  for (indx = 0; indx < m->itemCnt; indx++) {
    eptr = m->entries[indx];
    nested = nestedItem(eptr);

    eptr->xpos = x; 
    if (nested || eptr->flags & MH_TOGGLEITEM) {
      m->flags |= ARROW_OR_TICK;
      arrow = ITEM_GAP;
      eptr->height = MAX(fontHeight, Arrow_height);
      eptr->baseLine = f->max_bounds.ascent + (fontHeight - Arrow_height) / 2;
      if (fontHeight >= Arrow_height) {
	eptr->ypos = y;
	eptr->ppos = y + (fontHeight - Arrow_height) / 2;
      } else {
	eptr->ypos = y + (Arrow_height - fontHeight) / 2;
	eptr->ppos = y;
      }
    } else {
      eptr->ypos = y;
      eptr->height = fontHeight;
      eptr->baseLine = f->max_bounds.ascent;
    }

    len = exTextWidth(f, eptr->name);
    if (len > itemWidth) 
      itemWidth = len;
    
    y += eptr->height;

    if (eptr->key != EOS) {
      str[1] = eptr->key;
      if ((l = exTextWidth(m->itemFont, str)) > key)
	key = l;
    }
    
    if (indx < m->itemDisplayCnt)	/* height of visible item-area */
      menuHeight = y;
  }
  
  itemWidth += 2;			/* leave some space */

  if (arrow) 				/* consider arrow(s) */
    itemWidth += arrow + Arrow_width;
  
  if (key)				/* consider key(s) */
    itemWidth += key + ITEM_GAP;	
  
  if (m->flags & MULTI_SEL) {		/* consider done-box */
    int dw, ct;
    char msg[20];

    ct = (int)m->selType - (int)MDS_CNT_NULL - 1;
    if (ct > 0)
      sprintf(msg,"done (%s%d)", countType[ct][1], m->selCnt);
    else
      sprintf(msg,"done (%d)", m->selCnt);

    dw = exTextWidth(tf, msg);		/* done-message width */
    
    if (itemWidth + m->sb_width < 
			dw + 3 * MH_MARGIN + m->titleFontHeight + dw + 2)
      itemWidth = dw + 3 * MH_MARGIN + m->titleFontHeight + dw + 2;
  }
  
  *w = itemWidth + 6 * MH_BORDER;
  *h = menuHeight;

  for (indx = 0; indx < m->itemCnt; indx++) {
    if (key)  	  			/* at least one key specified */
      m->entries[indx]->labelOffset = key + ITEM_GAP;
    else
      m->entries[indx]->labelOffset = 0;
    m->entries[indx]->width = itemWidth;
  }
}

/* exPanelLayout: *************************************************************
 * determine panel representation and position on the screen
 ****************************************************************************/

static void exPanelLayout(m, w, h)
     exMenu *m;
     int *w, *h;
{
  int arrow, i, indx, x, y, l, len, key, width, height;
  int rowCnt, colCnt, col_gap;

  register menuEntry *eptr;
  static char str[4] = {'^', ' ', ' ', EOS};
  
  XFontStruct *f;
  Boolean nested;
  int fontHeight;
  
  x = ITEM_INDENT + MH_BORDER;

  m->baseLine = 0;
  col_gap = exTextWidth(m->itemFont, "   ");

  m->lm = x;				/* start of item area */

  f = m->itemFont;
  fontHeight = exFontHeight(m->itemFont);

  height = 0;

  for (colCnt = 0; colCnt < m->cols; colCnt++) {
    y = MH_BORDER;
    width = 0;
    key = 0;
    arrow = 0;

    for (rowCnt = 0; rowCnt < m->itemsPerCol; rowCnt++) {
      indx = colCnt * m->itemsPerCol + rowCnt;
      if (indx >= m->itemCnt)
	break;

      eptr = m->entries[indx];
      nested = nestedItem(eptr);

      eptr->xpos = x; 
      if (nested || eptr->flags & MH_TOGGLEITEM) {
	eptr->height = MAX(fontHeight, Arrow_height);
	eptr->baseLine = 
	  f->max_bounds.ascent + (fontHeight - Arrow_height) / 2;
	if (fontHeight >= Arrow_height) {
	  eptr->ypos = y;
	  eptr->ppos = y + (fontHeight - Arrow_height) / 2;
	} else {
	  eptr->ypos = y + (Arrow_height - fontHeight) / 2;
	  eptr->ppos = y;
	}
      } else {
	eptr->ypos = y;
	eptr->height = fontHeight;
	eptr->baseLine = f->max_bounds.ascent;
      }
      
      y += eptr->height;
      
      if (eptr->key != EOS) {
	str[1] = eptr->key;
	if ((l = exTextWidth(m->itemFont, str)) > key)
	  key = l;
      } else 
	key = 0;

      if (nested || eptr->flags & MH_TOGGLEITEM)
	arrow = 6;
      else
	arrow = 0;

      if (rowCnt < m->itemDisplayCnt)	/* height of visible item-area */
	if (y > height)
	  height = y;
      
      if (m->flags & FIXED_WIDTH) {
	width =  m->colWidth;
      } else {
	len = exTextWidth(f, eptr->name);
	if (arrow)
	  len = len + arrow + Arrow_width;
	if (key)
	  len = len + key + ITEM_GAP;

	if (len > width) 
	  width = len;
      }
    }
    
    for (i = colCnt * m->itemsPerCol; 
	 i < m->itemCnt && i < (colCnt + 1) * m->itemsPerCol; i++) {
    if (key)  	  			/* at least one key specified */
      m->entries[i]->labelOffset = key + ITEM_GAP;
    else
      m->entries[i]->labelOffset = 0;
      m->entries[i]->width = width;
    }
    
    x += width + col_gap;		/* advance x to next col */
  }
    
  width = x - col_gap;

  *w = width + MH_BORDER;
  *h = height + MH_BORDER;
}

/* postPanel: ****************************************************************
 * post a panel permanently on the screen;
 ****************************************************************************/

void postPanel(m, x, y)
     exMenu *m;
     int x, y;
{
  int width, height;
  
  width = m->menuWidth + m->sb_width;
  height = m->menuHeight;
  
  if (m->flags & POPUP) 
    height += m->titleHeight + m->sepHeight;

  if (m->flags & MULTI_SEL)
    height += m->doneHeight + m->sepHeight;

  exWidgetPos(m->w, x, y, width, height);
  XtMapWidget(m->w);
}

/* postMenu: ***************************************************************
 * post a menu wait until the user has finished menu selection 
 ***************************************************************************/

void postMenu(m)
     exMenu *m;
{
  int x, y, width, height;
  XEvent e;

  if (m->pos.posType == MDS_POS_ABS) {
    x = m->x;
    y = m->y;
  } else {
    getMouse_xy(XtDisplay(m->w), &x, &y);
    x += m->pos.x; 
    y += m->pos.y;
  }

  width = m->menuWidth + m->sb_width;
  height = m->menuHeight;
  
  if (m->flags & POPUP) 
    height += m->titleHeight + m->sepHeight;

  if (m->flags & MULTI_SEL)
    height += m->doneHeight + m->sepHeight;

  exWidgetPos(m->w, x, y, width, height);

  XtPopup(m->w, XtGrabExclusive);
  
  for (m->mapped = True; m->mapped; ) {
    XtNextEvent(&e);
    XtDispatchEvent(&e);
  }
  
  if (m->w->core.being_destroyed) 	/* this menu has been destroyed */
    return;

  exPopdown(m->w);			/* unmap menu */

  setMenuActive(m, False);		/* reset active flag */
  m->flags &= ~MENU_MAPPED;
  
  /* discard all pending events for the menu just popped down */
  XFlush(XtDisplay(m->w));
  while (XCheckWindowEvent(display, XtWindow(m->pane), -1, &e));
}

/****************************************************************************
 *
 *		PRIVATE AUXILIARY FUNCTIONS 
 *
 ***************************************************************************/

/* requestFromUser **********************************************************
 * handle all requests of an item and build the return value of an item by 
 * substituting all '$' in the return pattern; return this string
 ****************************************************************************/

static Boolean requestFromUser(m, indx)
     exMenu *m;
     int indx;
{
  register MHiReq *ireq;
  register i, len;
  RequestParm rparm;
  register menuEntry *iptr = m->entries[indx];
  Position rx, ry;

  if (m->flags & MH_NOEXEC)		/* do not execute, just display menu */
    return(False);

  setMenuActive(m, True);

  if (iptr->reqCnt < 1)			/* no requests specified */
    return(True);
  
  for (i = 0; i < iptr->reqCnt; i++) {	/* process all requests */
    ireq = iptr->requests[i];
    if (ireq->flags & MH_USERFUNCTION)  /* postpone execution ... */
      if (! (m->flags & PANEL)) 	/* ... if a regular menu */
	continue;			/* ... see assembleReturnValues */
    
    if (! ireq->reqFnct) 		/* data structure error */
      return(False);
    
    XtTranslateCoords(m->pane, 
		      iptr->xpos + m->menuWidth, 
		      iptr->ypos + iptr->height, 
		      &rx, &ry);

    rparm.menu = (caddr_t) m;
    rparm.parent = m->w;
    rparm.data = ireq->reqData;
    rparm.ptr = ireq->reqPtr;
    rparm.x = rx;
    rparm.y = ry;
    rparm.size = ireq->size;
    rparm.cnt = ireq->cnt;
    rparm.cntType = ireq->reqCntType;
    
    rparm.buffer = ireq->initialValue;

    len = callRequestFunction(m, ireq->reqFnct, &rparm);

    if (len < 0)  			/* user abort */
      return(False);
    
    ireq->returnValue = rparm.buffer;
    
    /* insert the returned value as new initial value if required */
    if (iptr->flags & MH_INITVALUE) {
      free(ireq->prelimInitialValue);
      ireq->prelimInitialValue = strCopy(ireq->returnValue);
    }
  }
  
  return(True);
}

/* assembleReturnValues ******************************************************
 * is called if a menu-selection is finished. It loops through all selected 
 * items and collects the corresponding returnValues. These values are 
 * concatenated and stored into the menu-component 'assembledReturnValue' 
 * for later access by the function 'exGetMenuReturnValue'. If a request is 
 * defined as user-function, the corresponding function is executed here.
 ****************************************************************************/

static void assembleReturnValues(m)
     exMenu *m;
{
  register i, j, len, cnt;
  register char *ret;
  MHiReq *ireq;
  menuEntry *iptr;
  char *returnBuffer;
  char *buffer[MDS_MAXREQ];
  RequestParm rparm;

  if (m->flags & PANEL) 	/* don't collect return values */
    return;

  if (m->flags & MH_NOEXEC)	/* in simulation mode */
    return;

  if (m->selItemCnt < 1)
    return;

  cnt = m->itemCnt;

  for (i = 0; i < cnt; i++) {			/* for all items */
    if (m->selections[i]) {			/* if selected .. */
      iptr = m->entries[i];
      for (j = 0; j < iptr->reqCnt; j++) {	/* check requests */
	ireq = iptr->requests[j];
	if (ireq->flags & MH_USERFUNCTION) {	/* execute it now */
	  if (ireq->initialValue && *(ireq->initialValue))
	    buffer[j] = strCopy(ireq->initialValue);
	  else {                                /* call function */
	    buffer[j] = (char *) 0;
	    if (iptr->flags & MH_TOGGLEITEM) {
	      if (iptr->flags & MH_TICKON) {
		iptr->flags &= ~MH_TICKON;
		rparm.data = "0";
	      } else {
		iptr->flags |= MH_TICKON;
		rparm.data = "1";
	      }
	    } else 
	      rparm.data = ireq->reqData;

	    rparm.menu = (caddr_t) m;
	    rparm.parent = XtParent(m->w);
	    rparm.x = m->x + iptr->xpos;
	    rparm.y = m->y;
	    rparm.buffer = (char *) 0;
	    rparm.size = ireq->size;
	    rparm.cnt = ireq->cnt;
	    rparm.cntType = ireq->reqCntType;
	    rparm.ptr = ireq->reqPtr;

	    if (ireq->reqFnct) {
/*	      exGrabMenu(m, exMenuClockCursor); */
	      if (callRequestFunction(m, ireq->reqFnct, &rparm) >= 0) {
		if (rparm.buffer) 
		  buffer[j] = rparm.buffer;
		else 
		  buffer[j] = strCopy("");
	      } else 
		buffer[j] = strCopy("");
	    }
	  }
	} else 					/* not a user function */
	  buffer[j] = ireq->returnValue;	/* retrieve orig. value */
      } /* end loop: for (j = 0; ... */

      insertReturnValues(iptr, iptr->i_return, buffer, &returnBuffer);
    
      for (j = 0; j < iptr->reqCnt; j++) {	/* free buffers */
	free(buffer[j]);
	iptr->requests[j]->returnValue = (char *) 0;
      }

      if (returnBuffer) {				
	if (m->returnValues[i])
	  free(m->returnValues[i]);		/* free old ret. value */
	m->returnValues[i] = returnBuffer;	/* copy new ret. value */
      }
    } 

    for (j = 0; j < m->entries[i]->reqCnt; j++)
      setInitialValue(m, i, j);

      if (m->selections[i] && (m->erase == False))
	setDefaultItem(m->mds_structure->i_table[i]);
      else
	resetDefaultItem(m->mds_structure->i_table[i]);
  } /* end loop: for (i = 0; ... */

  for (i = 0, len = 1; i < cnt; i++)
    if (m->selections[i] && m->returnValues[i])
      len += strlen(m->returnValues[i]);
  
  ret = Malloc(len);
  *ret = EOS;
  
  for (i = 0; i < cnt; i++) {
    if (m->selections[i] && m->returnValues[i])
      strcat(ret, m->returnValues[i]);		/* append ret. value */
  }
  
  m->assembledReturnValue = ret;
}

/* selectNestedItem: ********************************************************
 * this function is called if the user selects a nested item outside the 
 * arrow region. All requests are checked whether an initial value exists. 
 * If so they are combined and stored into the returnvalue of the 
 * corresponding item; otherwise all requests are processed.
 ****************************************************************************/
 
static Boolean selectNestedItem(m, indx)
     exMenu *m;
     int indx;
{
  register menuEntry *iptr = m->entries[indx];
  register MHiReq *ireq;
  register i;
  RequestParm rparm;
  Position rx, ry;

  if (m->flags & MH_NOEXEC)		/* do not execute, just display menu */
    return(False);
  
  setMenuActive(m, True);		/* do not hide menu */

  for (i = 0; i <  iptr->reqCnt; i++) {	/* check each request */
    ireq = iptr->requests[i];
    if (ireq->flags & MH_USERFUNCTION)   /* execution postponed ... */
      if (! (m->flags & PANEL)) 	/* .... if a regular menu */
	continue;			 /* .. see assembleReturnValues */
      
    
    if (ireq->initialValue && *(ireq->initialValue))
      ireq->returnValue = strCopy(ireq->initialValue);
    else {                           	 /* request function is called */
      XtTranslateCoords(m->pane, 
			iptr->xpos + m->menuWidth, 
			iptr->ypos + iptr->height, 
			&rx, &ry);

      rparm.menu = (caddr_t) m;
      rparm.parent = XtParent(m->w);
      rparm.ptr = ireq->reqPtr;
      rparm.data = ireq->reqData;
      rparm.x = rx;
      rparm.y = ry;
      rparm.buffer = (char *) 0;
      rparm.size = ireq->size;
      rparm.cnt = ireq->cnt;
      rparm.cntType = ireq->reqCntType;

      if (ireq->reqFnct) {			/* request-function defined */
	if (callRequestFunction(m, ireq->reqFnct, &rparm) >= 0) 
	  ireq->returnValue = rparm.buffer;
	else {					/* error in request-function */
	  if (ireq->returnValue) {
	    free(ireq->returnValue);
	    ireq->returnValue = (char *) 0;
	  }
	  return(False);			/* user abort */
	}
      }
    }
  }

  return(True);
}

/* clearMenuDescr **********************************************************
 * sets input menu to a state where no item is selected and no return value
 * is set
 **************************************************************************/

static void clearMenuDescr(m)
     exMenu *m;
{
  int i;

  if (! m)
    return;

  for (i = 0; i < MDS_MAXITEM; i++) {
    m->selections[i] = 0;
    if (m->returnValues[i])
      free(m->returnValues[i]);
    m->returnValues[i] = NULL;
  }

  m->selItemCnt = 0;    /* ? */
  AM_CPY(&m->initialMask, &m->activityMask);
}

/* eraseMenu **************************************************************
 * erases menu completly (erase = True) or preserves return values of items
 * (erase = False); default items and return values are reset 
 *************************************************************************/

static void eraseMenu(m, erase)
     exMenu *m;
     Boolean erase;
{
  register i;

  if (! m)
    return;
  
  m->topItem = 0;				/* reset to item */
  
  m->selItemCnt = 0;  
  
  for (i = 0; i < MDS_MAXITEM; i++) 
    m->selections[i] = 0;
  
  if (erase == True) {  
    for (i = 0; i < MDS_MAXITEM; i++) {
      if (m->returnValues[i])
	free(m->returnValues[i]);
      m->returnValues[i] = NULL;
    }
    AM_CPY(&m->initialMask, &m->activityMask);
  }
  
  exMakeDefaultMenu(m);                      	/* set default items */   
}

/* helpUser: ******************************************************************
 * display the help information provided with the current selection
 *****************************************************************************/

static void helpUser(m, indx)
     exMenu *m;
     int indx;
{
  RequestParm parms;
  
  bzero(&parms, sizeof(RequestParm));

  parms.menu = (caddr_t) m;
  parms.parent = m->w;
  parms.x = -1;
  parms.y = -1;

  exUngrabAll(grabbedWidget);

  if (indx == INTITLE) {
    parms.data = m->hlpMsg;
    (* (m->hlpFnct))(&parms);
  } else if (indx >= 0 && indx < m->itemDisplayCnt) {
    parms.data = m->entries[indx]->hlpMsg;
    (* (m->entries[indx]->hlpFnct))(&parms);
  }

  exGrabAll(grabbedWidget, exMenuCursor);
}

/* exMakeDefaultItem *******************************************************
 * if all requests of input item have initial values or are user defined 
 * functions the return value is build and the default selection is set
 **************************************************************************/

static Boolean exMakeDefaultItem(m, indx)
     exMenu *m;
     int indx;	/* indx of default item in entry-table of menu m */
{
  char *buffer[MDS_MAXREQ];
  char *returnbuffer;
  int i;
  
  if (!m || indx > m->itemCnt)
    return(False);
  
  for (i = 0; i < m->entries[indx]->reqCnt; i++) {
    if (! m->entries[indx]->requests[i]->initialValue) {
      if (m->entries[indx]->requests[i]->flags & MH_USERFUNCTION) {
	buffer[i] = (char *) 0;
	continue;
      }
      else              /* no initial value for a request of default item */
	return(False);
    } 
    else      
      buffer[i] = m->entries[indx]->requests[i]->initialValue;
  }
  
  insertReturnValues(m->entries[indx], 
		     m->entries[indx]->i_return, buffer, &returnbuffer);
  
  free(m->returnValues[indx]);
  m->returnValues[indx] = returnbuffer;
  m->selections[indx] = 1;
  m->selItemCnt++;
  
  return(True);
}

/* nestedItem: ****************************************************************
 * check whether the give item is nested, that is requests are defined. A 
 * user defined function is not considered a request in this situation 
 *****************************************************************************/

static Boolean nestedItem(iptr)
  register menuEntry *iptr;
{
  register i;

  if (iptr->reqCnt <= 0)
    return(False);

  for (i = 0; i < iptr->reqCnt; i++)
    if (! (iptr->requests[i]->flags & MH_USERFUNCTION))
      return(True);

  return(False);
}

/* exMakeDefaultMenu *******************************************************
 * look up MDS-structure of input menu in order to update or reset default
 * items for input menu;                                   (interface.c ?)
 **************************************************************************/

void exMakeDefaultMenu(m)
     exMenu *m;
{
  register i;
  register MHmenu *mptr = m->mds_structure;

  for (i = 0; i < mptr->itemCnt; i++) {
    if (mptr->i_table[i]->flag & MH_DEFAULTITEM)
      exMakeDefaultItem(m, i);
  }
  checkMenuActivityMask(m, False);
}

/* checkDefaultMenu *********************************************************
 * 
 ****************************************************************************/

static Boolean checkDefaultMenu(mptr)
     register MHmenu *mptr;
{
  register i;
  
  for (i = 0; i < mptr->itemCnt; i++)
    if (mptr->i_table[i]->flag & MH_DEFAULTITEM)
      return(True);

  return(False);
}

/* checkactMask **************************************************************
 * checks whether bit 'icnt' is set in activityMask of input item
 *****************************************************************************/

static Boolean checkactMask(iptr, icnt)
     menuEntry *iptr;
     int icnt;
{
  if (! iptr)
    return(False);

  return((AM_ISSET(&iptr->activityMask, icnt)) ? True : False);
}

/* checkMenuActivityMask ****************************************************
 * builds activity mask of input menu by composing activity masks of 
 * selected entries
 ****************************************************************************/

void checkMenuActivityMask(m, display)
     exMenu *m;
     Boolean display;        /* if True, adjust displayed menu items */
{
  register i, j;
  act_mask oldmask;

  AM_CPY(&m->activityMask, &oldmask);

  if (m->selItemCnt == 0)  /* no item selected, use initial activity mask  */ 
    AM_CPY(&m->initialMask, &m->activityMask); 
  else if (!(m->flags & MULTI_SEL)) {	/* single selection menu */
    for (i = 0; i < m->itemCnt; i++)
      if (m->selections[i] == 1) {    /* only if default selection is set */
 	AM_CPY(&m->entries[i]->activityMask, &m->activityMask); 
	break;
      }
  } else if (m->selItemCnt >= 1) {         
    /* at least one item is selected in a multi-selection menu, i.e. the 
     * activity masks of other selectable items have to be checked if they 
     * exclude the selection of the already selected item; these have to be 
     * enabled additionally; if several items are selected, these masks have 
     * to be combined */
 
    AM_ONES(&m->activityMask);
    for (i = 0; i < m->itemCnt; i++)
      if (m->selections[i]) {
	for (j = 0; j < sizeof(act_mask) / sizeof(long); j++)
	  m->activityMask.actmask_bits[j] 
	    		&= m->entries[i]->activityMask.actmask_bits[j];
	
	for (j = 0; j < m->itemCnt; j++) 
	  if (checkactMask(m->entries[j], i) == False)
	    AM_CLR(&m->activityMask, j);
	                      /* if selection of item j excludes selection */
                              /* of item i selection of j is forbidden for */
                              /* item i is already selected */ 
 
	if (m->selItemCnt == 1)
	  break;
      }
  }

  if (! bcmp(&m->activityMask, &oldmask, sizeof(act_mask)))
    return;

  if (display) {
    for (i = 0; i < m->itemCnt; i++)
      if (AM_ISSET(&oldmask, i) != AM_ISSET(&m->activityMask, i))
	displayItem(m, i, False);   /* display items which are affected by 
				     * new activity mask */
  }
}

/* setInitialValue **********************************************************
 * assign new initial value to request
 ****************************************************************************/

static setInitialValue(m, itemIndx, reqIndx)
     exMenu *m;
     int itemIndx, reqIndx;
{
  MHmenu *mptr;

  if (!m || itemIndx > m->itemCnt || !m->entries[itemIndx]->requests[reqIndx]
      || !m->entries[itemIndx]->requests[reqIndx]->prelimInitialValue)
    return(-1);
  
  if (m->entries[itemIndx]->requests[reqIndx]->initialValue)
    free(m->entries[itemIndx]->requests[reqIndx]->initialValue);

  m->entries[itemIndx]->requests[reqIndx]->initialValue  = 
   strCopy(m->entries[itemIndx]->requests[reqIndx]->prelimInitialValue);

  if ((mptr = m->mds_structure) == (MHmenu *) 0)
    return(-1);

  setRequestInitialValue(mptr->i_table[itemIndx]->i_requests[reqIndx], 
	m->entries[itemIndx]->requests[reqIndx]->initialValue);		 

  return(0);
}

/* freeMenu: ***************************************************************
 * free the menu structure
 ***************************************************************************/

void freeMenu(m)
     register exMenu *m;
{
  register i, j;
  register menuEntry *iptr;

  if (! m)
    return;

  /* not needed if we destroy it anyway */

  if (! m->w->core.being_destroyed) {	
    Arg args[2];
    int cnt = 0;
    XtSetArg(args[cnt], XmNuserData, NULL); cnt++;
    XtSetValues(m->pane, args, cnt);
  }

  free(m->name);

  for (i = 0; i < m->itemCnt; i++) {
    if (m->returnValues[i])
      free(m->returnValues[i]);

    if (m->entries[i]) {		/* free entry */
      iptr = m->entries[i];

      free(iptr->name);

      if (iptr->i_return)
	free(iptr->i_return);

      if (iptr->hlpMsg)
	free(iptr->hlpMsg);

      for (j = 0; j < MDS_MAXREQ; j++)	/* free requests */
	if (iptr->requests[j]) {
	  if (iptr->requests[j]->reqData)
	    free(iptr->requests[j]->reqData);
	  if (iptr->requests[j]->initialValue)
	    free(iptr->requests[j]->initialValue);
	  if (iptr->requests[j]->prelimInitialValue)
	    free(iptr->requests[j]->prelimInitialValue);
	  if (iptr->requests[j])
	    free(iptr->requests[j]);
	}
    }
  }

  if (m->assembledReturnValue)
    free(m->assembledReturnValue);

  free(m->hlpMsg);
      
  free(m);
}

/* exDisplayMenu: ***********************************************************
 * display the specified area of the menu. 
 ***************************************************************************/

static void exDisplayMenu(m, x, y, w, h)
     exMenu *m;
{
  int start_col, end_col, start_row, end_row, i, j, indx;

  if (m->flags & SCROLLABLE) {
    exSetScrollbar(m);
    m->slider_pos = -1;
  }
		
  if (m->flags & POPUP)
    displayTitle(m, True);

  /* the whole menu needs redrawing */
  if (x == 0 && y == 0 && w >= m->menuWidth && h >= m->menuHeight) {
    redrawMenu(m);
    return;
  }

  /* redraw only the new exposed part of the menu */
  if (w <= 1 || h <= 1)
    return;

  /* find items that are affected by exposed region */
  /* find starting col */
  if ((start_col = getCol(m, x)) < 0)
    start_col = 0;

  /* find end col */
  if ((end_col = getCol(m, x + w)) < 0)
    end_col = m->cols;

  if (m->flags & POPUP && 
      y < 0 && y > -m->titleHeight + m->sepHeight) /* check if in title */
    start_row = m->topItem;
  else {
    if ((start_row = getRow(m, y)) < 0)
      start_row = m->topItem;
  }

  if (m->flags & MULTI_SEL &&	      		/* check if in done box */
      	!(m->flags & PANEL) && y > -m->yb && y < -m->yb + m->titleFontHeight) {
    drawDoneBox(m);
    end_row = m->topItem + m->itemDisplayCnt;
  } else {
    if ((end_row = getRow(m, y + h)) < 0)
      end_row = m->topItem + m->itemDisplayCnt;

  }

  /* display selected items */
  for (i = start_col; i <= end_col; i++)
    for (j = m->topItem + start_row; j <= m->topItem + end_row; j++) {
      indx = i * m->itemsPerCol + j;
      if (indx >= 0 && indx < m->itemCnt)
	displayItem(m, indx, m->selections[indx]);
    }
}

/* redrawMenu: **************************************************************
 * redraw the whole menu, including shadow and border
 ****************************************************************************/

static void redrawMenu(m)
     exMenu *m;
{
  register cnt, i;
  Dimension w, h;
  Arg args[3];

  XClearWindow(display, XtWindow(m->pane));
   
  if (! (m->flags &  PANEL)) {
    cnt = 0;
    XtSetArg(args[cnt], XmNheight, &h); cnt++;
    XtSetArg(args[cnt], XmNwidth, &w); cnt++;
    XtGetValues(m->shape, args, cnt);

    exDrawShadow(m->shape, 0, 0, w - SHADOW_WIDTH, h - SHADOW_WIDTH);
  }

  if (m->flags & POPUP)
    displayTitle(m, True);

  for (i = 0; i < m->itemCnt; i++) 
    displayItem (m, i, (Boolean) m->selections[i]);
  
  if (m->flags & MULTI_SEL && !(m->flags & PANEL))
    drawDoneBox(m);
}

/* displayTitle: ************************************************************
 * display the title of a popup-menu 
 ***************************************************************************/

static void displayTitle(m, sel)
     exMenu *m;
     Boolean sel;
{
  int cnt;
  Arg args[2];
  Pixmap pm;

  cnt = 0;
  XtSetArg(args[cnt], XmNforeground, &pm); cnt++;
  XtGetValues(m->title, args, cnt);

  if (sel && pm == exWhitePixel(display))	/* already inverted */
    return;

  if (! sel && pm == exBlackPixel(display))	/* not yet inverted */
    return;

  cnt = 0;
  XtSetArg(args[cnt], XmNforeground, 
	   sel ? exWhitePixel(display) : exBlackPixel(display)); cnt++;
  XtSetArg(args[cnt], XmNbackground, 
	   sel ? exBlackPixel(display) : exWhitePixel(display)); cnt++;

  XtSetValues(m->title, args, cnt);
}
  
/* displayItem: *************************************************************
 * display the item specified by slot (slot is an indx into the menu 
 * itemtable) if sel is set the item is highlited. 
 ****************************************************************************/

void displayItem(m, slot, sel)
     exMenu *m;
     int slot; 	       	/* slot in item-table */
     Boolean sel;	/* True if item is selected */
{
  menuEntry *iptr;
  int y, indx, width;
  GC draw_gc, font_gc;
  Widget mw = m->pane;
  Window w = XtWindow(mw);
  Boolean disabled = False;

  if (slot < 0 || slot >= m->itemCnt)  	/* ill. slot into item table */
    return;

  if (! w)				/* has not been realized yet */
    return;

  indx = slot - m->topItem;

  if (! displayable(m, slot))
    return;

  iptr = m->entries[slot];

  /* calculate y-pos of item */
  y = iptr->ypos - m->entries[slot - indx]->ypos + MH_BORDER;

  if (m->flags & POPUP)		     	/* consider title */
    y += m->entries[0]->ypos - MH_BORDER;

  if (! (AM_ISSET(&m->activityMask, slot))) {		/* item disabled */
    disabled = True;
    draw_gc = disable_gc;
    font_gc = normal_gc;
  } else if (m->selections[slot] || sel) { 		/* item selected */
    font_gc = select_gc;
    draw_gc = normal_gc;
  } else { 						/* item not selected */
    draw_gc = select_gc;
    font_gc = normal_gc;
 } 

  XSetFont(display, font_gc, m->itemFont->fid);
  XSetFont(display, draw_gc, m->itemFont->fid);
  XSetFunction(display, draw_gc, GXcopy);

  width = (m->flags & ARROW_OR_TICK) ? 
    	iptr->width - Arrow_width - ITEM_GAP: iptr->width;

  drawItem(w, iptr, iptr->xpos, y, width, m->selections[slot] || sel, 
	   draw_gc, font_gc);

  y = iptr->ppos;

  if (iptr->flags & MH_TOGGLEITEM && iptr->flags & MH_TICKON)
    drawTick(w, iptr, y, iptr->width, m->selections[slot] || sel, disabled, 
	     draw_gc, font_gc);
  else if (nestedItem(iptr))  
    drawArrow(w, iptr, y, iptr->width, (m->selections[slot] || sel) && 
	      (iptr->flags & ARROW_REGION || iptr->flags & ARROW_SELECTED), 
	      disabled, draw_gc, font_gc);
}

/* drawItem *****************************************************************
 * draw the item at the appropriate position with background, foreground and 
 * font specified according to the items state (selected, enabled, disabled)
 ****************************************************************************/
     
static drawItem(w, iptr, lm, y, width, sel, draw_gc, font_gc)
     Window w;
     menuEntry *iptr;
     int lm;				/* start of item label area */
     int y;				/* y-pos of item label area */		
     int width;				/* width of item label area */
     int sel;				/* item selected */
     GC draw_gc, font_gc;
{
  static char key[3] = {'^',' ','\0'};
  int fy;

  XFillRectangle(display, w, draw_gc, lm, y,	width, iptr->height);
  
  fy = y + iptr->baseLine;

  if (iptr->key) {
    key[1] = iptr->key;
    XDrawString(display, w, font_gc, iptr->xpos, fy, key, 2);
  }

  XDrawString(display, w, font_gc, iptr->xpos + iptr->labelOffset, fy, 
	      				iptr->name, strlen(iptr->name));
  
  if (sel) 					/* draw tiny border */
    XDrawRectangle(display, w, font_gc, lm + 1, 
		   		y + 1, width - 3, iptr->height - 3);
}

/* drawArrow: **************************************************************
 * display the arrow with respect to the given item
 ***************************************************************************/

static void drawArrow(w, iptr, y, width, sel, disabled, draw_gc, font_gc)
     Window w;
     menuEntry *iptr;
     int y;
     int width;
     Boolean sel, disabled;
     GC draw_gc, font_gc;
{
  register arrow_x = iptr->xpos + width - Arrow_width - 2;

  if (sel) {
    XFillRectangle(display, w, draw_gc, arrow_x, y, 
		   		Arrow_width + 4, Arrow_height);
    XSetFunction(display, font_gc, GXxor);
  } else
    XClearArea(display, w, arrow_x, y, Arrow_width + 4, Arrow_height, False);

  XCopyArea(display, disabled ? disabledArrow : enabledArrow, w, font_gc,
	    0, 0, Arrow_width, Arrow_height, arrow_x + 2, y);

  if (sel) {					/* draw tiny border */
    XSetFunction(display, font_gc, GXcopy);
    XDrawRectangle(display, w, font_gc, arrow_x + 1, y + 1, 
		   Arrow_width + 1, Arrow_height - 3);
  }
}

/* drawTick: *****************************************************************
 * in the case of a toggle item, draw the appropriate tick 
 ****************************************************************************/

static drawTick(w, iptr, y, width, sel, disabled, draw_gc, font_gc)
     Window w;
     menuEntry *iptr;
     int y;
     int sel;
     Boolean disabled;
     GC draw_gc, font_gc;
{

  register arrow_x = iptr->xpos + width - Arrow_width;

  XFillRectangle(display, w, draw_gc, arrow_x, y, tick_width + 4, tick_height);

  if (sel)
    XSetFunction(display, font_gc, GXxor);

  XCopyArea(display, (disabled) ? disabledTick : enabledTick, w, font_gc, 
	    0, 0, tick_width, tick_height, arrow_x + 2, y);

  if (sel) {					/* draw tiny border */
    XSetFunction(display, font_gc, GXcopy);
    XDrawRectangle(display, w, font_gc, arrow_x + 1, y + 1, 
		   tick_width + 1, tick_height - 3);
  }
}

/* findItem: *****************************************************************
 * returns the index of the item corresponding to the x and y coordinates;
 * in the case of the title, -mptr->itemCnt is returned in the case of the 
 * done-box, +mptr->itemCnt is returned; the index returned is relative to the 
 * item table;
 *****************************************************************************/

static findItem(m, x, y, rx, ry)
     exMenu *m;
     int x, y, rx, ry;
{
  register i, v, row, indx;
  register offset;	/* offset of the first item displayed with 
			 * respect to the first item in the item table */
  int startItem;
  menuEntry *eptr;

  /* check if in title */
  if (m->flags & POPUP && 
                x > 0 - m->sb_width && x < m->menuWidth && 
      		y < 0 && y > -m->titleHeight + m->sepHeight)
    return(INTITLE);	
  else if (m->flags & PULLDOWN && (findButton(m, rx, ry) == m->button))
    return(INTITLE);   

  if (m->itemCnt < 1)				/* no items */
    return(OUTSIDE);

  if (m->flags & MULTI_SEL && ! (m->flags & PANEL)) {
    if (x > MH_MARGIN && x < MH_MARGIN + m->titleFontHeight &&
	y > -m->yb && y < -m->yb + m->titleFontHeight)
      return(INDONEBOX);			/* in done box */
  }

  if (x < -m->sb_width || x > m->menuWidth 
      	|| y < -m->titleHeight || y > m->menuHeight + m->doneHeight)
    return(OUTSIDE);

  if (m->flags & SCROLLABLE && x < 0 && x > -m->sb_width && 
      y > 0 && y < m->menuHeight)
    return(SB_INSIDE);				/* in scrollbar */

  offset = m->entries[0]->ypos - m->entries[m->topItem]->ypos;

  v = (m->itemCnt > 1) ? verticalDist(m, 0, 1) : 0;

  startItem = m->topItem + ((v) ? (y - m->titleHeight) / v : 0);
 
  if (startItem >= MIN(m->itemCnt, m->topItem + m->itemDisplayCnt))
    startItem = MIN(m->itemCnt, m->topItem + m->itemDisplayCnt) - 1;
  
  if (startItem < 0)
    startItem = 0;

  row = -1;

  /* check 1st col only for y-position */
  for (i = startItem; i < MIN(m->topItem + m->itemDisplayCnt,m->itemCnt); i++) 
    if (y >= m->entries[i]->ypos + offset && 
	y <= m->entries[i]->ypos + offset + 
	         m->entries[i]->height) {
      row = i;
      break;
    }

  if (row < 0) {
    for (i = startItem - 1; i >= m->topItem; i--)
      if (y >= m->entries[i]->ypos + offset && 
	  y <= m->entries[i]->ypos + offset + 
	  m->entries[i]->height) {
	row = i;
	break;
      }
  }

  if (row < 0 || row > m->itemsPerCol)			/* invalid row */
    return(-1);

  /* check cols now */
  for (i = 0; i < m->cols; i++) {
    indx = i * m->itemsPerCol + row;
    if (indx >= m->itemCnt) 
      break;
    eptr = m->entries[indx];
    if (x > eptr->xpos && x < eptr->xpos + eptr->width) 
      return(indx);
  }

  return(-1);
}

/************************************* verticalDist *************************
 * calculates the height of an area between two items; the height is from
 * the top-edge of the first item to the top-edge of the second item
 ***************************************************************************/

static verticalDist(m, item1, item2)
     exMenu *m;
     int item1, item2;
{
  if (!m->entries[item1] || !m->entries[item2])
    return(0);

  return(m->entries[item2]->ypos - m->entries[item1]->ypos);
}

/* selectDoneBox: ***********************************************************
 * select (fill box) or deselect (clear box) the done-box
 **************************************************************************/

static void selectDoneBox(m, set)
     exMenu *m;
     Boolean set;
{
  Window w = XtWindow(m->done_box);

  if (set) {
    XCopyArea(display, closePix, w, normal_gc, 2, 2, 
	      m->titleFontHeight - 4, m->titleFontHeight - 4, 
	      MH_MARGIN + 2, MH_MARGIN + 2);
    exWait(10);
  } else 
    XCopyArea(display, exWhitePM(display), w, normal_gc, 2, 2, 
	      m->titleFontHeight - 4, m->titleFontHeight - 4,
	      MH_MARGIN + 2, MH_MARGIN + 2);
}
  
/* drawDoneBox: ************************************************************
 * display the done box
 ***************************************************************************/

static void drawDoneBox(m)
     exMenu *m;
{
  register ct;
  char msg[20];
  Window w;

  if (!m->mapped)
    return;

  w = XtWindow(m->done_box);

  ct = (int)m->selType - (int)MDS_CNT_NULL - 1;

  if (ct > 0)
    sprintf(msg,"done (%s%d)", countType[ct][1], m->selCnt);
  else
    sprintf(msg,"done (%d)", m->selCnt);

  XClearWindow(display, w);

  XDrawRectangle(display, w, normal_gc, MH_MARGIN, MH_MARGIN, 
		 		m->titleFontHeight, m->titleFontHeight);
    
  XSetFont(display, normal_gc, m->itemFont->fid);
  XDrawString(display, w, normal_gc, 
	      MH_MARGIN + m->titleFontHeight + MH_MARGIN, 
	      MH_MARGIN + m->baseLine,  msg, strlen(msg));
}

/* pullMenuDown: ************************************************************
 * the coordinates are calculated where the menu is to appear. Then it is
 * moved and mapped and execution of the menu is started. 
 ****************************************************************************/

static exMenu *pullMenuDown(button, execute)
     Widget button;
     Boolean execute;
{
  exMenu *m = (exMenu *) 0;
  Arg args[5];
  int cnt = 0;

  Position x = 0, y = 0;
  Dimension h = 0;
  MDS_POS_TYPE pt;

  XtSetArg(args[cnt], XmNx, (XtArgVal) &x); cnt++;
  XtSetArg(args[cnt], XmNy, (XtArgVal) &y); cnt++;
  XtSetArg(args[cnt], XmNheight, (XtArgVal) &h); cnt++;
  XtSetArg(args[cnt], XmNuserData, (XtArgVal) &m); cnt++;

  XtGetValues(button, args, cnt);
  
  if (! m)
    return;

  XtTranslateCoords(XtParent(m->button), x, y, &m->x, &m->y);

  m->y = m->y + h + MH_PANEGAP;
  pt = m->pos.posType;			/* save pos type */
  m->pos.posType = MDS_POS_ABS;		/* temporarily set abs. pos */

  if (! execute)			/* do not really execute menu */
    m->flags |= MH_NOEXEC;

  postMenu(m);				/* start menu-processing */

  if (! execute)			/* reset flag, if necessary */
    m->flags &= ~MH_NOEXEC;
  m->pos.posType = pt;			/* restore pos type */

  return(m);
}

/* exGrabMenu: **************************************************************
 * the pointer is grabbed with respect to the menu's main window pane and 
 * a flag is set to indicate that the menu is mapped so as to avoid multiple 
 * grabs to occur.
 ***************************************************************************/

void exGrabMenu(m, csr)
     exMenu *m;
     Cursor csr;
{
  if (m->flags & PANEL)
    return;

  exUngrabAll(grabbedWidget);

  exGrabAll(m->pane, (csr == None) ? exMenuCursor : csr);

  grabbedWidget = m->pane;

  m->flags |= MENU_MAPPED;
}

/* findButton: *************************************************************
 * checks if the pointer is within the borders of a button and if so returns 
 * the button identifier 
 ***************************************************************************/

static Widget findButton(m, x, y)
     exMenu *m;
     int x, y;			/* root coordinates */
{
  exMenubar *mb;
  int i, rx, ry;
  Window win;
  Widget button;

  if ( !(mb = m->menubar))		/* a native popup menu */
    return((Widget) 0);

  /* get the window that containes the mouse */
  for (i = 0; i < mb->num_buttons; i++) {
    XTranslateCoordinates(display, DefaultRootWindow(display), 
			  XtWindow(XtParent(mb->buttons[i].button)),
			  x, y, &rx, &ry, &win);
    if (win)				/* found a button */
      break;
  }

  if (! win)				/* no button found */
    return((Widget) 0);

  button = XtWindowToWidget(display, win);
  
  for (i = 0; i < mb->num_buttons; i++) {
    if (button == mb->buttons[i].menu->button)
      return(button);
  }

  return((Widget) 0);
}

/*****************************************************************************
 *
 *     		           MENU ACTION ROUTINES 
 *
 ****************************************************************************/

/* exMenuEnterProc: *********************************************************
 * when the cursor enters a menu the title is selected (if a popup menu)
 ***************************************************************************/

/*ARGSUSED*/
static void exMenuEnterProc(w, event, params, num_params)
     Widget w;
     XEnterWindowEvent *event;
     String *params;
     Cardinal *num_params;
{
  exMenu *m = NULL;
  Arg args[2];

  XtSetArg(args[0], XmNuserData, &m);
  XtGetValues(w, args, 1);
  
  if (! m)
    return;

  if (m->flags & POPUP)
    displayTitle(m, True);
}

/* exMenuLeaveProc: *********************************************************
 * when the cursor leaves a menu the title is deselected (if a popup menu)
 * and the last item, is deselected.
 ***************************************************************************/

/*ARGSUSED*/
static void exMenuLeaveProc(w, event, params, num_params)
     Widget w;
     XLeaveWindowEvent *event;
     String *params;
     Cardinal *num_params;
{
  exMenu *m = (exMenu *) 0;
  Arg args[2];

  XtSetArg(args[0], XmNuserData, &m);
  XtGetValues(w, args, 1);

  if (! m)
    return;

  if (m->flags & POPUP)
    displayTitle(m, False);

  displayItem(m, m->prev, False);
  m->prev = -1;
}

/* exMenuExposeProc: *******************************************************
 * This callback is called, when the menu-pane window gets exposed. When the 
 * menu is mapped, the active grab has to be reset, so that we get all pointer 
 * events with respect to the menu-window. Otherwise, all pointer events go
 * to the previous window, until the mouse button is released. New active 
 * grabbing, however, is only necessary when the menu is newly mapped.
 ***************************************************************************/

/*ARGSUSED*/
static void exMenuExposeProc(w, event, params, num_params)
     Widget w;
     XExposeEvent *event;
     String *params;
     Cardinal *num_params;
{
  exMenu *m = (exMenu *) 0;
  Arg args[5];
  int cnt;
  static int xx = 0, yy = 0, ww = 0, hh = 0;
  Window wnd;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &m); cnt++;
  XtGetValues(w, args, cnt);

  if (! m)
    return;

  if (! (m->flags & MENU_MAPPED) && !(m->flags & PANEL))
    exGrabMenu(m,  exMenuCursor);

  /* collect the information of several consecutive expose events */
  xx = MIN(xx, event->x);
  yy = MIN(yy, event->y);
  ww = MAX(ww, event->x + event->width);
  hh = MAX(hh, event->y + event->height);
  
  if (event->count > 0)			/* collect more expose events */
    return;
  else {
    exDisplayMenu(m, xx, yy, ww, hh);	/* expose part of window */
    xx = yy = ww = hh = 0;
  }
  
  if (m->xs == 0 && m->flags & SCROLLABLE) 
    XTranslateCoordinates(display, XtWindow(m->pane), XtWindow(m->scrollbar),
			  0, 0, &m->ys, &m->ys, &wnd);

  /* calculate origin of done box relativ to item area */
  if (! (m->flags & PANEL) && m->yb == 0 && m->flags & MULTI_SEL) {
    XTranslateCoordinates(display, XtWindow(m->pane), XtWindow(m->done_box),
			  0, 0, &cnt, &m->yb, &wnd);
    m->yb -= MH_MARGIN;
  }
}

/* exMenuBtnDownProc: *******************************************************
 * callback that is called when the user presses a mouse button within the 
 * body of a menu. It is dicriminated between a mouse press inside the 
 * scrollbar (in which case the menu is scrolled) or if the right mouse 
 * button has been pressed (in which case the help window is displayed.
 ***************************************************************************/

/*ARGSUSED*/
static void exMenuBtnDownProc(w, event, params, num_params)
     Widget w;
     XButtonPressedEvent *event;
     String *params;
     Cardinal *num_params;
{
  exMenu *m;
  Arg args[2];
  int indx;

  XtSetArg(args[0], XmNuserData, &m);
  XtGetValues(w, args, 1);

  if (! m)
    return;

  indx = findItem(m, event->x, event->y, event->x_root, event->y_root);

  if (indx == SB_INSIDE) 			/* scrollbar event */ 
    exScrollbar(m, event);
  else if (event->button > Button2) 		/* user requested help */
    helpUser(m, indx);
}

/* exMenuBtnUpProc: ********************************************************* 
 * callback function that is called whenever the user releases the mouse
 * inside a menu-pane. This function decides which item is under the mouse 
 * cursor and decides which action to take with repect to the menu-type.
 ***************************************************************************/

/*ARGSUSED*/
static void exMenuBtnUpProc(w, event, params, num_params)
     Widget w;
     XButtonReleasedEvent *event;
     String *params;
     Cardinal *num_params;
{
  exMenu *m = (exMenu *) 0;
  Arg args[2];
  Boolean ReqResult;
  int indx;

  XtSetArg(args[0], XmNuserData, &m);
  XtGetValues(w, args, 1);

  if (! m)
    return;

  exScrollbar(m, event);

  indx = findItem(m, event->x, event->y, event->x_root, event->y_root);

  if (indx == OUTSIDE || indx == INTITLE) {	/* button released outside */
    if (m->assembledReturnValue)
      free(m->assembledReturnValue);
    m->assembledReturnValue = NULL;
    eraseMenu(m, m->erase);
    m->mapped = False;
    m->flags |= MH_ABORT;
    return;
  }

  if (! (m->flags & MULTI_SEL)) {
    if (indx < 0) {
      m->flags |= MH_ABORT;
      return;
    }

    if (! AM_ISSET(&m->activityMask, indx)) {	/* disabled */
      eraseMenu(m, m->erase);
      m->mapped = False;
      return;
    }

    clearMenuDescr(m);

    if (nestedItem(m->entries[indx])) { 	/* if nested, ... */
      if (inArrowRegion(m->entries[indx], event->x)) {
	m->entries[indx]->flags |= ARROW_SELECTED;
	ReqResult = requestFromUser(m, indx);
      } else {
	m->entries[indx]->flags &= ~ARROW_SELECTED;
	ReqResult = selectNestedItem(m, indx);
      }
    } else 					/* not nested */
      ReqResult = requestFromUser(m, indx);

    if (ReqResult) {				/* return value requested */
      m->selections[indx] = 1;
      m->selItemCnt = 1;
    }
    
    m->mapped = False;
    assembleReturnValues(m);
    eraseMenu(m, m->erase);
  } else { 					/* multiselection menu */
    if (indx >= 0) {				/* legal item */

      if (! AM_ISSET(&m->activityMask, indx)) 	/* item disabled */
	return;

      if (m->selections[indx]) {		/* already selected */
	if (nestedItem(m->entries[indx])        /* in arrow */
	                  && inArrowRegion(m->entries[indx], event->x)) {
	  (void) requestFromUser(m, indx);
	} else {				/* not nested, not in arrow */
	  if (m->selItemCnt > 0)
	    m->selItemCnt--;
	  m->selections[indx] = 0;
	  if (m->returnValues[indx])
	    free(m->returnValues[indx]);
	  m->returnValues[indx] = NULL;
	}
      } else {					/* not selected yet */
	if (nestedItem(m->entries[indx])) {	/* if nested, ... */
	  if (inArrowRegion(m->entries[indx], event->x)) {
	    m->entries[indx]->flags |= ARROW_SELECTED;
	    ReqResult = requestFromUser(m, indx);
	  } else {
	    m->entries[indx]->flags &= ~ARROW_SELECTED;
	    ReqResult = selectNestedItem(m, indx);
	  }
	} else 					/* not nested */
	  ReqResult = requestFromUser(m, indx);
	
	if (ReqResult) {			/* return value requested */
	  m->selections[indx] = 1;
	  m->selItemCnt++;
	}
      }
    } else if (indx == INDONEBOX) {
      displayItem(m, m->prev, False);
      m->prev = -1;
      selectDoneBox(m, True);			/* fill done-box */
      if (countCondition(m->selType,  m->selItemCnt, m->selCnt)) {	
	m->mapped = False;			/* unmap window */
	assembleReturnValues(m);
	eraseMenu(m, m->erase);
      } else { 
	exUngrabAll(grabbedWidget);
	grabbedWidget = NULL;
	announceIllegalMenuCount(m, exMenuCursor);/* leave not, inform user */
	
	exGrabMenu(m, exMenuCursor);
	selectDoneBox(m, False);		/* clear done-box */ 
      }
      return;
    }

    m->prev = -1;
    checkMenuActivityMask(m, True);
    displayItem(m, indx, (Boolean) m->selections[indx]);
  }						/* end multiselection */
}
 
/* exMenuMotionProc: *******************************************************
 * This callback is called, when the cursor moves with respect to a mapped
 * menu pane. As a result of the pointer beeing grabbed (see exMenuExposeProc)
 * this function is called even when the mouse-cursor is not within the
 * menu-pane. This allows to know when the cursor moves to another button in
 * the menu-bar or the done-button (if any).
 ***************************************************************************/

/*ARGSUSED*/
static void exMenuMotionProc(wd, event, params, num_params)
     Widget wd;
     XMotionEvent *event;
     String *params;
     Cardinal *num_params;
{
  exMenu *m = (exMenu *) 0;
  Arg args[2];
  int x, y, indx;
  Window w;
  Display *d;
  Widget button;

  XtSetArg(args[0], XmNuserData, &m);
  XtGetValues(wd, args, 1);
  x = event->x;
  y = event->y;

  if (! m)
    return;

  w = XtWindow(m->pane);
  d = event->display;

  /* check if the event has been generated by this routine (this happens
   * if the user selects to browse through different menus if already one 
   * multiselection menu is active on the screen). In this case we simply
   * pull the appropriate menu down (the menu determined by the button whose
   * window is specified in the event).
  */

  if (event->send_event == True) {	/* event generated here (see above) */
    button = findButton(m, event->x_root, event->y_root);
    if (! button || button == m->button)
      return;
    pullMenuDown(button, False);	/* display associated menu */
    exGrabMenu(m, exMenuCursor);	/* reset grab to active menu */
    return;
  }

  indx = findItem(m, x, y, event->x_root, event->y_root);

  if (indx == -1)
    return;

  if (indx == INTITLE) {		/* title */
    displayItem(m, m->prev, False);
    m->prev = -1;
    return;
  } 

  if (indx == INDONEBOX) {			/* in menu's done box */
    displayItem(m, m->prev, False);		/* deselect previous item */
    m->prev = -1;
    return;
  }

  if (indx == SB_INSIDE){ 			/* scrollbar event */ 
    displayItem(m, m->prev, False);		/* deselect previous item */
    m->prev = -1;
    exScrollbar(m, event);
    return;
  }

  if (indx == OUTSIDE) {			/* not in menu's item-list */ 
    displayItem(m, m->prev, False);		/* deselect previous item */
    m->prev = -1;
    button = findButton(m, event->x_root, event->y_root);

    if (! button || button == m->button)	/* not in any (new) button */
      return;
    
    if (event->state & Button1Mask) {		/* left mouse button pressed */
      if (thisMenuActive(m)) {			/* do not erase current menu */
	pullMenuDown(button, False);		/* only display menu */
	exGrabMenu(m, exMenuCursor);		/* grab mouse again */
      } else {
	XMotionEvent me;
	XCrossingEvent ce;
	exMenu *activeMenu;
	
	m->mapped = False;			/* make menu disappear */
	
	if (activeMenu = getActiveMenu(m)) {	/* any active menu displayed */
	  me.type = MotionNotify;
	  me.display = event->display;
	  me.window = XtWindow(activeMenu->pane);
	  me.root = event->root;
	  me.subwindow = NULL;
	  me.time = CurrentTime;
	  me.x = 0;
	  me.y = 0;
	  me.x_root = event->x_root;
	  me.y_root = event->y_root;
	  me.state = event->state;
	  me.is_hint = 0;
	  me.same_screen = event->same_screen;
	  
	  /* send MotionNotify event; if a menu is already displayed and 
	   * active and a second menu is current and displayed, this event is 
	   * handled in the context of the first menu's exMenuMotionProc (see
	   * code above). Execution results in unmapping the current menu 
	   * (m->mapped = False) and displaying the menu
	   * associated with the identified button.
	   */
	  
	  XSendEvent(d, PointerWindow, False, 0L, (XEvent *) &me);
	} else { 				/* no active menu */
	  ce.type = EnterNotify;
	  ce.display = event->display;
	  ce.window = XtWindow(button);
	  ce.root = event->root;
	  ce.subwindow = NULL;
	  ce.time = CurrentTime;
	  ce.x = 0;
	  ce.y = 0;
	  ce.x_root = event->x_root;
	  ce.y_root = event->y_root;
	  ce.mode = NotifyNormal;
	  ce.detail = NotifyNonlinear;
	  ce.same_screen = event->same_screen;
	  ce.focus = False;
	  ce.state = event->state;
	  
	  /* send EnterNotify event; if a menu is displayed but not active 
	   * (i.e. no item has been selected in a multiselection menu) this 
	   * event is sent to the identified button. As a result of the 
	   * specified translations, the function exButtonProc is called in 
	   * the context of this button which results in the menu being 
	   * displayed as if the user had pressed the left mouse button in 
	   * the associated menubar-button. The pointer grabbing enshures 
	   * that the event is delivered to the specified window (the button 
	   * window) 
	   */
	  
	  XSendEvent(d, PointerWindow, False, 0L, (XEvent *) &ce);
	  XGrabPointer(d, XtWindow(button), False, 
		ButtonPressMask | ButtonReleaseMask | PointerMotionMask,
		GrabModeAsync, GrabModeAsync, None, exMenuCursor, CurrentTime);
	  XFlush(d);
	}
      }
      return;
    } else					/* no mouse button pressed */
      return;
  }

  if (indx < 0)				/* illegal item */
    return;

  if (indx == m->prev) {			/* same item */
    if (! AM_ISSET(&m->activityMask, indx))	/* disabled */
      return;
    else if (! nestedItem(m->entries[indx]))	/* not a nested item */
      return;
    else {
      if (inArrowRegion(m->entries[indx], x)) {
	if (m->entries[indx]->flags & ARROW_REGION)
	  return;
	else {
	  m->entries[indx]->flags |= ARROW_REGION;
	  drawArrow(w, m->entries[indx], m->entries[indx]->ppos, 
		m->entries[indx]->width, True, False, normal_gc, select_gc);
	  return;
	}
      } else {					/* not in arrow region */
	if (m->entries[indx]->flags & ARROW_REGION) {
	  m->entries[indx]->flags &= ~ARROW_REGION;
	  drawArrow(w, m->entries[indx], m->entries[indx]->ppos, 
	       m->entries[indx]->width, False, False, select_gc, normal_gc);
	  return;
	} else
	  return;
      }
    }
  }
    
  displayItem(m, m->prev, False);		/* deselect previous item */
  if (inArrowRegion(m->entries[indx], x)) {
    m->entries[indx]->flags |= ARROW_REGION;   	/* select arrow region */
    displayItem(m, indx, True);		/* select new one */
  } else {
    m->entries[indx]->flags &= ~ARROW_REGION;  	/* deselect arrow region */
    displayItem(m, indx, True);			/* select new one */
  }    
  m->prev = indx;				/* store selection */
}

/* exConfigurePane: **********************************************************
 * is called when the size of the menu pane has been changed by means of the
 * window manager in the case of a panel. The panels new width and height are
 * provided to the menu system;
 ****************************************************************************/
/* ARGSUSED */
static void exConfigurePane(w, event, params, num_params)
     Widget w;
     XMotionEvent *event;
     String *params;
     Cardinal *num_params;
{
  exMenu *m = NULL;
  Dimension width = 0, height = 0;
  Arg args[3];
  int cnt, row_cnt;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &m); cnt++;
  XtGetValues(w, args, cnt);

  if (! m) 
    return;

  cnt = 0;
  XtSetArg(args[cnt], XmNwidth, &width); cnt++;
  XtSetArg(args[cnt], XmNheight, &height); cnt++;
  XtGetValues(m->pane, args, cnt);

  if (width != m->menuWidth || height != m->menuHeight) {
    m->menuHeight = height;
    m->menuWidth = width;

    if (m->itemCnt > 0) {
      row_cnt = height / m->entries[0]->height;
      if (row_cnt != m->itemDisplayCnt) {
	m->itemDisplayCnt = row_cnt;
	exSetScrollbar(m);
      }
    }
  }
}
 
/* exButtonDownProc: ********************************************************
 * callback that is called when the user presses a mouse button within a 
 * button in the menubar. The corresponding menu is pulled down. After 
 * execution, the return value is added to the menubars speific atom. During 
 * creation of the menubar the user has specified a function that is called 
 * whenever the value of the atom is changed this function is called. 
 * Addtionally, a proper<ty notify event with respect to the button is 
 * generated that that makes the button call its unhighlighte procedure 
 ***************************************************************************/

/*ARGSUSED*/
static void exButtonDownProc(w, event, params, num_params)
     Widget w;
     XButtonEvent *event;
     String *params;
     Cardinal *num_params;
{
  exMenu *m;
  char *ptr;
  
  m = pullMenuDown(w, True);			/* display and execute menu */

  if (m->w->core.being_destroyed)
    return;

  if (ptr = exGetMenuReturnValue(m)) 		/* make ret. value available */
    XChangeProperty(display, XtWindow(m->menubar->menubar), 	  
		    exMenuReturnValue, XA_STRING, 8, PropModeReplace,
		    (unsigned char *) ptr, strlen(ptr));


  XChangeProperty(display,  			/* trig. button to unhilight */
		  XtWindow(m->button), exMenuToggleButton, 
		  XA_STRING, 8, PropModeReplace, NULL, 0);
}

/*ARGSUSED*/
static void exButtonHelpProc(w, event, params, num_params)
     Widget w;
     XLeaveWindowEvent *event;
     String *params;
     Cardinal *num_params;
{  
  exMenu *m = (exMenu *) 0;
  Arg args[2];

  if (w->core.being_destroyed)
    return;

  XtSetArg(args[0], XmNuserData, &m);
  XtGetValues(w, args, 1);

  if (! m)
    return;

  helpUser(m, INTITLE);
}

/***************************************************************************
 *
 *		scrollbar handling routines
 *
 ***************************************************************************/

/* exScrollbar: ************************************************************
 * called whenever a motion or button event appears in the scrollbar area 
 **************************************************************************/

static void exScrollbar(m, e)
     exMenu *m;
     XEvent *e;
{
  register XmScrollBarPart *sbp;
  int y, moved, dir, amount, value, height, newtop;
  float fraction;

  sbp = & ((XmScrollBarRec *)m->scrollbar)->scrollBar;

  switch (e->type) {
  case ButtonRelease:
    if (e->xbutton.button != Button1)			/* not left button */
      return;
    m->slider_pos = -1;
    if (m->sb_timer)
      XtRemoveTimeOut(m->sb_timer);			/* remove time-out */
    m->sb_timer = NULL;					/* clear time-out */
    break;
  case ButtonPress:					/* butten pressed */
    if (e->xbutton.button != Button1)			/* not left button */
      return;
    y = e->xbutton.y;

    if (y >= sbp->arrow2_y) {				/* down-arrow */
      exScrollMenu(m, DOWN, 1, True);
      XSync(display, False);
      m->flags |= MH_FIRST_SCRL;
      m->sb_timer = XtAppAddTimeOut(XtWidgetToApplicationContext(m->w),
                          (unsigned long) MH_INITIAL_DELAY,
                          (XtTimerCallbackProc) exTimerEvent, (caddr_t) m);
    } else if (y <= sbp->arrow1_y + sbp->arrow_height) {/* up-arrow */
      exScrollMenu(m, UP, 1, True);
      XSync(display, False);
      m->flags |= MH_FIRST_SCRL;
      m->sb_timer = XtAppAddTimeOut(XtWidgetToApplicationContext(m->w),
                          (unsigned long) MH_INITIAL_DELAY,
                          (XtTimerCallbackProc) exTimerEvent, (caddr_t) m);
    } else if (y > sbp->arrow1_y + sbp->arrow_height && y < sbp->slider_y)
      exScrollMenu(m, UP, m->itemDisplayCnt / 4, True);
    else if (y > sbp->slider_y + sbp->slider_height && y < sbp->arrow2_y)
      exScrollMenu(m, DOWN, m->itemDisplayCnt / 4, True);
    else if (y > sbp->slider_y && y < sbp->slider_y + sbp->slider_height)
      m->slider_pos = y;
    break;
  case MotionNotify:
    if (! (e->xmotion.state & Button1Mask))	/* not left button */
      break;

    y = e->xmotion.y;

    if (m->slider_pos == y)			/* nothing changed */
      return;
    if (m->slider_pos < 0)			/* set it the first time */
      m->slider_pos = y;
    
    moved = y - m->slider_pos; 			/* move since last motion */
    
    if (moved == 0)				/* did not move */
      return;
    else if (moved < 0 && sbp->slider_y <= 0)	/* at top */
      return;
    else if (moved > 0 && 			/*at bottom */
	     sbp->slider_y + sbp->slider_height >= sbp->arrow2_y)
      return;
    
    m->slider_pos = y;		/* remember mouse pos for next time */
    
    XmScrollBarGetValues(m->scrollbar, &value, &height, &y, &y);
    
    value += moved;				/* adjust value */
    
    if (value < 0)			
      value = 0;
    else if (value > m->menuHeight - height)
      value = m->menuHeight - height;

    XmScrollBarSetValues(m->scrollbar, value, height, 0, 0, False);

    fraction = (float) value / (float) sbp->slider_area_height;
    newtop = (int) (fraction * m->itemsPerCol);

    if (m->topItem > newtop) {
      dir = UP;
      amount = m->topItem - newtop;
    } else {
      dir = DOWN;
      amount = newtop - m->topItem;
    }
    
    if (amount)
      exScrollMenu(m, dir, amount, False);
    break;
  }
}

/* exPanelScroll: ***********************************************************
 * called when a scrollbar event appears whithin a panel; this has to be 
 * handeled differently from menus, because we don't have a pointer grab set
 * and thus get no scrollbar events with respect to the menu pane 
 ****************************************************************************/
/*ARGSUSED*/
static void exPanelScroll(w, event, params, num_params)
     Widget w;
     XButtonEvent *event;
     String *params;
     Cardinal *num_params;
{
  exMenu *m = (exMenu *) 0;
  Arg args[2];

  XtSetArg(args[0], XmNuserData, &m);
  XtGetValues(w, args, 1);

  if (! m)
    return;

  exScrollbar(m, event);
}

/* exScrollMenu: *********************************************************** 
 * do the visual scrolling in a scrollable menu
 **************************************************************************/

static void exScrollMenu(m, dir, amount, autoset)
     exMenu *m;
     int dir;
     int amount;
     Boolean autoset;
{
  register i, j;
  int real_amount;
  int width, height;
  int oldtop = m->topItem;
  menuEntry *eptr;

  Window w = XtWindow(m->pane);

  m->dir = dir;

  switch (dir) {
  case DOWN:						/* scroll up */
    real_amount = 
      MIN(amount, m->itemsPerCol - (m->topItem + m->itemDisplayCnt));
    if (real_amount <= 0)
      return;
    m->topItem += real_amount;
    height = verticalDist(m, oldtop, m->topItem);
    width = m->menuWidth;
    XCopyArea(display, w, w, normal_gc, 		/* move items */
	      0, height, width, m->menuHeight, 0, 0);

    i = m->topItem + (m->cols - 1) * m->itemsPerCol + m->itemDisplayCnt;
    if (i > m->itemCnt) {
      eptr = m->entries[m->itemCnt - 1];
      XClearArea(display, w, eptr->xpos, 
		 eptr->ypos - m->entries[m->topItem]->ypos + eptr->height,
		 eptr->width, 
		 m->menuHeight - (i - m->itemCnt) * eptr->height, False);
    }

    for (j = 0; j < m->cols; j++) 		/* display new items */
      for (i = m->topItem + j * m->itemsPerCol + 
	   				m->itemDisplayCnt - real_amount; 
	     i < m->topItem + j * m->itemsPerCol + m->itemDisplayCnt; i++)
	if (i < m->itemCnt)
	  displayItem(m, i, (Boolean) m->selections[i]);
    break;
  case UP:
    real_amount = MIN(amount, m->topItem);
    if (real_amount <= 0)
      return;
    m->topItem -= real_amount;
    height = verticalDist(m, m->topItem, oldtop);
    width = m->menuWidth;
    XCopyArea(display, w, w, normal_gc, 0, 0, width, m->menuHeight, 0, height);

    for (j = 0; j < m->cols; j++) 
      for (i = m->topItem + j * m->itemsPerCol;
	   i < m->topItem + j * m->itemsPerCol + real_amount; i++)
	if (i < m->itemCnt)
	  displayItem(m, i, (Boolean) m->selections[i]);
    break;
  default:
    m->dir = -1;
    break;
  }

  if (autoset) 					/* adjust scrollbar */
    exSetScrollbar(m);
}

/* exSetScrollbar: *******************************************************
 * set the scrollbar according to the current top item 
 ************************************************************************/

static void exSetScrollbar(m)
     exMenu *m;
{
  int size = 1, value = 0;
  Arg args[5];
  int cnt;

  if (! m->scrollbar)
    return;

  if (m->itemDisplayCnt > 0) {
    size = (float)(m->menuHeight * m->itemDisplayCnt) / (float)m->itemsPerCol;
    if (size > m->menuHeight)
      size = m->menuHeight;

    if (m->topItem == 0)
      value = 0;		/* to top */
    else if (m->topItem >=  m->itemsPerCol - m->itemDisplayCnt)
      value = m->menuHeight - size; /* to bottom */
    else 
      value = 
	(int)((float) (m->menuHeight * m->topItem) / (float) m->itemsPerCol);
  }

  cnt = 0;
  XtSetArg(args[cnt], XmNmaximum, m->menuHeight); cnt++;
  XtSetArg(args[cnt], XmNvalue, value); cnt++;
  XtSetArg(args[cnt], XmNsliderSize, size); cnt++;
  XtSetValues(m->scrollbar, args, cnt);
}

/***************************************************************************
 *
 *            		   UTILITY FUNCTIONS
 *
 ***************************************************************************/

/* getActiveMenu: ***********************************************************
 * returns the active menu with respect to a menubar; this functionn is used 
 * to determine, whether a menu is already active (highlighted) in the case 
 * that the user moves the mouse into a button in the menubar. If an active
 * menu exists, it is unhighlighted (see exMenuMotionProc).
 ****************************************************************************/

static exMenu *getActiveMenu(m)
     register exMenu *m;
{
  register i;

  if (! m->menubar)			/* pure popup menu */
    return((exMenu *) 0);

  for (i = 0; i < m->menubar->num_buttons; i++)
    if (m->menubar->buttons[i].active == True)
      return(m->menubar->buttons[i].menu);
  
  return((exMenu *) 0);
}

/* thisMenuActive: **********************************************************
 * checks whether the current menu is active 
 ***************************************************************************/

static Boolean thisMenuActive(m)
     register exMenu *m;
{
  register i;

  if (! m->menubar)			/* popup menu */
    return(False);

  for (i = 0; i < m->menubar->num_buttons; i++)
    if (m->menubar->buttons[i].menu->button == m->button)
      return(m->menubar->buttons[i].active);
  
  return(False);
}

/* setMenuActive: *********************************************************
 * sets the active flag in the button structure to the specified value;
 * if True, the specified menu is displayed with at least one item selected;
 **************************************************************************/

void setMenuActive(m, active)
     exMenu *m;
     Boolean active;
{
  register i;

  if (! m->menubar)			/* popup menu */
    return;

  for (i = 0; i < m->menubar->num_buttons; i++)
    if (m->button == m->menubar->buttons[i].menu->button) {
       m->menubar->buttons[i].active = active;
       return;
     }
}

/* destroyMenu: *************************************************************
 * callback that is called when the toplevel widget of a menu is destroyed 
 ***************************************************************************/

/*ARGSUSED*/
static void destroyMenu(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
  exMenu *m = (exMenu *) client_data;
  register i, j;

  if (m->flags & PULLDOWN) {	   	/* update button information */
    exMenubar *mb = m->menubar;
    for (i = 0; i < MH_MENUBUTTON; i++)
      if (mb->buttons[i].button == m->button) {
	mb->num_buttons--;
	for (j = i; j < mb->num_buttons; j++) {
	  mb->buttons[j].menu   = mb->buttons[j + 1].menu;
	  mb->buttons[j].active = mb->buttons[j + 1].active;
	  mb->buttons[j].button = mb->buttons[j + 1].button;
	  mb->buttons[j].ptt    = mb->buttons[j + 1].ptt;
	}

	mb->buttons[mb->num_buttons].menu    = (exMenu *) NULL;
	mb->buttons[mb->num_buttons].active  = False;
	mb->buttons[mb->num_buttons].button  = (Widget) NULL;
	mb->buttons[mb->num_buttons].ptt     = (XtTranslations) NULL;

	break;
      }
  }

  freeMenu(m);

}

/* exTimerEvent: ************************************************************
 * callback function that is called when the user holds down a mouse button 
 * in the arrow region of a scrollbar.
 ***************************************************************************/
/*ARGSUSED*/
static void exTimerEvent(m, iid)
     exMenu *m;
     XtIntervalId *iid;
{
  int x, y;

  getMouse_xy(display, &x, &y);
  
  if (x < m->x + m->xs                 || /* to the left of the scrollbar */
      x > m->x + m->xs + m->sb_width   || /* to the right of the scrollbar */
      y < m->y + m->ys                 || /* above scrollbar */
      y > m->y + m->yb + m->menuHeight) { /* below scrollbar */
    if (m->sb_timer) {
      XtRemoveTimeOut(m->sb_timer);
      m->sb_timer = NULL;
    }
    return;
  }
  m->sb_timer = NULL;
  if (m->flags & MH_FIRST_SCRL) {
    XSync(display, False);
    m->flags &= ~MH_FIRST_SCRL;
    m->sb_timer = XtAppAddTimeOut(XtWidgetToApplicationContext(m->w),
                          (unsigned long) MH_INITIAL_DELAY,
                          (XtTimerCallbackProc) exTimerEvent, (caddr_t) m);
  } else {
    if (m->dir == UP || m->dir == DOWN)
      exScrollMenu(m, m->dir, 1, True);
    XSync(display, False);
    m->sb_timer = XtAppAddTimeOut(XtWidgetToApplicationContext(m->w),
                          (unsigned long) MH_INITIAL_DELAY,
                          (XtTimerCallbackProc) exTimerEvent, (caddr_t) m);
  }
}

/* checkEvent: **************************************************************
 * checks whether the event is a property-notify event that notifies that 
 * a new menu property has been added to the menubar. This function is used 
 * to discriminate between different properties of different clients 
 * (see function keys and macros for more information)
 ***************************************************************************/


Boolean checkEvent(event)
     XPropertyEvent *event;
{
  if (event->state != PropertyNewValue) {	/* improper event */
    setError(ERR_TYPE);
    return(False);
  }
  
  if (event->atom != exMenuReturnValue) {	/* improper atom */
    setError(ERR_TYPE);
    return(False);
  }
  return(True);
}   
     
/* dummy: ********************************************************************
 * a dummy function used as a noop action (see exCreateMenuBar)
 ****************************************************************************/
/*ARGSUSED*/
static void dummy(w, event, params, num_params)
     Widget w;
     XLeaveWindowEvent *event;
     String *params;
     Cardinal *num_params;
{
}

/* exMenubarExposeProc: *******************************************************
 * This callback is called, when the menu-pane window gets exposed. When the 
 * menu is mapped, the active grab has to be reset, so that we get all pointer 
 * events with respect to the menu-window. Otherwise, all pointer events go
 * to the previous window, until the mouse button is released. New active 
 * grabbing, however, is only necessary when the menu is newly mapped.
 ***************************************************************************/

/*ARGSUSED*/
static void  exMenubarExposeProc(w, event, params, num_params)
     Widget w;
     XExposeEvent *event;
     String *params;
     Cardinal *num_params;
{
  XDefineCursor(display, XtWindow(w), exMenuBarCursor);
}

/* inArrowRegion: ************************************************************
 * check if the cursor is located in the arrow-region of the specified item 
 ****************************************************************************/

static Boolean inArrowRegion(iptr, x)
     menuEntry *iptr;
     int x;
{
  register arrow_x = iptr->xpos + iptr->width - Arrow_width;

  return((x > arrow_x && x < arrow_x + Arrow_width) ? True : False);
}

/* displayable: *************************************************************
 * check if menu item with index indx lies within the displayable subset of 
 * items 
 ***************************************************************************/

static Boolean displayable(m, indx)
     exMenu *m;
     int indx;
{
  register i;

  if (indx < 0 || indx >= m->itemCnt)
    return(False);

  for (i = 0; i < m->cols; i++)			/* check each col */
    if (indx >= m->topItem + i * m->itemsPerCol &&
	indx < m->topItem + i * m->itemsPerCol + m->itemDisplayCnt)
      return(True);

  return(False);
}

/* setItemName: **************************************************************
 * assigns a new item name; used heavily in the file tree browser
 ****************************************************************************/

Boolean setItemName(m, slot, name)
     exMenu *m;
     int slot;
     char *name;
{

  if (slot < 0 || slot >= m->itemCnt)
    return(False);

  free(m->entries[slot]->name);
  m->entries[slot]->name = strCopy(name);
  return(True);
}

/* getCol: *******************************************************************
 * given a x-coordinate find the correspoding col;
 ****************************************************************************/

getCol(m, x)
     exMenu *m;
     int x;
{
  register i;
  register menuEntry *eptr;

  for (i = 0; i < m->cols; i++) {
    if (i * m->itemsPerCol >= m->itemCnt)
      break;
    eptr =  m->entries[i * m->itemsPerCol];
    if (x > eptr->xpos && x < eptr->xpos + eptr->width)
      return(i);
  }
  return(-1);
}
  
/* getRow: *******************************************************************
 * given a y-coordinate find the correspoding row;
 ****************************************************************************/

getRow(m, y)
     exMenu *m;
     int y;
{
  register menuEntry *eptr;
  register i;
  
  y += m->entries[m->topItem]->ypos - m->entries[0]->ypos;

  for (i = m->topItem;i< m->itemDisplayCnt + m->topItem && i< m->itemCnt;i++) {
    eptr = m->entries[i];
    if (y > eptr->ypos && y < eptr->ypos + eptr->height)
      return(i - m->topItem);
  }

  return(-1);
}

/* exPopdown: ****************************************************************
 * ungrab any widget that might has been grabbed within the request function;
 * then the menu is popped down. The calling menu is grabbed in the function 
 * callRequestFunction().
 ****************************************************************************/

exPopdown(w)
     Widget w;			/* popup-shell to be popped down */
{
  exUngrabAll(grabbedWidget);
  grabbedWidget = NULL;

  XtPopdown(w);					/* pop down requested widget */
}

/* callRequestFunction: ******************************************************
 * calls the function with the given parms; correct ungrabbing and 
 * grabbing is done
 ****************************************************************************/

int callRequestFunction(m, f, parms)
     exMenu *m;				/* calling menu */
     int (* f)();	   		/* function to call */
     RequestParm * parms;		/* parameter structure */
{
  int len;

  exUngrabAll(grabbedWidget);		/* ungrab any grabbed widget */
  grabbedWidget = NULL;			/* no menu grabbed now */

  len = (* f)(parms); 			/* invoke request function */

  exGrabAll(m->pane, exMenuCursor);

  grabbedWidget = m->pane;		/* re-grab the calling menu */

  return(len);
}

/* titleHeight: **************************************************************
 * calculate the height of the title label for the given menu including the 
 * corresponding separator
 *****************************************************************************/

static titleHeight(m)
     exMenu *m;
{
  Arg args[10];
  int cnt, h;
  Dimension mbot, mtop, mh, bw, st;
  XmFontList fl = NULL;
  XmString xstr;

  if (! m->title)
    return(0);

  cnt = 0;
  XtSetArg(args[cnt], XmNfontList, &fl); cnt++;
  XtSetArg(args[cnt], XmNlabelString, &xstr); cnt++;
  XtSetArg(args[cnt], XmNmarginBottom, &mbot); cnt++;
  XtSetArg(args[cnt], XmNmarginTop, &mtop); cnt++;
  XtSetArg(args[cnt], XmNmarginHeight, &mh); cnt++;
  XtSetArg(args[cnt], XmNshadowThickness, &st); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, &bw); cnt++;

  XtGetValues(m->title, args, cnt);

  h = mbot + mtop + 2 * (st + bw + mh) + XmStringHeight(fl, xstr);
  h += sepHeight(m->title_sep);

  return(h);
}

/* doneHeight: ****************************************************************
 * calculate the height of the done box for the given menu including the 
 * corresponding separator
 *****************************************************************************/

static doneHeight(m)
     exMenu *m;
{
  Arg args[3];
  int cnt, h;
  Dimension bw, mh, st;

  if (! m->done_box)
    return(0);

  cnt = 0;
  XtSetArg(args[cnt], XmNmarginHeight, &mh); cnt++;
  XtSetArg(args[cnt], XmNshadowThickness, &st); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, &bw); cnt++;
  XtGetValues(m->done_box, args, cnt);

  h = 2 * (st + bw + mh) + m->titleFontHeight + 3 * MH_MARGIN;
  h += sepHeight(m->done_sep);

  return(h);
}

/* sepHeight: ****************************************************************
 * calculate the height of the given separator
 ****************************************************************************/

static sepHeight(s)
     Widget s;
{
  Arg args[3];
  int cnt;
  Dimension ht, st, bw;

  if (! s)
    return(0);

  cnt = 0;
  XtSetArg(args[cnt], XmNhighlightThickness, &ht); cnt++;
  XtSetArg(args[cnt], XmNshadowThickness, &st); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, &bw); cnt++;
  XtGetValues(s, args, cnt);

  return(2 * (ht + st + bw));
}

/* redrawDoneBox: ************************************************************
 * redraw the done box whenever an expose event occures 
 ****************************************************************************/
/* ARGSUSED */
static void redrawDoneBox(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
  drawDoneBox((exMenu *) client_data);
}
