/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
/*****************************************************************************
 * This file contains the program to generate the FunctionKeyManagers menus;
 * the generated application node "FKM-functions" is stored in file "fk.menu"
 ****************************************************************************/

#include <stdio.h>
#include <exBasics.h>
#include <exErrors.h>
#include <exFonts.h>
#include <exMDS.h>
#include <exMenu.h>

/* Quit */
extern fkmeditQuit();

/* function key sets */
extern fkmeditLoadSet();
extern fkmeditNewSet();
extern fkmeditSaveSet();
extern fkmeditSaveSetAs();
extern fkmeditDeleteSet();
extern fkmeditShowSet();
extern fkmeditHideSet();

/* function keys */
extern fkmeditLoadKey();
extern fkmeditNewKey();
extern fkmeditSaveKey();
extern fkmeditSaveKeyAs();
extern fkmeditSaveKeyIn();
extern fkmeditDeleteKey();

extern fkmeditAssignKey();
extern fkmeditCopyKey();
extern fkmeditPasteKey();

Widget top;

main(argc, argv)
     int argc;
     char **argv;
{
  char *getenv();
  char str[200];
  MHmenu *mptr;
  MHitem *iptr;
  MHiReq *req;
  int mask[12];

#ifdef DEBUG
  malloc_debug(2);
#endif

  top = XtInitialize(NULL, "fkm menus", (XrmOptionDescRec *) NULL, 0, &argc, 
		     argv);

  initMDS();
  exInitLayout(XtDisplay(top));
  exInitMenus(XtDisplay(top));

  newApplicationNode("FKM-functions");
  
  insertFunction(currentApplication(), fkmeditQuit, "fkmeditQuit");

  insertFunction(currentApplication(), fkmeditLoadSet, "fkmeditLoadSet");
  insertFunction(currentApplication(), fkmeditNewSet, "fkmeditNewSet");
  insertFunction(currentApplication(), fkmeditSaveSet, "fkmeditSaveSet");
  insertFunction(currentApplication(), fkmeditSaveSetAs, "fkmeditSaveSetAs");
  insertFunction(currentApplication(), fkmeditDeleteSet, "fkmeditDeleteSet");
  insertFunction(currentApplication(), fkmeditShowSet, "fkmeditShowSet");
  insertFunction(currentApplication(), fkmeditHideSet, "fkmeditHideSet");

  insertFunction(currentApplication(), fkmeditLoadKey, "fkmeditLoadKey");
  insertFunction(currentApplication(), fkmeditNewKey, "fkmeditNewKey");
  insertFunction(currentApplication(), fkmeditSaveKey, "fkmeditSaveKey");
  insertFunction(currentApplication(), fkmeditSaveKeyAs, "fkmeditSaveKeyAs");
  insertFunction(currentApplication(), fkmeditSaveKeyIn, "fkmeditSaveKeyIn");
  insertFunction(currentApplication(), fkmeditDeleteKey, "fkmeditDeleteKey");
  insertFunction(currentApplication(), fkmeditAssignKey, "fkmeditAssignKey");
  insertFunction(currentApplication(), fkmeditCopyKey, "fkmeditCopyKey");
  insertFunction(currentApplication(), fkmeditPasteKey, "fkmeditPasteKey");

/* ------------------------------------------------------------ */
   mptr = newMenu("quit");
  setMenuType(mptr, MDS_M_PULLD);
  setSelType(mptr, MDS_CNT_EQ);
  setSelCnt(mptr, 1);
  sprintf(str, "Option to quit the function-key editor.");
  mptr->m_hlp->hlpmsg = strCopy(str);

  iptr = newItem(mptr, "quit", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Quit the function-key editor.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(fkmeditQuit, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

/* ------------------------------------------------------------ */

  mptr = newMenu("key sets");
  setMenuType(mptr, MDS_M_PULLD);
  setSelType(mptr, MDS_CNT_EQ);
  setSelCnt(mptr, 1);
  sprintf(str, "Options to manage sets of function-keys.");
  mptr->m_hlp->hlpmsg = strCopy(str);

  iptr = newItem(mptr, "load set", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Load a function-key set and show the first key description.\nYou get a menu containing the list of function-keys\nfrom which you can select the desired one.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(fkmeditLoadSet, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  iptr = newItem(mptr, "new set", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Create a new function-key set.\nYou are requested to enter the name for the set.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(fkmeditNewSet, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  iptr = newItem(mptr, "save set", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Save the current set to a file.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(fkmeditSaveSet, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  iptr = newItem(mptr, "save set as", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Duplicate the current function-key set and store it under a different name.\nYou are requested to enter the new name for the set.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(fkmeditSaveSetAs, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  iptr = newItem(mptr, "delete set", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Delete a set and remove the corresponding file.\nYou get a menu containing the list of function-key sets\nfrom which you can select the desired one.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(fkmeditDeleteSet, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  iptr = newItem(mptr, "display set", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Preview  the current function-key set.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(fkmeditShowSet, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  iptr = newItem(mptr, "hide set", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Make the preview window go away.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(fkmeditHideSet, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

/* -------------------------------------------------------------------------*/

  mptr = newMenu("keys");
  setMenuType(mptr, MDS_M_PULLD);
  setSelType(mptr,MDS_CNT_EQ);
  setSelCnt(mptr,1);
  sprintf(str, "Options to manage the keys of the current set.");
  mptr->m_hlp->hlpmsg = strCopy(str);

  iptr = newItem(mptr, "load key", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Load a key from the current function-key set.\nYou get a menu containing the list of function-keys\nfrom which you can select the desired one.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(fkmeditLoadKey, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  iptr = newItem(mptr, "new key", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Create a new key in the current function-key set.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(fkmeditNewKey, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  iptr = newItem(mptr, "save key", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Store the current key into the current set.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(fkmeditSaveKey, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  iptr = newItem(mptr, "save key as", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Duplicate the current key and store it in the current set.\nYou are requested to enter the new name for the key.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(fkmeditSaveKeyAs, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  iptr = newItem(mptr, "save key in", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Store the current key in a different set.\nYou get a menu containing the list of available function-key sets\nfrom which you can select the desired one.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(fkmeditSaveKeyIn, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  iptr = newItem(mptr, "delete key", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Delete a key from the current set.\nYou get a menu containing the list of function-keys\nfrom which you can select the desired one.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(fkmeditDeleteKey, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  iptr = newItem(mptr, "------------------", EOS);

  iptr = newItem(mptr, "assign key", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Assign a physical function-key to the current key.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(fkmeditAssignKey, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  iptr = newItem(mptr, "------------------", EOS);

  iptr = newItem(mptr, "copy key", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Preserve a copy of the current key description in an internal buffer.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(fkmeditCopyKey, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  iptr = newItem(mptr, "paste key", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Paste the contents of the internal buffer in the current key.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(fkmeditPasteKey, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  setFile(currentApplication(), "fk.menu");
  dumpToFile(top);
  exit(0);
}

fkmeditLoadKey()
{
}

fkmeditSaveKey()
{
}

fkmeditSaveSet()
{
}

fkmeditSaveSetAs()
{
}

fkmeditLoadSet()
{
}

fkmeditDeleteSet()
{
}

fkmeditCloseSet()
{
}

fkmeditQuit()
{
}

fkmeditNewKey()
{
}

fkmeditNewSet()
{
}

fkmeditAssignKey()
{
}

fkmeditCopyKey()
{
}

fkmeditPasteKey()
{
}

fkmeditDeleteKey()
{
}

fkmeditShowSet()
{
}

fkmeditHideSet()
{
}

fkmeditSaveKeyAs()
{
}

fkmeditSaveKeyIn()
{
}
