/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
/****************************************************************************
 * This file contains the functions generating the menu editor and the main
 * event loop
 ****************************************************************************/

#include <stdio.h>
#include <sys/file.h>
#include <signal.h>
#include <Xm/ScrolledW.h>
#include <Xm/ScrollBar.h>
#include <Xm/Form.h>
#include <Xm/RowColumn.h>
#include <Xm/LabelG.h>

#include <exMenu.h>
#include <ife.h>
#include "editDescr.h"
#include "formEntry.h"
#include "icons/menu"
#include "version.h"

static Boolean initEditMenus();
static eshHandler();
static processEvents();
static processMessage();
static menuEditorHints();
static formDescr *createEditForm();
static headerDescr *createHeadForm();
static meditWindow *initEditWindow();
static MenuEditorFont();
static MenuEditorQuit();
static void getmeMenuResult();
static void meDoScroll();

char msg[NOTELEN];                            /* global array for messages */
struct runMenuEdit *rme;

MenuDescr EditMenuWin[MDS_MAXMENU];
int WinCnt = MDS_MAXMENU;

static void closeME();
static void openME();

static XtActionsRec ifeCmds[] = {
  {"closeME", closeME},
  {"openME", openME}
};

static char ifeCmdTable[] = 
  "<Map>	        : openME() \n \
   <Unmap>	        : closeME()";

/* main program **************************************************************
 * initializes menu editor; contains event loop
 ****************************************************************************/

main(argc, argv)
     int argc;
     char *argv[]; 
{
  Widget w;
  XtAppContext app_context;
  int DELETED = 1, fw, fh,
      nfds, maxfd = getdtablesize();
  fd_set readfds[2];
  int dpy;
  char msg[50];
  static XtTranslations ptt;
  Display *d;
  Boolean connected = False;

  /*********** initialize "menuEdit" as top-level shell *******************/

#ifdef DEBUG
  malloc_debug(2);
#endif

  if (argc > 1 && strcmp(argv[1], "-c") == 0)
    connected = True;

  w = XtInitialize(NULL, "MENUEDIT", (XrmOptionDescRec *) NULL, 0, &argc, 
		   argv);
  d = XtDisplay(w);

  app_context = XtWidgetToApplicationContext(w);
  XSynchronize(d, TRUE); 
  XSetErrorHandler(eshHandler);
  signal(SIGHUP, MenuEditorQuit);

  XtAddActions(ifeCmds, 2);
  ptt = XtParseTranslationTable(ifeCmdTable);
  XtOverrideTranslations(w, ptt);

  /********************* initialize fonts *********************************/

  exInitLayout(d);                      

  /********************* initialize menu editor ***************************/

  sprintf(msg, "-------- loading menu editor ... -------");
  initThermometer(d, w, 100, 100, strlen(msg), msg, NULL);
  setThermometer(100, 100, 0);

  rme = (struct runMenuEdit *) Malloc(sizeof(struct runMenuEdit));
  rme->menuEditDescr = initEditWindow(w, XPOS, YPOS); /* initialize widgets */

  if (!initEditMenus(rme->menuEditDescr))             /* initialize menus */
    exit(1);

  if (! generateApplList(ApplName))
    exit(1);

  rme->menuPtr = initialMenu((MHmenu *) 0);  
  rme->findPtr = (MHmenu *) 0;

  if (! (rme->applPtr = createInitApplication(UNTITLEDAPPL))) {
                                                    /* find first menu set */
    rme->applPtr = currentApplication(); 
    if (rme->findPtr = findFirstMenu(rme->applPtr))          
      copyMenuDescr(rme->findPtr, rme->menuPtr);      /* find first accessable
							 menu */
  }

  setThermometer(100, 100, 70);

  if (!generateMenuList(rme->applPtr, MenuName) || 
      !generateFuncList(rme->applPtr, FuncName, ApplFuncName))
    exit(1);

  setThermometer(100, 100, 80);

  fillFormWithAppl(rme->menuEditDescr, rme->applPtr);        /* fill in form */
  setThermometer(100, 100, 85);
  fillFormWithMenu(rme->menuEditDescr, rme->menuPtr); 
  setThermometer(100, 100, 90);
  fillEditPopup(rme->applPtr, rme->menuPtr);      /* set menu default values */
  setThermometer(100, 100, 95);
  fillEditPulldown(rme->menuPtr);
  setThermometer(100, 100, 100); 

  hideThermometer(10);

  /********************* start menu editor ********************************/

  setWorkApplication();
  
  XtRealizeWidget(rme->menuEditDescr->meShell);  

  if (connected)
    exSendMsg(fileno(stdout), getpid(), IFE_OPEN, 0);

  rme->menuEditDescr->flag |= MEDITOR_OPEN;
  menuEditorHints(rme->menuEditDescr);
  
  setbuf(stdin, NULL);
  setbuf(stdout, NULL);

  dpy = ConnectionNumber(d);

  FD_ZERO(&readfds[0]);
  FD_SET(dpy, &readfds[0]);
  FD_SET(fileno(stdin), &readfds[0]);

  while (DELETED) {

    bcopy((char *) &readfds[0], (char *) &readfds[1], sizeof(readfds[0])); 
    processEvents();
   
    nfds = select(maxfd, &readfds[1], (fd_set *) 0, (fd_set *) 0, 
		  (struct timeval *) 0); 
    if (FD_ISSET(dpy, &readfds[1]))                   /* process events */
      processEvents();    
    else if (FD_ISSET(fileno(stdin), &readfds[1]))  	  /* process IO */
      processMessage(rme->menuEditDescr);    
  }
}

/* processEvents *************************************************************
 * process all pending events 
 ****************************************************************************/

static processEvents()
{	
  static XEvent event;
  
  while (XtPending()) {
    XtNextEvent(&event);
    XtDispatchEvent(&event);
  }
}

/* processMessage: **********************************************************
 * process the messages that are sent by the interface editor
 ***************************************************************************/

static processMessage(mEdit)
     meditWindow *mEdit;
{
  int pid, opc, info;
  
  if (exReceiveMsg(fileno(stdin), &pid, &opc, &info) < 0)
    return;

  if (pid != getpid())
    return;
    
  fprintf(stderr, "Message received(menu editor): pid=%d, opc=%d, info=%d\n",
	  pid, opc, info);

  switch (opc) {
  case IFE_OPEN:
    XtMapWidget(mEdit->meShell);
    mEdit->flag |= MEDITOR_OPEN;
    break;
  case IFE_CLOSE:
    XtUnmapWidget(mEdit->meShell);
    mEdit->flag &= MEDITOR_CLOSED;
    break;
  default:
    break;
  }
  XFlush(XtDisplay(mEdit->meShell));
}

/* menuEditorHints ***********************************************************
 * set window manager hints and size hints
 ****************************************************************************/

static menuEditorHints(mEdit)
     meditWindow *mEdit;
{
  XWMHints wm_hints;
  XSizeHints size_hints;
  int mb_height, fh, fw;
  Display *d = XtDisplay(mEdit->meShell);
  XFontStruct *formFont = 
    getEditorFont(mEdit->form->ItemEntries[0]->entry->entryEditor);

  mb_height  = menuBarHeight(mEdit->menuBar); 
  fh = exFontHeight(formFont);
  fw = exFontWidth(formFont);

  wm_hints.flags = IconPixmapHint | IconPositionHint | PMinSize | PResizeInc;
  wm_hints.icon_pixmap = 
    XCreateBitmapFromData(d, XDefaultRootWindow(d), 
			  menu_bits, menu_width, menu_height);
  wm_hints.icon_x = 1060;
  wm_hints.icon_y = 800;

  size_hints.flags = PMinSize | PBaseSize | PResizeInc | PMaxSize;

  size_hints.base_width  = WIDTH;
  size_hints.base_height = HEIGHT;

  size_hints.min_width  = fw * (E_ENTRYLEN3 + E_ENTRYLEN1 + 20);
  size_hints.min_height = mb_height + fh * 20;

  size_hints.max_width  = 1.2 * WIDTH;
  size_hints.max_height = 2 * HEIGHT;

  size_hints.width_inc  = fw;
  size_hints.height_inc = getFormEntryHeight(mEdit->form->ItemEntries[0]);

  XSetWMHints(d, XtWindow(mEdit->meShell), &wm_hints);
  XSetNormalHints(d, XtWindow(mEdit->meShell), &size_hints);
}

/* fillFormWithAppl **********************************************************
 * fills in form with application info
 ****************************************************************************/

fillFormWithAppl(editw, aptr)
     meditWindow *editw;
     ApplicationNode *aptr;
{
  register cnt;
  Arg args[10];
  char *msg;

  if (!aptr) {
    setError(ERR_NULL);
    return;
  }

  cnt = 0;                                /* application name */
  XtSetArg(args[cnt], XmNlabelString, XmStringCreateLtoR
	   (aptr->applicationName, XmSTRING_DEFAULT_CHARSET)); cnt++;
  XtSetValues(editw->head->setName[1], args, cnt);

  if ((aptr->a_access & (MDS_READACC | MDS_CHNGACC)) == 0)      
    msg = "no access";
  else if ((aptr->a_access & MDS_READACC) != 0)
    msg = "read access";
  else if ((aptr->a_access & MDS_CHNGACC) != 0)
    msg = "change access";

  cnt = 0;                                /* application access */
  XtSetArg(args[cnt], XmNlabelString, XmStringCreateLtoR
	   (msg, XmSTRING_DEFAULT_CHARSET)); cnt++;
  XtSetValues(editw->head->setAccess[1], args, cnt);

  if ((aptr->a_access & MDS_SYS) != 0)
    msg = "yes";
  else if ((aptr->a_access & MDS_SYS) == 0)
    msg = "no";  

  cnt = 0;                                /* system menu set */
  XtSetArg(args[cnt], XmNlabelString, XmStringCreateLtoR
	   (msg, XmSTRING_DEFAULT_CHARSET)); cnt++;
  XtSetValues(editw->head->systemMenu[1], args, cnt);

  setError(ERR_NOERR);
}

/* fillFormWithMenu *********************************************************
 * fills in form with values of menu description
 ****************************************************************************/

fillFormWithMenu(editw, mptr)
     meditWindow *editw;
     MHmenu *mptr;
{ 
  register i;
  Boolean readOnly = False;

  if (!mptr) {
    setError(ERR_NULL);
    return;
  }

  XtMoveWidget(editw->meRowCol, 0, 0);

  if ((mptr->m_flags & MDS_READACC) != 0)
    readOnly = TRUE;

  fillEntry(editw->head->MenuEntry, mptr->m_repr->name, readOnly);

  for (i = 0; i < mptr->itemCnt; i++) {
    fillEntry(editw->form->ItemEntries[i], mptr->i_table[i]->i_repr->name, 
	      readOnly);
    fillEntry(editw->form->ReturnEntries[i], 
	      mptr->i_table[i]->i_return->retPattern, readOnly);
  }

  for (i = mptr->itemCnt; i < editw->form->entryCnt; i++){
    fillEntry(editw->form->ItemEntries[i], (char *) NULL, readOnly);
    fillEntry(editw->form->ReturnEntries[i], (char *) NULL, readOnly);
  }

  updateReqCnt(editw, mptr, editw->form->entryCnt);

  setError(ERR_NOERR);
}

/* fillEditPopup ***********************************************************
 * fills in popup menus (in work application) with values of application
 ****************************************************************************/

fillEditPopup(aptr, mptr)
     ApplicationNode *aptr;
     MHmenu *mptr;
{
  register i;
  MHmenu *meMptr;

  /******************************** menu 'CHANGE' **************************/

  setWorkApplication();

  if (!(meMptr = getMenuDescrByName(CHANGE))){
    setError(ERR_NULL);
    return;
  }
 
  setRequestInitialValue(meMptr->i_table[0]->i_requests[0], 
			 aptr->file);                 /* application file */
  setRequestInitialValue(meMptr->i_table[1]->i_requests[0], 
			 aptr->applicationName);      /* application name */

  /*if (aptr->a_access & MDS_SYS)
    setToggle(meMptr->i_table[3], TRUE);
  else 
    setToggle(meMptr->i_table[3], False);*/

  /********************** menu 'application access' ************************/

  if (!(meMptr = getMenuDescrByName(MENUSETACCESS))) {
    setError(ERR_NULL);
    return;
  } 
  
  for (i = 0; i < meMptr->itemCnt; i++)
    resetDefaultItem(meMptr->i_table[i]);
  
  if ((aptr->a_access & MDS_READACC) != 0) 
    setDefaultItem(meMptr->i_table[0]); 
  else if ((aptr->a_access & MDS_CHNGACC) != 0)
    setDefaultItem(meMptr->i_table[1]); 

  /*********** menu 'application list' and 'delete application' ***********/

  adjustDefaultAppl(aptr);

  /********* menu 'menu list','delete menu' and 'menu copy' ***************/

  adjustDefaultMenu(mptr);

  setCurrentApplication(aptr);
  setError(ERR_NOERR);
}

/* fillEditPulldown ********************************************************
 * fills in pulldown menus (in application EDITMENUS) with values of menu 
 * description
 ****************************************************************************/

fillEditPulldown(mptr)
     MHmenu *mptr;
{
  ApplicationNode *aptr;
  MHmenu *meMptr;
  exMenu *menu;
  char str[20];

  aptr = currentApplication();
  setCurrentApplication(getApplByName(EDITMENUS));

  /************************ menu 'menu type' *******************************/

  if (!(meMptr = getMenuDescrByName(MENUTYPE)) 
      || (!(menu = getMenuByName(MENUTYPE, EditMenuWin, WinCnt))))
    return(setError(ERR_NULL));

  resetSelection(meMptr, menu);

  if (mptr->m_type == MDS_M_PULLD) 
    setDefaultItem(meMptr->i_table[0]);  
  else if (mptr->m_type == MDS_M_POPUP) 
    setDefaultItem(meMptr->i_table[1]); 
  else if (mptr->m_type == MDS_L_PULLD) 
    setDefaultItem(meMptr->i_table[2]);
  else if (mptr->m_type == MDS_L_POPUP) 
    setDefaultItem(meMptr->i_table[3]);

  menu->erase = False;
  exMakeDefaultMenu(menu);

  /************************ menu 'menu access' *****************************/

  if (!(meMptr = getMenuDescrByName(MENUACCESS)) 
      || (!(menu = getMenuByName(MENUACCESS, EditMenuWin, WinCnt))))
    return(setError(ERR_NULL));

  resetSelection(meMptr, menu);

  if ((mptr->m_flags & MDS_READACC) != 0) 
    setDefaultItem(meMptr->i_table[0]); 
  else if ((mptr->m_flags & MDS_CHNGACC) != 0)
    setDefaultItem(meMptr->i_table[1]); 

  menu->erase = False;
  exMakeDefaultMenu(menu);

  /************************ menu 'menu position' ***************************/

  if (!(meMptr = getMenuDescrByName(MENUPOSITION)) 
      || (!(menu = getMenuByName(MENUPOSITION, EditMenuWin, WinCnt))))
    return(setError(ERR_NULL));

  resetSelection(meMptr, menu);

  setDefaultItem(meMptr->i_table[0]);     
  if (mptr->m_pos.posType == MDS_POS_ABS) 
    setDefaultItem(meMptr->i_table[1]);	
  else if (mptr->m_pos.posType == MDS_POS_CREL) 
    setDefaultItem(meMptr->i_table[2]); 

  sprintf(str, "(%d, %d)", mptr->m_pos.x, mptr->m_pos.y);
  meSetInitialValue(meMptr, menu, 0, 0, str); /* new initial value in entry */

  menu->erase = False;
  exMakeDefaultMenu(menu);

  /************************ menu 'item selection' ***************************/

  if (!(meMptr = getMenuDescrByName(ITEMSELECTION)) 
      || (!(menu = getMenuByName(ITEMSELECTION, EditMenuWin, WinCnt))))
    return(setError(ERR_NULL));

  resetSelection(meMptr, menu);

  setDefaultItem(meMptr->i_table[0]);     
  if (mptr->i_selType == MDS_CNT_EQ) 
    setDefaultItem(meMptr->i_table[1]); 
  else if (mptr->i_selType == MDS_CNT_LT) 
    setDefaultItem(meMptr->i_table[2]);   
  else if (mptr->i_selType == MDS_CNT_LE) 
    setDefaultItem(meMptr->i_table[3]);  
  else if (mptr->i_selType == MDS_CNT_GT) 
    setDefaultItem(meMptr->i_table[4]);  
  else if (mptr->i_selType == MDS_CNT_GE) 
    setDefaultItem(meMptr->i_table[5]); 

  sprintf(str, "%d", mptr->i_selCnt);
  meSetInitialValue(meMptr, menu, 0, 0, str); /* new initial value in entry */

  menu->erase = False;
  exMakeDefaultMenu(menu);

  setCurrentApplication(aptr);
  return(setError(ERR_NOERR));
}

/* initEditMenus *************************************************************
 * initialize global variables of menu editor
 *****************************************************************************/

static Boolean initEditMenus(editw)
     meditWindow *editw;
{
  register i;
  char msg[50];
  char *menupath;
  char *getenv();
  ApplicationNode *aptr;

  /**************** load menus of menu editor from file *********************/

  MH_FUNCTIONS[0].func = (int (*)()) 0;
  MH_FUNCTIONS[0].funcName = (char *) 0;

  i = 1;      /* functions must have the same sequence as in definition! */
  MH_FUNCTIONS[i].func = (int (*)()) adminSave;
  MH_FUNCTIONS[i].funcName = "adminSave"; i++;

  MH_FUNCTIONS[i].func = (int (*)()) adminQuit;
  MH_FUNCTIONS[i].funcName = "adminQuit"; i++;

  MH_FUNCTIONS[i].func = (int (*)()) adminNewSet;
  MH_FUNCTIONS[i].funcName = "adminNewSet"; i++;

  MH_FUNCTIONS[i].func = (int (*)()) adminDuplSet;
  MH_FUNCTIONS[i].funcName = "adminDuplSet"; i++;

  MH_FUNCTIONS[i].func = (int (*)()) adminDuplMenu;
  MH_FUNCTIONS[i].funcName = "adminDuplMenu"; i++;

  MH_FUNCTIONS[i].func = (int (*)()) newApplList;
  MH_FUNCTIONS[i].funcName = "newApplList"; i++;

  MH_FUNCTIONS[i].func = (int (*)()) adminUserFunct;
  MH_FUNCTIONS[i].funcName = "adminUserFunct"; i++;

  MH_FUNCTIONS[i].func = (int (*)()) ApplChange;
  MH_FUNCTIONS[i].funcName = "ApplChange"; i++;

  MH_FUNCTIONS[i].func = (int (*)()) adminNewMenu;
  MH_FUNCTIONS[i].funcName = "adminNewMenu"; i++;

  MH_FUNCTIONS[i].func = (int (*)()) adminPreview;
  MH_FUNCTIONS[i].funcName = "adminPreview"; i++;

  MH_FUNCTIONS[i].func = (int (*)()) newMenuList;
  MH_FUNCTIONS[i].funcName = "newMenuList"; i++;

  MH_FUNCTIONS[i].func = (int (*)()) editMenuType;
  MH_FUNCTIONS[i].funcName = "editMenuType"; i++;

  MH_FUNCTIONS[i].func = (int (*)()) editAccessType;
  MH_FUNCTIONS[i].funcName = "editAccessType"; i++;

  MH_FUNCTIONS[i].func = (int (*)()) editMenuPos1;
  MH_FUNCTIONS[i].funcName = "editMenuPos1"; i++;

  MH_FUNCTIONS[i].func = (int (*)()) editMenuPos2;
  MH_FUNCTIONS[i].funcName = "editMenuPos2"; i++;

  MH_FUNCTIONS[i].func = (int (*)()) editMenuActMask;
  MH_FUNCTIONS[i].funcName = "editMenuActMask"; i++;

  MH_FUNCTIONS[i].func = (int (*)()) editSelCnt;
  MH_FUNCTIONS[i].funcName = "editSelCnt"; i++;

  MH_FUNCTIONS[i].func = (int (*)()) editSelType;
  MH_FUNCTIONS[i].funcName = "editSelType"; i++;

  MH_FUNCTIONS[i].func = (int (*)()) editRequest;
  MH_FUNCTIONS[i].funcName = "editRequest"; i++;

  MH_FUNCTIONS[i].func = (int (*)()) editMenuHelp;
  MH_FUNCTIONS[i].funcName = "editMenuHelp"; i++;

  setThermometer(100, 100, 25);
  hideThermometer(2);

  if (!LoadEditMenus(editw->meShell, getMenuDir(), MEDITORMENUS)) {  
    fprintf(stderr, "menus of menu editor cannot be loaded\n");
    return(False);
  }


  /****** create popup menus of menu editor in WorkApplication ***********/

  sprintf(msg, "-------- loading menu editor ... -------");
  initThermometer(XtDisplay(editw->meShell), editw->meShell, 100, 100, 
		  strlen(msg), msg, NULL);
  setThermometer(100, 100, 30);

  if (!createInternalMenus(editw->meShell)) {
    fprintf(stderr, "internal popup menus of menu editor cannot be created\n");
    return(False);
  } 

  setThermometer(100, 100, 60);
  hideThermometer(2);

  /************* load user menus in MDS-structure ************************/
  
  for (i = 0; i < MDS_MAXFUNC; i++) {
    MH_FUNCTIONS[i].func = (int (*)()) (i);
    MH_FUNCTIONS[i].funcName = (char *) 0;
  }

  if ((menupath = getenv("MENUPATH")) == (char *) 0) {
    fprintf(stderr, "Cannot find menu-directory.\n");
    return(False);
  }

  if (!loadMenuDir(editw->meShell, menupath)) {
    fprintf(stderr, "problems with user environment or user defined menus\n");
    return(False);
  }

  /************* load menus of menu editor into menubar ********************/

  sprintf(msg, "-------- loading menu editor ... -------");
  initThermometer(XtDisplay(editw->meShell), editw->meShell, 100, 100, 
		  strlen(msg), msg, NULL);
  setThermometer(100, 100, 65);

  aptr = getApplByName(EDITMENUS);
  WinCnt = aptr->menuCnt;    

  exInitMenus(XtDisplay(editw->meShell));

  if (!loadMenus(editw->menuBar, editw->meShell, aptr, EditMenuWin, 
		&WinCnt)) {
    fprintf(stderr, "error while creating menu description\n");
    return(False);
  }

  return(TRUE);
}

/* initEditWindow ************************************************************
 * generates windows and handler for menu editor 
 ****************************************************************************/

static meditWindow *initEditWindow(parent, xval, yval)
     Widget parent;
     int xval, yval;                   /* x-/y-value of menu editor window */
{
  Arg args[20];
  register cnt;
  Widget scrlWidget;
  meditWindow *editw;
  int we, he, scrlbWidth;
  Dimension sbHeight, sbWidth; 

  scrlbWidth = 3 * exFontHeight(exBaseFont(XtDisplay(parent))) / 2;

  editw = (meditWindow *) Malloc(sizeof(meditWindow));
  editw->meShell = parent;

  he = HEIGHT;
  we = WIDTH;

  cnt = 0;
  XtSetArg(args[cnt], XmNx, xval); cnt++;
  XtSetArg(args[cnt], XmNy, yval); cnt++;
  XtSetArg(args[cnt], XmNtitle, VERSION); cnt++;
  XtSetArg(args[cnt], XmNiconName, "ERGO Menu Editor"); cnt++;
  XtSetArg(args[cnt], XmNiconX, 1060); cnt++;
  XtSetArg(args[cnt], XmNiconY, 800); cnt++;
  XtSetArg(args[cnt], XmNkeyboardFocusPolicy, XmPOINTER); cnt++;
  XtSetValues(editw->meShell, args, cnt);

  /*** create widgets of menu editor ***/

  cnt = 0;
  XtSetArg(args[cnt], XmNwidth, we); cnt++;
  XtSetArg(args[cnt], XmNheight, he); cnt++;
  XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;
  XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;

  editw->meForm = XtCreateManagedWidget("meForm", xmFormWidgetClass, 
					    editw->meShell, args, cnt);

  setThermometer(100, 100, 5);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;

  editw->menuBar = exCreateMenuBar(editw->meForm, args,
				   cnt, getmeMenuResult);

  editw->head = createHeadForm(editw);
  
  cnt = 0;
  XtSetArg(args[cnt], XmNwidth, scrlbWidth); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, editw->head->labelSet[0]); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;

  editw->scrollBar = XtCreateManagedWidget("formScrollB", 
					   xmScrollBarWidgetClass, 
					   editw->meForm, args, cnt); 

  setThermometer(100, 100, 10);

  cnt = 0;
  XtSetArg(args[cnt], XmNheight, &sbHeight); cnt++;
  XtSetArg(args[cnt], XmNwidth, &sbWidth); cnt++;
  XtGetValues(editw->scrollBar, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNsliderSize, (int) ((float) sbHeight / (float) 10)); 
  cnt++;
  XtSetArg(args[cnt], XmNminimum, 0); cnt++;
  XtSetArg(args[cnt], XmNmaximum, sbHeight); cnt++;
  XtSetValues(editw->scrollBar, args, cnt);

  XtAddCallback(editw->scrollBar, XmNincrementCallback,     meDoScroll, 
		editw);
  XtAddCallback(editw->scrollBar, XmNdecrementCallback,     meDoScroll, 
		editw);
  XtAddCallback(editw->scrollBar, XmNdragCallback,          meDoScroll, 
		editw);
  XtAddCallback(editw->scrollBar, XmNpageDecrementCallback, meDoScroll, 
		editw);
  XtAddCallback(editw->scrollBar, XmNpageIncrementCallback, meDoScroll, 
		editw);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, editw->head->labelSet[0]); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, editw->scrollBar); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightOffset, 5); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNbottomOffset, 5); cnt++;

  XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;
  XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;

  scrlWidget = XtCreateManagedWidget("scrollW", 
				     xmDrawingAreaWidgetClass, 
				     editw->meForm, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNwidth, we - sbWidth); cnt++;
  XtSetArg(args[cnt], XmNheight, MAXHEIGTH); cnt++;
  XtSetArg(args[cnt], XmNpacking, XmPACK_TIGHT); cnt++;
  XtSetArg(args[cnt], XmNorientation, XmHORIZONTAL); cnt++;
  XtSetArg(args[cnt], XmNresizeWidth, False); cnt++;
  XtSetArg(args[cnt], XmNresizeHeight, False); cnt++;
  XtSetArg(args[cnt], XmNmarginWidth, 5); cnt++;

  editw->meRowCol = XtCreateManagedWidget("meRowCol", 
					  xmRowColumnWidgetClass,
					  scrlWidget, args, cnt);

  setThermometer(100, 100, 15);

  editw->form = createEditForm(editw);

  setThermometer(100, 100, 20);

  return(editw);
}

/* createHeadForm ************************************************************
 * create form which contains fields for the application information, the
 * menu name entry and the titles of entry columns
 ****************************************************************************/

static headerDescr *createHeadForm(me)
     meditWindow *me;
 {
  Arg args[20];
  register cnt = 0;
  headerDescr *form;
  Widget label, intForm;
  Display *d = XtDisplay(me->meShell);

  me->head = (headerDescr *) Malloc(sizeof(headerDescr));
  form = me->head;

  form->mFontPtr = exBaseFont(d);

  /******* create label "menu name" and corresponding entry ***********/

  form->MenuEntry = createFormEntry(me->meForm, "menu name:", E_ENTRYLEN1, 1);
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_POSITION); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetValues(form->MenuEntry->entry->entryWindow, args, cnt);  

  /************** create title labels for columns ***********************/

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, form->MenuEntry->entry->entryWindow); 
  cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR("item number/name:", XmSTRING_DEFAULT_CHARSET));
  cnt++;
  form->labelSet[0] = XtCreateManagedWidget("itemname", xmLabelGadgetClass, 
					    me->meForm, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, form->MenuEntry->entry->entryWindow); 
  cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, form->labelSet[0]); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR("return pattern:", XmSTRING_DEFAULT_CHARSET));
  cnt++;
  form->labelSet[1] = XtCreateManagedWidget("returnpattern", 
				   xmLabelGadgetClass, me->meForm, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, form->MenuEntry->entry->entryWindow); 
  cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, form->labelSet[1]); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR("request count:", XmSTRING_DEFAULT_CHARSET));
  cnt++;
  form->labelSet[2] = XtCreateManagedWidget("requestcount", 
			         xmLabelGadgetClass, me->meForm, args, cnt);

  /************** create form of application info ***********************/

  cnt = 0;
  XtSetArg(args[cnt], XmNborderWidth, 1); cnt++;
  XtSetArg(args[cnt], XmNverticalSpacing, 7); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, me->menuBar->menubar); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_POSITION); cnt++;

  form->applForm = XtCreateManagedWidget("applForm", xmFormWidgetClass, 
					  me->meForm, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR("MENU SET INFO", XmSTRING_DEFAULT_CHARSET));
  cnt++;
  label = XtCreateManagedWidget("menusetinfo", xmLabelGadgetClass, 
					    form->applForm, args, cnt);

  /* internal form for application info */
  cnt = 0;
  XtSetArg(args[cnt], XmNborderWidth, 0); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, label); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;

  intForm = XtCreateManagedWidget("internalForm1", xmFormWidgetClass, 
				  form->applForm, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR("menu set name: ", XmSTRING_DEFAULT_CHARSET));
  cnt++;
  form->setName[0] = XtCreateManagedWidget("menusetname", xmLabelGadgetClass, 
					   intForm, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, form->setName[0]); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR("menu set access: ", XmSTRING_DEFAULT_CHARSET));
  cnt++;
  form->setAccess[0] = XtCreateManagedWidget("menusetaccess", 
		                     xmLabelGadgetClass, intForm, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, form->setAccess[0]); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR("system menu set: ", XmSTRING_DEFAULT_CHARSET));
  cnt++;
  form->systemMenu[0] = XtCreateManagedWidget("systemmenuset", 
				      xmLabelGadgetClass, intForm, args, cnt);

  /* internal form for application info */
  cnt = 0;
  XtSetArg(args[cnt], XmNborderWidth, 0); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, label); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, intForm); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;

  intForm = XtCreateManagedWidget("internalForm2", xmFormWidgetClass, 
				  form->applForm, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR("", XmSTRING_DEFAULT_CHARSET));
  cnt++;
  form->setName[1] = XtCreateManagedWidget("setname", xmLabelGadgetClass, 
					   intForm, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, form->setName[1]); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR("", XmSTRING_DEFAULT_CHARSET));
  cnt++;
  form->setAccess[1] = XtCreateManagedWidget("setaccess", xmLabelGadgetClass, 
					     intForm, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget, form->setAccess[1]); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString, 
	   XmStringCreateLtoR("", XmSTRING_DEFAULT_CHARSET));
  cnt++;
  form->systemMenu[1] = XtCreateManagedWidget("systemset", xmLabelGadgetClass,
					      intForm, args, cnt);

  return(form);
}

/* createEditForm *************************************************************
 * initialize formular of menu editor
 *****************************************************************************/

static formDescr *createEditForm(me)
     meditWindow *me;
{ 
  Arg args[10];
  char str[10];
  register i, cnt = 0;

  me->form = (formDescr *) Malloc(sizeof(formDescr));

  me->form->entryCnt = ME_MAXITEM;
  for (i = 0; i < me->form->entryCnt; i++) {

    if (i < 9)
      sprintf(str, "%d: ", i + 1);
    else
      sprintf(str, "%d:", i + 1);

    me->form->ItemEntries[i] = createFormEntry(me->meRowCol, str, 
					       E_ENTRYLEN1, 1);
    me->form->ReturnEntries[i] = createFormEntry(me->meRowCol, "",
						 E_ENTRYLEN3, 1);
    cnt = 0;
    XtSetArg(args[cnt], XmNmarginLeft, 100); cnt++;
    XtSetArg(args[cnt], XmNlabelString, 
	                   XmStringCreateLtoR("0", XmSTRING_DEFAULT_CHARSET));
    cnt++;
    me->form->RequestLabels[i] = XtCreateManagedWidget("reqlabel", 
			         xmLabelGadgetClass, me->meRowCol, args, cnt);
  }

  me->form->formInfo = makeStxt(me->meShell, "missing information in form", 
				50, 10, me->form->entryCnt * 4 + 1); 

  return(me->form);
}

/* getmeMenuResult ***********************************************************
 * 
 ****************************************************************************/

static void getmeMenuResult(w, event, params, num_params)
     Widget w;
     XPropertyEvent *event;
     String *params;
     Cardinal *num_params;
{
  char *value;

  value = strCopy(getMenuEventReturnValue(event));

  free(value);
}


/* meDoScroll ***************************************************************
 * Callback procedure invoked whenever the user requests scrolling by
 * clicking over the scrollbar 
 ****************************************************************************/ 

static void meDoScroll(widget, closure, callData)
     Widget widget;
     caddr_t closure;
     caddr_t callData;
{
  XmScrollBarCallbackStruct *data = (XmScrollBarCallbackStruct *) callData;
  meditWindow *mEdit = (meditWindow *) closure;
  register cnt = 0;
  static int mx = 0, my = 0;
  static short unit = 0;
  Arg args[5];
  int maxy, maxDisplay = 9, maxHeight;

  if (unit <= 0) {
    unit = getFormEntryHeight(mEdit->form->ItemEntries[0]);
    
    if (unit <= 0) 
      unit = editorHeight(mEdit->form->ItemEntries[0]->entry->entryEditor);

    cnt = 0;
    XtSetArg(args[cnt], XmNheight, unit * ME_MAXITEM); cnt++;
    XtSetValues(mEdit->meRowCol, args, cnt);       /* calculate height */
  }

  maxHeight = maxDisplay * unit; 
  maxy = unit * ME_MAXITEM - maxHeight;

  switch(data->reason) {
  case XmCR_DRAG:
    if (data->value > 100 * (ME_MAXITEM - maxDisplay) / ME_MAXITEM)
      my = -maxy;
    else
      my = -unit * (data->value * ME_MAXITEM / 100);
    break;
  case XmCR_DECREMENT:
    if (my + unit <= 0)
      my = my + unit;
    break;
  case XmCR_INCREMENT:
    if (ABS(my - unit) < ABS(maxy))
      my = my - unit;
    break;
  case XmCR_PAGE_DECREMENT:
    if (my + maxHeight <= 0)
      my = my + maxHeight;
    break;
  case XmCR_PAGE_INCREMENT:
    if (ABS(my - maxHeight) < ABS(maxy))
      my = my - maxHeight;
    break;
  case XmCR_TO_BOTTOM:
    my = -maxy;
    break;
  case XmCR_TO_TOP:
    my = 0;
    break;
  }

  XtMoveWidget(mEdit->meRowCol, mx, my);
}

/* MenuEditorFont ***********************************************************
 * zoom menu editor window; (same entry positions in 'createEditEntries')
 ****************************************************************************/

static MenuEditorFont(editw)
meditWindow *editw;
{
  /* siehe alte Version in /Mrest/checkRest.c */
}

/* closeME **************************************************************
 * application function to inform ife that menu editor is iconized
 ****************************************************************************/
/*ARGSUSED*/
static void closeME(w, event, params, num_params)
     Widget w;
     XKeyEvent *event;
     String *params;
     Cardinal *num_params;
{
  exSendMsg(fileno(stdout), getpid(), IFE_CLOSE, 0);  
}

/* openME **************************************************************
 * application function to inform ife that menu editor is mapped
 ****************************************************************************/
/*ARGSUSED*/
static void openME(w, event, params, num_params)
     Widget w;
     XKeyEvent *event;
     String *params;
     Cardinal *num_params;
{
  exSendMsg(fileno(stdout), getpid(), IFE_OPEN, 0);  
}

/* MenuEditorQuit *********************************************************
 * this function is called upon receipt of a SIGHUP signal from the eshell in
 * order to terminate;
 ****************************************************************************/

static MenuEditorQuit()
{
  exit(0);
}

static eshHandler(d,e)
     Display *d;
     XErrorEvent *e;
{
  kill(getpid(), SIGQUIT);
}
